unit ChildForm;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses WhatsThis, Classes, Forms, MNode, ComCtrls;

type

  { TChildForm }

  TChildForm = class(THelpForm)
  private
    FBaseNode: TMNode;
    FCurrentMNode: TMNode;
    FCurrentMNodeName: string;
    FReStuffPending: Boolean;
    procedure SetBaseNode(const Value: TMNode);
    procedure SetCurrentMNode(const AValue: TMNode);
  protected
    procedure MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer); virtual;
    procedure ParameterChanged(MNode: TMNode; Index: Integer); virtual; abstract;
    procedure SettingChanged(MNode: TMNode; Index: Integer); virtual; abstract;
    procedure PropNameChanged(MNode: TMNode; Index: Integer); virtual;
    procedure Stuff(MNode: TMNode; Prefix: string; ParentNode: TTreeNode);
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; virtual; abstract;
    procedure ComboBoxChange(Sender: TObject);
    class  function Lockout: Integer;
    property CurrentMNode: TMNode read FCurrentMNode write SetCurrentMNode;
    property CurrentMNodeName: string read FCurrentMNodeName;
    procedure DoShow; override;
    procedure UpdateActions; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ReStuff; virtual;
    procedure EndUpdate(DoReStuff:Boolean = True; Postpone: Boolean = False);
    function BaseNodeClass: TClass; virtual; abstract;
    procedure ShowItem(MNode: TMNode); virtual; abstract;
    procedure StuffSelector(MNode: TMNode); virtual;
    property BaseNode: TMNode read FBaseNode write SetBaseNode;
  end;

procedure BeginUpdate;
procedure EndUpdate(Postpone: Boolean = false);

implementation

uses SysUtils, StdCtrls, Mixture;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

var
  FLockout: Integer; { Global Variable, so all forms locked }

{ TChildForm }

constructor TChildForm.Create(AOwner: TComponent);
begin
  inherited;
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
end;

destructor TChildForm.Destroy;
begin
  FBaseNode := nil;
  CurrentMNode := nil;
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  inherited;
end;

procedure TChildForm.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
  procedure NodeGone;
  begin
    if FBaseNode = Sender then
      FBaseNode := nil;
    if CurrentMNode = Sender then
      CurrentMNode := CurrentMNode.Owner as TMNode;
    ReStuff;
  end;
begin
  if not Visible then Exit;
  if (Change <> mnNewBaseNode)
     and (not Assigned(BaseNode) or (Sender.RootNode <> BaseNode.RootNode)) then Exit;
  case Change of
    mnDestroy: NodeGone;
    mnParameterChange:
               if Sender = CurrentMNode then
                 ParameterChanged(CurrentMNode, Index);
    mnSettingChange:
               if Sender = CurrentMNode then
                 SettingChanged(CurrentMNode, Index);
    mnNameChange, mnCreate:  ReStuff;
    mnNewBaseNode: if Sender is BaseNodeClass then
                     BaseNode := Sender;
    mnPropNameChange:
               if Sender = CurrentMNode then
                 PropNameChanged(CurrentMNode, Index);
  end;

end;

procedure TChildForm.ReStuff;
var
  MNode: TMNode;
begin
  FReStuffPending := True;
  CurrentMNode := CurrentMNode; { Force update of saved name }
  if FLockout > 0 then Exit;
  FReStuffPending := False;
  MNode := CurrentMNode;
  StuffSelector(BaseNode);
  if Assigned(MNode) and Assigned(BaseNode) then
    ShowItem(MNode);
end;

procedure TChildForm.SetBaseNode(const Value: TMNode);
begin
  if FBaseNode <> Value then begin
    CurrentMNode := nil;
    FBaseNode := Value;
  end;
  ReStuff;
end;

procedure TChildForm.SetCurrentMNode(const AValue: TMNode);
begin
  FCurrentMNode := AValue;
  if Assigned(CurrentMNode) then
    FCurrentMNodeName := CurrentMNode.UniqueName
  else
    FCurrentMNodeName := '??';
end;

procedure TChildForm.Stuff(MNode: TMNode; Prefix: string; ParentNode: TTreeNode);
var
  i: Integer;
  s, c: string;
  function OptState(MNode: TMNode): string;
  begin
    if Assigned(MNode) then
      Result := MNode.Name
    else
      Result := '?';
  end;
begin
  if not Assigned(MNode) or (csDestroying in MNode.ComponentState) then Exit;
  c := Copy(MNode.ClassName,2,255) + ': ';
  if MNode is TMixture then begin
    s := MNode.Name;
    StuffObject(c, s, MNode, ParentNode);
    Prefix := '';
  end else if MNode is TMNodePair then
    with TMNodePair(MNode) do begin
      s := '<' + OptState(Bra) + '|' + PairName + '|' +
               OptState(Ket) + '>';
      StuffObject(c + Prefix, s, MNode, ParentNode);
      Prefix := Prefix + s + ' - ';
    end
  else begin
    s := MNode.Name;
    StuffObject(c + Prefix, s, MNode, ParentNode);
    Prefix := Prefix + MNode.Name + ' - ';
  end;
  for i := 0 to MNode.Count-1 do
    Stuff(MNode.Items[i], Prefix, ParentNode);
end;

procedure TChildForm.StuffSelector(MNode: TMNode);
var
  E: Exception;
begin
  FLockout := 1;
  if Assigned(MNode) and not (csDestroying in MNode.ComponentState) then begin
    E := MNode.FixupProperties;
    if Assigned(E) then begin
      Application.Hint := E.Message;
      E.Free;
    end;
  end else
    CurrentMNode := nil;
  FLockout := 0;
end;

procedure TChildForm.ComboBoxChange(Sender: TObject);
var
  ComboBox: TComboBox;
begin
  ComboBox := Sender as TComboBox;
  Inc(FLockout);
  if FLockout = 1 then
    try
      if Assigned(FBaseNode) and (ComboBox.ItemIndex >= 0) then
        ShowItem(TMNode(ComboBox.Items.Objects[ComboBox.ItemIndex]))
      else
        ShowItem(nil)
    except
      Dec(FLockout);
      raise;
    end;
  Dec(FLockout);
end;

procedure TChildForm.PropNameChanged(MNode: TMNode; Index: Integer);
begin
 { Ignore }
end;

procedure BeginUpdate;
begin
  Inc(FLockout);
end;

procedure TChildForm.EndUpdate(DoReStuff: Boolean; Postpone: Boolean);
begin
  FReStuffPending := FReStuffPending or DoReStuff;
  ChildForm.EndUpdate(Postpone);
end;

procedure EndUpdate(Postpone: Boolean);
var
  i: Integer;
begin
  Dec(FLockout);
  if Postpone then Exit;
  if (FLockout = 0) then begin
    for i := 0 to Screen.FormCount-1 do
      if Screen.Forms[i] is TChildForm then
        if TChildForm(Screen.Forms[i]).FReStuffPending then
          TChildForm(Screen.Forms[i]).ReStuff;
  end;
end;

class function TChildForm.Lockout: Integer;
begin
  Result := FLockout;
end;

procedure TChildForm.DoShow;
begin
  FLockout := 0;
  inherited;
end;

procedure TChildForm.UpdateActions;
begin
  inherited;
  if FReStuffPending then
    ReStuff;
end;

end.
