unit GraphRange;

interface

uses Graphs;

type
  TPlotRange = record
  { TODO: May want to add x range in for completeness, though not currently used }
   y1, y2: Double; { Fraction of current map used }
   yi, yx: Double; { Corresponding plot range }
   Tag: Integer;
  end;
  TPlotRanges = array of TPlotRange;

procedure AddPlotRange(var PlotRanges: TPlotRanges; G: TGraphD; yu1, yu2, ya1, ya2: Double; ATag: Integer = 0);
function FindYRange(var YValue: Double; G:TgraphD; const PlotRanges: TPlotRanges): Integer;

implementation

uses Math, MathSup;

function YFraction(YValue: Double; G:TgraphD):Double;
begin
  if G.Ymax <> G.Ymin then
    Result := (YValue-G.Ymin)/(G.Ymax-G.Ymin)
  else
    Result := 0;
end;

function FractionToValue(Frac: Double; G: TGraphD; const PlotRange: TPlotRange):Double;
begin
  with PlotRange do
    if y2 <> y1 then
      Result := (Frac-y1)/(y2-y1)*(yx-yi) + yi
    else
      Result := yi { Arbitrary, but avoid divide by zero }
end;

procedure AddPlotRange(var PlotRanges: TPlotRanges; G: TGraphD; yu1, yu2, ya1, ya2: Double; ATag: Integer = 0);
begin
  SetLength(PlotRanges, Length(PlotRanges)+1);
  with PlotRanges[High(PlotRanges)] do begin
    { The actual fraction of the plot area used }
    y1 := YFraction(yu1, G);
    y2 := YFraction(yu2, G);
    { The corresponding mathematical plot range }
    if y2 > y1 then begin
      yi := ya1;
      yx := ya2;
    end else begin
      { Out of order }
      InOrder(y1, y2);
      yi := ya2;
      yx := ya1;
    end;
    Tag := ATag;
  end;
end;

function FindYRange(var YValue: Double; G:TgraphD; const PlotRanges: TPlotRanges): Integer;
var
  i: Integer;
  dist, distMin: Double;
begin
  YValue := YFraction(YValue, G);
  Result := -1;
  for i := High(PlotRanges) downto 0 do
    with PlotRanges[i] do
      if (YValue >= y1) and (YValue <= y2) then begin
        Result := i;
        Break;
      end else begin
       { keep track of closest range }
       dist := Min(Abs(YValue-y1), Abs(YValue-y2));
       if (Result < 0) or (dist < distMin) then begin
         distMin := dist;
         Result := i;
      end;
    end;
  if Result >= 0 then
    YValue := FractionToValue(YValue, G, PlotRanges[Result]);
end;

end.
