unit ImportACDD;
{ Import of data files from
  ACDD
  ASYTOP, A PROGRAM FOR DETAILED ANALYSIS OF GAS PHASE
  MAGNETIC RESONANCE SPECTRA OF ASYMMETRIC TOP MOLECULES.
  T.J. SEARS. REF. IN COMP. PHYS. COMMUN. 34 (1984) 123
}

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

interface

uses
  Classes, SysUtils, Mixture, TextStrm;

function ImportACDDFile(S: TFreeFormatStream): TMixture;

implementation

uses OAsymT, QuantumNumber;

function ImportACDDFile(S: TFreeFormatStream): TMixture;
var
  nManifolds: Integer;
  Species: TSpecies;
  Molecule: TAsymmetricMolecule;
  ManifoldL, ManifoldU: TAsymmetricManifold;
  Line: string;
  Origin: Double;
  Spin: TQuantumNumber;
  procedure ReadLine;
  begin
    Line := S.RestOfLine;
    S.NextLine;
  end;
  function ipar(n:Integer):Integer;
  var
    s: string;
  begin
    s := Trim(Copy(line,(n-1)*2+1,2));
    s := StringReplace(s, ' ', '', [rfReplaceAll]);
    if s = '' then
      Result := 0
    else
      Result := StrToInt(s);
  end;
  function par(n:Integer):Double;
  var
    s: string;
  begin
    s := Trim(Copy(line,(n-1)*10+1,10));
    s := StringReplace(s, ' ', '', [rfReplaceAll]);
    s := StringReplace(s, 'd', 'E', [rfReplaceAll, rfIgnoreCase]);
    if s = '' then
      Result := 0
    else
      Result := StrToFloat(s);
  end;
  procedure ReadManifold(Manifold: TManifold);
  var
    State: TAsymmetricTop;
  begin
    State := TAsymmetricTop.Create(Manifold);
    State.Name := 'X';
    State.S := Spin;
    ReadLine; { A1,B1,C1 - Upper }
    State.A := Par(1);
    State.B := Par(2);
    State.C := Par(3);
    ReadLine; { DN1,DNK1,DK1,DELN1,DELK1 }
    State.DJ  := Par(1);
    State.DJK := Par(2);
    State.DK  := Par(3);
    State.deltaJ := Par(4);
    State.deltaK := Par(5);
    ReadLine; { HN1,HNK1,HKN1,HK1,SHN1,SHNK1,SHK1 }
    State.HJ := Par(1);
    State.HJK := Par(2);
    State.HKJ := Par(3);
    State.HK := Par(4);
    State.phiJ := Par(5);
    State.phiJK := Par(6);
    State.phiK := Par(7);
    ReadLine; { LK1 }
    State.LK := Par(1);
    ReadLine; { TODO: ALPHA1,BETA1 }
    ReadLine; { EAA1,EBB1,ECC1,EAB1 }
    State.eaa := Par(1);
    State.ebb := Par(2);
    State.ecc := Par(3);
    State.eabbar := Par(4);
    ReadLine; { DKS1,DNPKS1,DNKS1,DNS1,DLKS1,DLNS1 }
    State.DsK := Par(1);
    State.DsKN := Par(2); { Read KN + NK }
    State.DsNK := Par(3);
    State.DsKN := State.DsKN - State.DsNK;
    State.DsN := Par(4);
    State.ds2 := Par(5);
    State.ds2 := Par(6);
    ReadLine; { TODO: HKS1 }
    ReadLine; { TODO: AF,AI,BI,ABI }
    ReadLine; { TODO: AQ,BQ }
    ReadLine; { TODO: GAA1,GBB1,GCC1,GAB1,GN1 }
    ReadLine; { TODO: GRA1,GRB1,GRC1 }
  end;
begin
  Result := TMixture.Create(nil);
  try
    Result.Name := 'Mixture';
    Species := TSpecies.Create(Result);
    Species.Name := 'S';
    Molecule := TAsymmetricMolecule.Create(Species);
    Molecule.Name := 'M';
    Molecule.SReduction := False;
    ReadLine; { Skip IDN,IDK,IDDK,I2S,I2I }
    Spin := Ipar(4);
    ReadLine; { Skip IFLD }
    ReadLine; { XNU }
    ManifoldL := TAsymmetricManifold.Create(Molecule);
    ManifoldL.Name := 'Lower';
    ManifoldL.Initial := True;
    Origin := Par(1);
    if Origin <> 0 then begin
      ManifoldU := TAsymmetricManifold.Create(Molecule);
      ManifoldU.Name := 'Upper';
    end else
      ManifoldU := ManifoldL;
    ReadLine; { TEMP }
    Result.Temperature := Par(1);
    ReadManifold(ManifoldU);
    ManifoldU.State[0].Origin := Origin;
    if ManifoldL <> ManifoldU then begin
      ReadManifold(ManifoldL);
      ManifoldU.State[0].Name := 'A';
    end;
    { ILSQS }
    { IPRED }
    { IEIGS }
    { NMAX,NMIN }
    { Fake a transition moment }
    TCartesianTransitionMoment.Create(TTransitionMoments.Create(Molecule, ManifoldL.Name, ManifoldU.Name),
                                      ManifoldL.State[0].Name, ManifoldU.State[0].Name).Strength := 1;
  except
    Result.Free;
    Raise;
  end;
end;

end.
