unit Main;
{$I DFS.inc}
{
PGOPHER - a program for simulating rotational structure
Copyright (C) 2005 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}

interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ToolWin, ComCtrls, ToolBarEx, StdCtrls, ExtCtrls,
  ScrlCtrl, Graphs, Mixture, Buttons, SpectrumPlot, Overlay, AlEdit,
  MRUFList, ProjAbout, ExActn, ActnList, Split, Vector, MNode, GPenDlg,
  PrintPre, TextStrm, WhatsThis, GraphRange;

type

  TGraphStyle = (gsNoOffset, gsOffset, gsOffsetAll, gsInverted, gsInvertedAll);

  TMixtureReader = function (Stream: TFreeFormatStream):TMixture;

  { TMainForm }

  TMainForm = class(THelpForm)
    ToolBar: TToolBarEx;
    MainMenu: TMainMenu;
    ViewMenu: TMenuItem;
    LogItem: TMenuItem;
    G: TGraphD;
    FileMenu: TMenuItem;
    ExitItem: TMenuItem;
    PrintSetupItem: TMenuItem;
    PrintItem: TMenuItem;
    N2: TMenuItem;
    SaveAsItem: TMenuItem;
    SaveItem: TMenuItem;
    OpenItem: TMenuItem;
    NewItem: TMenuItem;
    SimulateButton: TSpeedButton;
    AllButton: TSpeedButton;
    XZoomButton: TSpeedButton;
    LeftButton: TSpeedButton;
    RightButton: TSpeedButton;
    CompressButton: TSpeedButton;
    ExpandButton: TSpeedButton;
    ZoomButton: TSpeedButton;
    ZoomOutButton: TSpeedButton;
    ResetZoomButton: TSpeedButton;
    Foffset: TAlignedEdit;
    Label1: TLabel;
    EditMenu: TMenuItem;
    N4: TMenuItem;
    CopyItem: TMenuItem;
    ClearOverlayItem: TMenuItem;
    PasteOverlayItem: TMenuItem;
    PasteAnotherItem: TMenuItem;
    ClearLogItem: TMenuItem;
    PasteOverlayButton: TSpeedButton;
    GraphPrintPreview: TGraphPrintPreview;
    HelpMenu: TMenuItem;
    AboutItem: TMenuItem;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    Label2: TLabel;
    Lorentzian: TAlignedEdit;
    Label3: TLabel;
    Gaussian: TAlignedEdit;
    Label4: TLabel;
    Temperature: TAlignedEdit;
    OpenButton: TSpeedButton;
    MRUFileList: TdfsMRUFileList;
    FileExitSeparator: TMenuItem;
    ConstantsItem: TMenuItem;
    LineListItem: TMenuItem;
    OC: TGraphD;
    Split: TSplit;
    StatusBar: TStatusBar;
    OverlaysItem: TMenuItem;
    ActionList: TActionList;
    ExCopy: TExCopy;
    LoadOverlayItem: TMenuItem;
    OpenOverlayDialog: TOpenDialog;
    CopyPictureItem: TMenuItem;
    ExPaste: TExPaste;
    PasteItem: TMenuItem;
    ShowPartsButton: TSpeedButton;
    ShowSumButton: TSpeedButton;
    ShowFortratButton: TSpeedButton;
    StatesItem: TMenuItem;
    TransitionsItem: TMenuItem;
    ClearCacheItem: TMenuItem;
    LevelsItem: TMenuItem;
    PlotStyleButton: TSpeedButton;
    ImportItem: TMenuItem;
    ImportDialog: TOpenDialog;
    StylesItem: TMenuItem;
    GraphPenDialog: TGraphPenDialog;
    EmptyMixtureItem: TMenuItem;
    NewLinearMoleculeItem: TMenuItem;
    OldPgopherItem: TMenuItem;
    OldD3hItem: TMenuItem;
    ExportItem: TMenuItem;
    ExportPlotItem: TMenuItem;
    ExportPlotDialog: TSaveDialog;
    TimingItem: TMenuItem;
    OverlaysMenu: TMenuItem;
    PasteOverlay1Item: TMenuItem;
    OverlayPropertiesItem: TMenuItem;
    InvertItem: TMenuItem;
    DeleteLastPeakItem: TMenuItem;
    N1: TMenuItem;
    ClearAllPeaksItem: TMenuItem;
    SelectRegionItem: TMenuItem;
    ClearSelectionItem: TMenuItem;
    WhatIsItem: TMenuItem;
    IndexItem: TMenuItem;
    SearchItem: TMenuItem;
    ProceduresItem: TMenuItem;
    UsingHelpItem: TMenuItem;
    PlotMenu: TMenuItem;
    LeftItem: TMenuItem;
    RightItem: TMenuItem;
    CompressItem: TMenuItem;
    ExpandItem: TMenuItem;
    PlotSelectionItem: TMenuItem;
    AllItem: TMenuItem;
    N3: TMenuItem;
    MagnifyPlotItem: TMenuItem;
    ShrinkplotItem: TMenuItem;
    ResetZoomItem: TMenuItem;
    N5: TMenuItem;
    SimulateItem: TMenuItem;
    NextPlotstyleItem: TMenuItem;
    ShowPartsItem: TMenuItem;
    ShowSumItem: TMenuItem;
    ShowFortratItem: TMenuItem;
    DumpItem: TMenuItem;
    CALPGMfileItem: TMenuItem;
    ExportLineListItem: TMenuItem;
    ExportLineListDialog: TSaveDialog;
    UnitsItem: TMenuItem;
    WavenumbersItem: TMenuItem;
    MHzItem: TMenuItem;
    KelvinItem: TMenuItem;
    eVItem: TMenuItem;
    nmVacuumItem: TMenuItem;
    nmAirItem: TMenuItem;
    FortratItem: TMenuItem;
    UpperStateItem: TMenuItem;
    Project: TProject;
    N6: TMenuItem;
    MergeItem: TMenuItem;
    BinaryfileItem: TMenuItem;
    procedure BinaryfileItemClick(Sender: TObject);
    procedure UpperStateItemClick(Sender: TObject);
    procedure FoffsetKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure WavenumbersItemClick(Sender: TObject);
    procedure ExportLineListItemClick(Sender: TObject);
    procedure CALPGMfileItemClick(Sender: TObject);
    procedure MRUFileListGetDisplayName(Sender: TObject; AFilename: string;
      var ADisplayName: string);
    procedure StatusBarHint(Sender: TObject);
    procedure GContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure DumpItemClick(Sender: TObject);
    procedure ShowFortratItemClick(Sender: TObject);
    procedure ShowSumItemClick(Sender: TObject);
    procedure ShowPartsItemClick(Sender: TObject);
    procedure UsingHelpItemClick(Sender: TObject);
    procedure ProceduresItemClick(Sender: TObject);
    procedure SearchItemClick(Sender: TObject);
    procedure IndexItemClick(Sender: TObject);
    procedure WhatIsItemClick(Sender: TObject);
    procedure ClearSelectionItemClick(Sender: TObject);
    procedure SelectRegionItemClick(Sender: TObject);
    procedure LogItemClick(Sender: TObject);
    procedure SimulateButtonClick(Sender: TObject);
    procedure AllButtonClick(Sender: TObject);
    procedure XZoomButtonClick(Sender: TObject);
    procedure LeftButtonClick(Sender: TObject);
    procedure RightButtonClick(Sender: TObject);
    procedure CompressButtonClick(Sender: TObject);
    procedure ExpandButtonClick(Sender: TObject);
    procedure ZoomButtonClick(Sender: TObject);
    procedure ZoomOutButtonClick(Sender: TObject);
    procedure ResetZoomButtonClick(Sender: TObject);
    procedure GMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ClearLogItemClick(Sender: TObject);
    procedure ClearOverlayItemClick(Sender: TObject);
    procedure PasteAnotherItemClick(Sender: TObject);
    procedure PasteOverlayItemClick(Sender: TObject);
    procedure ExitItemClick(Sender: TObject);
    procedure PrintItemClick(Sender: TObject);
    procedure PrintSetupItemClick(Sender: TObject);
    procedure AboutItemClick(Sender: TObject);
    procedure OpenItemClick(Sender: TObject);
    procedure SaveAsItemClick(Sender: TObject);
    procedure MRUFileListMRUItemClick(Sender: TObject; AFilename: String);
    procedure ConstantsItemClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure LineListItemClick(Sender: TObject);
    procedure SaveItemClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure OverlaysItemClick(Sender: TObject);
    procedure FoffsetChange(Sender: TObject);
    procedure CopyItemClick(Sender: TObject);
    procedure LoadOverlayItemClick(Sender: TObject);
    procedure CopyPictureItemClick(Sender: TObject);
    procedure ShowSumButtonClick(Sender: TObject);
    procedure StatesItemClick(Sender: TObject);
    procedure TransitionsItemClick(Sender: TObject);
    procedure ClearCacheItemClick(Sender: TObject);
    procedure GMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure LevelsItemClick(Sender: TObject);
    procedure PlotStyleButtonClick(Sender: TObject);
    procedure StylesItemClick(Sender: TObject);
    procedure EmptyMixtureItemClick(Sender: TObject);
    procedure NewLinearMoleculeItemClick(Sender: TObject);
    procedure OldPgopherItemClick(Sender: TObject);
    procedure OldD3hItemClick(Sender: TObject);
    procedure ExportPlotItemClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TimingItemClick(Sender: TObject);
    procedure InvertItemClick(Sender: TObject);
    procedure DeleteLastPeakItemClick(Sender: TObject);
    procedure ClearAllPeaksItemClick(Sender: TObject);
  private
    FMixture: TMixture;
    FLastGraph: TGraphD;
    FLastPlot: TPlot;
    FSelectedLineList: Boolean;
    GraphStyle: TGraphStyle;
    FLastLineInfo: string;
    OvrYl, OvrYu: Double; { Map Y range when overlays plotted }
    SpectrumPlotRanges, FortratPlotRanges: TPlotRanges;
    FCalculationInProgress: Boolean;
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
    procedure FormResize(Sender: TObject);
    procedure SetLineList(const Value: TLineList);
    procedure SetMixture(AMixture: TMIxture);
    procedure ChangeMixtureSetting(AIndex: Integer; Sender: TObject);
    function GetLineList: TLineList;
    function LastGraph: TGraphD;
    function LastPlot(WithPeak: Boolean = False):TPlot;
    function ShowYValue(GD: TgraphD; const PlotRanges: TPlotRanges):string;
  public
    Overlays: TOverlays;
    LastObsWasLineForm: Boolean;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddOverlay(FileName: string; Merge:Boolean);
    function ClearOverlays(Prompt: Boolean):Boolean;
    function QueryDiscardOverlays: Boolean;
    procedure LoadFile(AFileName: string; Merge: Boolean);
    procedure SetName(AFileName: string);
    procedure UpdateToolbar;
    procedure AddOverlayParameters;
    procedure ContourFit;
    procedure ShowContour(Spectra: TOverlays; Simulation: TColumnVector);
    procedure ShowSimulation;
    procedure LoadGeneralFile(AFileName: string; Merge: Boolean);
    procedure ImportFile(DefaultProc: TMixtureReader); overload;
    procedure ImportFile(DefaultProc: TMixtureReader; AFileName: string; Merge: Boolean); overload;
    procedure DoPlot(All: Boolean = False; PurgeLineSelection: Boolean = True);
    procedure ConsiderReplot(All: Boolean = False; PurgeLineSelection: Boolean = True);
    procedure LoadAnyObservations;
    procedure DoLineFit;
    function BeginCalculation:Boolean;
    procedure EndCalculation;
    function FindLimits:Boolean;
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
    property CalculationInProgress: Boolean read FCalculationInProgress;
    property Mixture: TMixture read FMixture;
    property SelectedLineList: Boolean read FSelectedLineList;
    property LineList: TLineList read GetLineList write SetLineList;
  end;

var
  MainForm: TMainForm;

implementation

{$DEFINE OffRecyclerMM}

{$IFDEF VCL_}
{$R version.res}
{$ENDIF}
uses {$IFDEF LCL_} LResources, AlignHelp, LCLType, {$ENDIF}
  {$IFDEF VCL_} PHtmlHelpViewer, {$ENDIF}
  {$IFDEF MSWINDOWS} ShellAPI, {$ENDIF}
  {$IFDEF RecyclerMM} FRMMSnapShot, {$ENDIF}
  {$IFDEF HELPHELP} MakeHelp, {$ENDIF}
  Math, Huge, ClipBrd, LogUnit, Profile, ExpPlot, ExpBitmap,
  Obs, ConstFrm, LineFrm, MathSup, Fitter, OvrlyFrm, StateInfo,
  TranInfo, LevelFrm, PgopherImport, D3hImport, ImportACDD, ImportCALPGM,
  ExpCan, ExportG, Units, Progress, WCompat, ProfileForm, LineExportOpt,
  ReadBinary, ChildForm;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

procedure ReShow(Form: TCustomForm);
begin
  if not Form.Visible then
    Form.Show;
  if Form.WindowState in [wsMinimized, wsMaximized] then
    Form.WindowState := wsNormal;
  Form.BringToFront;
end;

procedure TMainForm.LogItemClick(Sender: TObject);
begin
  ReShow(LogForm);
end;

procedure TMainForm.SimulateButtonClick(Sender: TObject);
begin
  DoPlot;
end;

function TMainForm.FindLimits: Boolean;
  procedure AddLimits(FMin, FMax: Double);
  var
    t: Double;
  begin
    if FMin > FMax then begin
      t := FMin;
      FMin := FMax;
      FMax := t;
    end;
    if not Result then begin
      Mixture.Fmin := Fmin;
      Mixture.Fmax := Fmax;
      Result := True;
    end else begin
      if Fmin < Mixture.Fmin then Mixture.Fmin := Fmin;
      if Fmax > Mixture.Fmax then Mixture.Fmax := Fmax;
    end;
  end;
begin
  Result := False;
  if not Assigned(Mixture) then Exit;
  if Assigned(LineList) and (LineList.nLines > 0) then
    AddLimits(Convert(LineList.FMin,Mixture.Units,Mixture.PlotUnits), Convert(LineList.FMax,Mixture.Units,Mixture.PlotUnits));
  if Assigned(Overlays) then begin
    Overlays.FixupProperties;
    AddLimits( Convert(Overlays.HMin,Overlays.Units,Mixture.PlotUnits) +Mixture.Foffset,
               Convert(Overlays.HMax,Overlays.Units,Mixture.PlotUnits)+Mixture.Foffset);
  end;
  if Mixture.Fmin = Mixture.Fmax then begin
    Mixture.Fmin := Mixture.Fmin - 0.5;
    Mixture.Fmax := Mixture.Fmax + 0.5;
  end;
end;

procedure TMainForm.AllButtonClick(Sender: TObject);
begin
  DoPlot(True)
end;

procedure TMainForm.XZoomButtonClick(Sender: TObject);
begin
  if not LastGraph.AreaSelected then Exit;
  Mixture.Fmin := LastGraph.SelectedMathArea.Left + Mixture.Foffset;
  Mixture.Fmax := LastGraph.SelectedMathArea.Right + Mixture.Foffset;
  ShowSimulation;
end;

procedure TMainForm.LeftButtonClick(Sender: TObject);
var
  d: Double;
begin
  d := (Mixture.Fmax-Mixture.Fmin)/2;
  Mixture.Fmin := Mixture.Fmin - d;
  Mixture.Fmax := Mixture.Fmax - d;
  ShowSimulation;
end;

procedure TMainForm.RightButtonClick(Sender: TObject);
var
  d: Double;
begin
  d := (Mixture.Fmax-Mixture.Fmin)/2;
  Mixture.Fmin := Mixture.Fmin + d;
  Mixture.Fmax := Mixture.Fmax + d;
  ShowSimulation;
end;

procedure TMainForm.CompressButtonClick(Sender: TObject);
var
  d: Double;
begin
  d := (Mixture.Fmax-Mixture.Fmin)/4;
  Mixture.Fmin := Mixture.Fmin - d;
  Mixture.Fmax := Mixture.Fmax + d;
  ShowSimulation;
end;

procedure TMainForm.ExpandButtonClick(Sender: TObject);
var
  d: Double;
begin
  d := (Mixture.Fmax-Mixture.Fmin)/4;
  Mixture.Fmin := Mixture.Fmin + d;
  Mixture.Fmax := Mixture.Fmax - d;
  ShowSimulation;
end;

procedure TMainForm.ZoomButtonClick(Sender: TObject);
begin
  if LastGraph.AreaSelected then
    LastGraph.ZoomSelected(True);
end;

procedure TMainForm.ZoomOutButtonClick(Sender: TObject);
begin
  LastGraph.ZoomOut;
end;

procedure TMainForm.ResetZoomButtonClick(Sender: TObject);
begin
  LastGraph.ResetZoom;
end;

procedure TMainForm.GMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Range, P: Double;
  Area: TDRect;
  procedure BackTransform(var y: Double);
  begin
    { Convert current Y co-ordinate to equivalent y co-ordinate when overlays plotted }
    y := (y - G.Ymin)/(G.Ymax-G.Ymin)*(OvrYu-OvrYl) + OvrYl;
  end;
begin
  if (ssAlt in Shift) and (Button = mbLeft) and LastGraph.AreaSelected then begin
    Mixture.Foffset := Mixture.Foffset + LastGraph.SelectedMathArea.Right - LastGraph.SelectedMathArea.Left;
    Foffset.Modified := False;
    ShowSimulation;
  end else if (Button = mbRight) then
    if (not LastGraph.AreaSelected or LastGraph.CloseToX(LastGraph.SelectedMathArea.Left, LastGraph.SelectedMathArea.Right))
       and Assigned(LineList) then begin
      Range := Max(Mixture.Lorentzian + Mixture.Gaussian, 3.0*LastGraph.PixelSize(True));
      LineList.FindLines(LastGraph.LastMouseX+Mixture.Foffset, Range, Mixture.PlotUnits);
      LineForm.ShowLines(Mixture.Temperature, Mixture.Tvib);
    end else if (LastGraph = G) and LastGraph.AreaSelected and Assigned(Overlays) then begin
      Area := G.SelectedMathArea;
      BackTransform(Area.Top);
      BackTransform(Area.Bottom);
      if Overlays.FindPeak(Area, P, FLastPlot, Mixture.PlotUnits) then begin
        ShowSimulation;
        if LineForm.AutoCheck.Checked then
          LineForm.AssignButtonClick(nil);
      end;
    end;
end;

destructor TMainForm.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  ClearOverlays(False);
  SetMixture(nil);
  inherited;
end;

procedure TMainForm.ClearLogItemClick(Sender: TObject);
begin
  LogForm.ClearButtonClick(nil);
end;

procedure TMainForm.ExitItemClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.PrintItemClick(Sender: TObject);
begin
  GraphPrintPreview.Show;
end;

procedure TMainForm.PrintSetupItemClick(Sender: TObject);
begin
  GraphPrintPreview.PrinterSetup;
end;

procedure TMainForm.AboutItemClick(Sender: TObject);
begin
  Project.Execute;
end;

procedure TMainForm.OpenItemClick(Sender: TObject);
begin
  if CloseQuery and OpenDialog.Execute then
    LoadGeneralFile(OpenDialog.FileName,MergeItem.Checked);
end;

procedure TMainForm.SaveAsItemClick(Sender: TObject);
begin
  if not Assigned(Mixture) then Exit;
  SaveDialog.FileName := OpenDialog.FileName;
  if SaveDialog.Execute then begin
    Mixture.SaveToFile(SaveDialog.FileName);
    SetName(SaveDialog.FileName);
  end;
end;

procedure TMainForm.MRUFileListMRUItemClick(Sender: TObject; AFilename: String);
begin
  if LowerCase(ExtractFileExt(AFilename)) = '.pgo' then begin
    if CloseQuery then
      LoadFile(AFilename,MergeItem.Checked);
  end else
    LoadGeneralFile(AFileName,MergeItem.Checked);
  ConsiderReplot;
end;

procedure TMainForm.LoadFile(AFileName: string; Merge: Boolean);
var
  NewMixture: TMixture;
  Stream: TFileStream;
begin
  if Merge and Assigned(Mixture) and (Mixture.nMolecules > 0) then begin
    Stream := TFileStream.Create(AFileName, fmOpenRead);
    try
      BeginUpdate;
      Mixture.MergeFromStream(Stream, AFileName);
      SetName('');
    finally
      EndUpdate(True);
      Stream.Free;
    end;
  end else begin
    NewMixture := TMixture.Create(nil);
    try
      NewMixture.LoadFromFile(AFilename);
      SetMixture(NewMixture);
      SetName(AFileName);
    except
      NewMixture.Free;
      raise;
    end;
  end;
  ClearObservations;
  ConsiderReplot;
end;

procedure TMainForm.SetName(AFileName: string);
begin
  OpenDialog.FileName := AFileName;
  if AFileName = '' then
    Caption := 'Pgopher'
  else begin
    Caption := AFileName + ' - Pgopher';
    MRUFileList.AddItem(ExpandFileName(AFileName));
  end;
end;

procedure TMainForm.ConstantsItemClick(Sender: TObject);
begin
  if not Assigned(Mixture) then Exit;
  ReShow(ConstantsForm);
  ConstantsForm.BaseNode := Mixture;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  s: string;
begin
  CanClose := not Assigned(Mixture) or not Mixture.Modified;
  if CanClose then Exit;
  s := Mixture.Name;
  if s = '' then s := 'Mixture';
  CanClose := MessageDlg(Concat(s, ' has changed: discard?'), mtWarning, mbOkCancel, 0) = mrOK ;
end;

procedure TMainForm.LineListItemClick(Sender: TObject);
begin
  ReShow(LineForm);
end;

procedure TMainForm.SaveItemClick(Sender: TObject);
begin
  if OpenDialog.FileName = '' then
    SaveAsItemClick(nil)
  else
    Mixture.SaveToFile(OpenDialog.FileName);
end;

procedure TMainForm.AddOverlayParameters;
begin
  if Assigned(Overlays) then
    Mixture.AddNodeParameters(Overlays);
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Temperature.Tag := TemperatureIndex;
  Gaussian.Tag := GaussianIndex;
  Lorentzian.Tag := LorentzianIndex;
  FOffset.Tag := FOffsetIndex;
  ShowSumButton.Tag := ShowSumIndex;
  ShowPartsButton.Tag := ShowPartsIndex;
  ShowFortratButton.Tag := ShowFortratIndex;
  Split.OneWindowVisible(G);
  Split.Visible := False;
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  OnResize := {$IFDEF FPC}@{$ENDIF}FormResize; { LAZARUS bug work-around }
  DragAcceptFiles(Handle, True);
  {$IFDEF LCL}
  Application.OnHint := @StatusBarHint;
  AlControl(nil);
  AlControl(Foffset);
  AlControl(Label4);
  AlControl(Temperature);
  AlControl(Label3);
  AlControl(Gaussian);
  AlControl(Label2);
  AlControl(Lorentzian);
  AlForm;
  {$ENDIF}
  StatusBar.SimplePanel := False; { Defaults incompatible between VCL and LCL }
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  Split.ParentResized;
end;

procedure TMainForm.GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  GD: TGraphD;
begin
  GD := Sender as TGraphD;
  StatusBar.Panels[0].Text := RoundedStr(GD.LastMouseX, (GD.Xmax-GD.Xmin)/Width);
  if GD <> G then
    StatusBar.Panels[1].Text := ShowYValue(GD, FortratPlotRanges)
  else
    StatusBar.Panels[1].Text := ShowYValue(G, SpectrumPlotRanges);
  if GD.AreaSelected then
    with GD.SelectedMathArea do begin
      StatusBar.Panels[3].Text := RoundedStr(Right-Left, (GD.Xmax-GD.Xmin)/Width);
      StatusBar.Panels[4].Text := RoundedStr(Top-Bottom, (GD.Ymax-GD.Ymin)/Height);
    end
  else begin
    StatusBar.Panels[3].Text := '';
    StatusBar.Panels[4].Text := '';
  end;
end;

procedure TMainForm.ContourFit;
var
  i: Integer;
begin
  if not Assigned(Overlays) then
    raise EReportException.Create('Overlays required');
  if BeginCalculation then Exit;
  try
    PushProgress('Contour Fit', Mixture.FitCycles);
    for i := 1 to Mixture.FitCycles do begin
      SetProgress(i);
      Mixture.FillParameters;
      AddOverlayParameters;
      Mixture.PushParameters(ftContour);
      try
        FitContour(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowContour, Mixture.FittingParameters);
      except
        Mixture.DiscardFit;
        raise;
      end;
    end;
    PopProgress;
  finally
    EndCalculation;
    ResetProgress;
  end;
end;

procedure TMainForm.ShowContour(Spectra: TOverlays; Simulation: TColumnVector);
var
  i, j, l: Integer;
  v: Double;
  dmin, dmax: Double; { Limits of obs-calc }
  omin, omax: Double;
  XLabel: string;
begin
  SetLength(FortratPlotRanges,0);
  SetLength(SpectrumPlotRanges,0);
  Split.Visible := True;
  Split.MakeBothVisible;
  { Determine Limits on obs-calc, obs }
  l := 1;
  for i := 0 to Spectra.Count-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      for j := StartPoint to EndPoint do begin
        v := Intensity[j];
        if l = 1 then begin
          omin := v;
          omax := v;
        end else begin
          if v < omin then omin := v;
          if v > omax then omax := v;
        end;
        v := v - Simulation[l];
        if l = 1 then begin
          dmin := v;
          dmax := v;
        end else begin
          if v < dmin then dmin := v;
          if v > dmax then dmax := v;
        end;
        Inc(l);
      end;
  if l = 1 then Exit; { No points }
  { Plot obs - calc }
  OC.Start;
  OC.Map(Mixture.Fmin, Mixture.Fmax, dmin, dmax);
  OC.Axes('', 'Obs-Calc');
  l := 1;
  for i := 0 to Spectra.Count-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      for j := StartPoint to EndPoint do begin
        v := Intensity[j] - Simulation[l];
        if l = 1 then
          OC.ClipPoint(Frequency[j], v)
        else
          OC.ClipJoin(Frequency[j], v);
        Inc(l);
      end;
  OC.Finish;
  G.Start;
  G.Map(Mixture.Fmin, Mixture.Fmax, omin, omax);
  XLabel := Mixture.UnitsTypeName + '/' + Mixture.UnitsName;
  G.Axis(Mixture.Fmin, Mixture.Fmax, omin, true, XLabel);
  l := 1;
  for i := 0 to Spectra.Count-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      for j := StartPoint to EndPoint do begin
        v := Simulation[l];
        if l = 1 then
          G.ClipPoint(Frequency[j], v)
        else
          G.ClipJoin(Frequency[j], v);
        Inc(l);
      end;
  G.Style(1);
  l := 1;
  for i := 0 to Spectra.Count-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      for j := StartPoint to EndPoint do begin
        v := Intensity[j];
        if l = 1 then
          G.ClipPoint(Frequency[j], v)
        else
          G.ClipJoin(Frequency[j], v);
        Inc(l);
      end;
  G.Finish;
end;

function TMainForm.ClearOverlays(Prompt: Boolean):Boolean;
begin
  if Prompt and not QueryDiscardOverlays then begin
    Result := False;
    Exit;
  end;
  Result := True;
  BeginUpdate;
  try
    FreeAndNil(Overlays);
  finally
    EndUpdate(True);
  end;
end;

function TMainForm.QueryDiscardOverlays: Boolean;
var
  s: string;
begin
  if Assigned(Overlays) and Overlays.Modified then begin
    s := Overlays.Name;
    if s = '' then s := 'Overlay';
    if MessageDlg(Concat(s, ' has changed: discard?'), mtWarning, mbOkCancel, 0) <> mrOK then begin
      Result := False;
      Exit;
    end;
  end;
  Result := True;
end;

procedure TMainForm.AddOverlay(FileName: string; Merge:Boolean);
begin
  if not Merge and not QueryDiscardOverlays then
    Exit;
  try
    BeginUpdate;
    AddToOverlay(Overlays, FileName, Merge);
  finally
    EndUpdate(True);
  end;
  MRUFileList.AddItem(ExpandFileName(FileName));
end;

procedure TMainForm.ClearOverlayItemClick(Sender: TObject);
begin
  if ClearOverlays(True) then
    ShowSimulation;
end;

procedure TMainForm.PasteAnotherItemClick(Sender: TObject);
begin
  AddOverlay('',True);
  ShowSimulation;
end;

procedure TMainForm.LoadOverlayItemClick(Sender: TObject);
var
  i: Integer;
begin
  OpenOverlayDialog.Filter :=
    'Data File (*.dat)|*.dat|Overlay File (*.ovr)|*.ovr|'
    +PlotLoaderFilter
    +GraphicFilter(TGraphic)
    +'All Files (*.*)|*.*';
  if OpenOverlayDialog.Execute then begin
    for i := 0 to OpenOverlayDialog.Files.Count-1 do
      LoadGeneralFile(OpenOverlayDialog.Files[i],MergeItem.Checked);
    ShowSimulation;
  end;
end;

procedure TMainForm.PasteOverlayItemClick(Sender: TObject);
begin
  if not MergeItem.Checked then
    if not ClearOverlays(True) then
      Exit;
  PasteAnotherItemClick(nil);
end;

procedure TMainForm.OverlaysItemClick(Sender: TObject);
begin
  ReShow(OverlayForm);
  OverlayForm.BaseNode := Overlays;
end;

procedure TMainForm.MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
begin
  if Change = mnGraphicChanged then
    G.PlotChanged;
  if Sender <> Mixture then Exit;
  if (Change = mnNewBaseNode)
     or ((Change = mnParameterChange)
          and (Index in [TemperatureIndex, GaussianIndex, LorentzianIndex, FoffsetIndex]) )
     or ((Change = mnSettingChange)
         and (Index in [ShowSumIndex, ShowPartsIndex, ShowFortratIndex, ShowFortratIndex+1])) then
    UpdateToolbar;
end;

procedure TMainForm.UpdateToolbar;
var
  i: Integer;
  procedure UpdateEditControl(Edit: TCustomEdit; Value: Double);
  begin
    if not Edit.Focused then
      Edit.Text := FloatToStr(Value);
  end;
begin
  UpdateEditControl(Temperature, Mixture.Temperature);
  UpdateEditControl(Gaussian, Mixture.Gaussian);
  UpdateEditControl(Lorentzian, Mixture.Lorentzian);
  UpdateEditControl(Foffset, Mixture.Foffset);
  ShowSumButton.Down := Mixture.ShowSum;
  ShowPartsButton.Down := Mixture.ShowParts;
  ShowFortratButton.Down := Mixture.ShowFortrat;
  ShowSumItem.Checked := Mixture.ShowSum;
  ShowPartsItem.Checked := Mixture.ShowParts;
  ShowFortratItem.Checked := Mixture.ShowFortrat;
  UpperStateItem.Checked := Mixture.UseUpper;
  for i := 0 to UnitsItem.Count-1 do
    if UnitsItem.Items[i].Tag = Ord(Mixture.PlotUnits) then
      UnitsItem.Items[i].Checked := True;
end;

procedure TMainForm.FoffsetChange(Sender: TObject);
var
  NText: string;
  Value: Double;
begin
  NText := Trim((Sender as TEdit).Text);
  if Assigned(Mixture) and (NText <> '') and TryStrToFloat(NText, Value) then
    Mixture.SetParameter(TComponent(Sender).Tag, Value);
end;

procedure TMainForm.ShowSumButtonClick(Sender: TObject);
begin
  ChangeMixtureSetting(TComponent(Sender).Tag, Sender);
end;

procedure TMainForm.LoadGeneralFile(AFileName: string; Merge: Boolean);
var
  Ext: string;
begin
  Ext := LowerCase(ExtractFileExt(AFileName));
  if (Ext = '.pgo') or (Ext = '.par') or (Ext = '.var') then begin
    if Merge or CloseQuery then
      if Ext = '.pgo' then
        LoadFile(AFileName,Merge)
      else
        ImportFile({$IFDEF FPC}@{$ENDIF}ImportCALPGMFile, AFileName, Merge);
  end else if (Ext = '.lin') or (Ext = '.obs') or (Ext = '.cat') then
    LogForm.ObsFileName.Text := AFileName
  else begin
    AddOverlay(AFileName, Merge);
  end;
end;

procedure TMainForm.CopyItemClick(Sender: TObject);
begin
  G.CopyToClipboard;
end;

procedure TMainForm.CopyPictureItemClick(Sender: TObject);
begin
  LastGraph.CopyToClipboard;
end;

constructor TMainForm.Create(AOwner: TComponent);
begin
  inherited;
  {$IFDEF VCL_}
  StatusBar.AutoHint := True; { TODO: LCL equivalent }
  {$ENDIF}
  FLastGraph := G;
end;

procedure TMainForm.StatesItemClick(Sender: TObject);
begin
  ReShow(StatesForm);
  StatesForm.BaseNode := Mixture;
end;

procedure TMainForm.SetMixture(AMixture: TMIxture);
begin
  if AMixture <> FMixture then begin
    FreeAndNil(FMixture);
    FMixture := AMixture;
    PerformMNodeNotification(Mixture, mnNewBaseNode, 0);
  end;
end;

procedure TMainForm.TransitionsItemClick(Sender: TObject);
begin
  ReShow(TransitionForm);
  TransitionForm.BaseNode := Mixture;
end;

procedure TMainForm.ClearCacheItemClick(Sender: TObject);
begin
  Mixture.ClearCache;
  LineList := nil;
end;

procedure TMainForm.SetLineList(const Value: TLineList);
begin
  if Assigned(Mixture) then
    Mixture.LineList := Value;
end;

function TMainForm.GetLineList: TLineList;
begin
  if Assigned(Mixture) then
    Result := Mixture.LineList
  else
    Result := nil;
end;

procedure TMainForm.GMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FLastGraph := Sender as TGraphD;
  if Sender = G then
    OC.AreaSelected := False
  else
    G.AreaSelected := False;
end;

function TMainForm.LastGraph: TGraphD;
begin
  Result := FLastGraph;
  if Assigned(Result) and Result.Visible then
    Exit;
  if Result = G then
    Result := OC
  else
    Result := G;
  FLastGraph := Result;
end;

procedure TMainForm.LevelsItemClick(Sender: TObject);
begin
  ReShow(LevelForm);
  LevelForm.BaseNode := Mixture;
end;

procedure TMainForm.ShowSimulation;
var
  SpectrumPlot: TSpectrumPlot;
  XLabel, YLabel: string;
  AnySpectrum: Boolean;
  xl, xu: Double;
  function RV(Edit: TCustomEdit; Value: Double): Double;
  begin
    if Edit.Modified and (Edit.Text <> '') then
      Result := StrToFloat(Edit.Text)
    else begin
      Result := Value;
      Edit.Text := FloatToStr(Value);
      Edit.Modified := False;
    end;
  end;
  procedure ShowObservedLines;
  var
    i, nPlotted: Integer;
    v, vcalc: Double;
  begin
    nPlotted := 0;
    for i := 0 to High(Observations) do begin
      v := Observations[i].Frequency;
      try
        if not CalculateObs(Observations[i],vcalc) then
          vcalc := v;
      except
        vcalc := v;
      end;
      v := Convert(v,Mixture.Units,Mixture.PlotUnits);
      vcalc := Convert(vcalc,Mixture.Units,Mixture.PlotUnits);
      if (v >= Mixture.Fmin) and (v <= Mixture.Fmax) then begin
          G.StyleP(Observations[i].Colour, G.LineStyle.Width, G.LineStyle.Pattern);
          G.Point(v, 1.05);
          G.Join(vcalc, 1.02);
          Inc(nPlotted);
        end;
    end;
    if nPlotted > 0 then
      AddPlotRange(SpectrumPlotRanges, G, 1.02, 1.05, 1.02, 1.05, -2);
  end;
begin
  if BeginCalculation then Exit;
  try
    SetLength(FortratPlotRanges,0);
    SetLength(SpectrumPlotRanges,0);
    FLastLineInfo := '';
    if not Assigned(Mixture) then Exit;
    Mixture.Foffset := RV(Foffset, Mixture.Foffset);
    Mixture.Temperature := RV(Temperature, Mixture.Temperature);
    Mixture.Gaussian := RV(Gaussian, Mixture.Gaussian);
    Mixture.Lorentzian := RV(Lorentzian, Mixture.Lorentzian);
    AnySpectrum := Mixture.ShowSum or Mixture.ShowParts;
    if not (AnySpectrum or Mixture.ShowFortrat) then Exit;
    xl := Mixture.Fmin-Mixture.Foffset;
    xu := Mixture.Fmax-Mixture.Foffset;
    if xl = xu then begin
      AllButtonClick(nil);
      Exit;
    end;

    if AnySpectrum and Mixture.ShowFortrat then begin
      Split.MakeBothVisible;
      Split.Visible := True;
    end else begin
      if AnySpectrum then
        Split.OneWindowVisible(G)
      else
        Split.OneWindowVisible(OC);
      Split.Visible := False;
    end;

    if Mixture.ShowParts then
      SpectrumPlot := TMultiColouredSpectrumPlot.Create(Mixture)
    else if AnySpectrum then
      SpectrumPlot := TSpectrumPlot.Create(Mixture.PlotUnits)
    else
      SpectrumPlot := nil;
    if Mixture.Foffset = 0 then
      XLabel := TypeOfUnit(Mixture.PlotUnits) + '/' + NameOfUnit(Mixture.PlotUnits)
    else
      XLabel := Format('(%s-%g)/%s', [Mixture.UnitsTypeName, Mixture.Foffset, Mixture.UnitsName]);
    if Assigned(SpectrumPlot) then
    try
      if not SpectrumPlot.MakeFromMixture(Mixture) then
        Application.Hint := 'No transition moments set';
      G.ResetZoom;
      G.Start;
      G.Map(xl, xu, -0.05, 1);

      if Assigned(Overlays) then begin
        case GraphStyle of
          gsOffset, gsOffsetAll:     G.Map(xl, xu, -1.1, 1);
          gsInverted, gsInvertedAll: G.Map(xl, xu, -1.1, 1);
        end;
        G.Style(1);
        Overlays.Show(G, Mixture.PlotUnits, SpectrumPlotRanges, GraphStyle in [gsOffsetAll, gsInvertedAll]);
        OvrYl := G.Ymin;
        OvrYu := G.Ymax;
      end;
      ShowObservedLines;
      if Assigned(Overlays) then
        case GraphStyle of
          gsOffset, gsOffsetAll:     G.Map(xl, xu, -0.1, 2);
          gsInverted, gsInvertedAll: G.Map(xl, xu, 1.1, -1);
        end
      else
        case GraphStyle of
          gsInverted, gsInvertedAll: G.Map(xl, xu, 1.05, 0);
        end;
      G.Style(0);
      G.Axis(xl, xu, G.Ymin, true, XLabel);
      YLabel := NameOfUnit(Mixture.IntensityUnits);
      if Mixture.IntensityUnits >= nm2MHzperMolecule then
        YLabel := 'Intensity (' + YLabel + ' / ' + NameOfUnit(Mixture.PlotUnits) + ')';
      SpectrumPlot.Show(G, SpectrumPlotRanges, GraphStyle in [gsOffsetAll, gsInvertedAll], YLabel);
      FLastLineInfo := Format('%d lines in range with total intensity %s',
                              [SpectrumPlot.nLinesInRange, DtosM(SpectrumPlot.IntensityInRange,12,-1)]);
      StatusBar.Panels[5].Text := FLastLineInfo;
      G.Finish;
    finally
      ResetProgress;
      SpectrumPlot.Free;
    end;
    if Mixture.ShowFortrat then begin
      OC.ResetZoom;
      OC.Start;
      with TFortratPlot.Create(Mixture.PlotUnits) do
        try
          MakeFromMixture(Mixture);
          OC.Map(xl, xu, IMin-1, IMax+1);
          OC.Axis(xl, xu, IMin-1, true, XLabel);
          Show(OC, FortratPlotRanges, GraphStyle in [gsOffsetAll, gsInvertedAll], 'J');
        finally
          Free;
        end;
      OC.Finish;
    end;
    Mixture.HasPlotted := True;
  finally
    EndCalculation;
  end;
end;

function TMainForm.ShowYValue(GD: TgraphD; const PlotRanges: TPlotRanges):string;
var
  Index: Integer;
  Value: Double;
  procedure ShowObservationHint;
  var
    i: Integer;
    s: string;
  begin
    s := '';
    for i := 0 to High(Observations) do
      if GD.CloseToX(GD.LastMouseX, Observations[i].Frequency) then
        s := s + Format(' %s:%d', [ExtractFileName(Observations[i].FileName), Observations[i].LineNumber]);
    G.Hint := s;
  end;
begin
  { Find plot relating to given YValue and display it }
  Value := GD.LastMouseY;
  Index := FindYRange(Value, GD, PlotRanges);
  if Index >= 0 then
    with PlotRanges[Index] do
      if Tag = -2 then begin
        ShowObservationHint;
        Result := '';
        Exit;
      end else
        Result := RoundedStr(Value, (yx-yi)/((y2-y1)*GD.Height))
  else
    Result := RoundedStr(GD.LastMouseY, (GD.Ymax-GD.Ymin)/GD.Height);
  G.Hint := '';
end;

procedure TMainForm.PlotStyleButtonClick(Sender: TObject);
begin
  if GraphStyle = High(GraphStyle) then
    GraphStyle := Low(GraphStyle)
  else
    Inc(GraphStyle);
  ShowSimulation;
end;

procedure TMainForm.ImportFile(DefaultProc: TMixtureReader);
begin
  if CloseQuery and ImportDialog.Execute then
    ImportFile(DefaultProc, ImportDialog.FileName, MergeItem.Checked);
end;

procedure TMainForm.ImportFile(DefaultProc: TMixtureReader; AFileName: string; Merge: Boolean);
var
  Stream: TFreeFormatStream;
  FStream: TFileStream;
  Line: string;
  NewMixture: TMixture;
  MemoryStream: TMemoryStream;
begin
  FStream := TFileStream.Create(AFileName, fmOpenRead);
  Stream := nil;
  try
    Stream := TFreeFormatStream.Create(FStream);
    Line := LowerCase(Stream.RestOfLine);
    Stream.Position := 0;
    if Pos('this is a d3h ', Line) > 0 then
      DefaultProc := {$IFDEF FPC}@{$ENDIF}ImportD3hFile
    else if Pos(' pgopher ', Line) > 0 then
      DefaultProc := {$IFDEF FPC}@{$ENDIF}ImportPgopherFile
    else if Pos('idn', Line) > 0 then
      DefaultProc := {$IFDEF FPC}@{$ENDIF}ImportACDDFile;
    NewMixture := DefaultProc(Stream);
    if Merge then
      try
        MemoryStream := TMemoryStream.Create;
        try
          BeginUpdate;
          NewMixture.SaveToStream(MemoryStream);
          Mixture.MergeFromStream(MemoryStream);
        finally
          EndUpdate(True);
          MemoryStream.Free;
        end;
      except
        NewMixture.Free;
      end
    else begin
      SetMixture(NewMixture);
      SetName('');
      Mixture.Comment := 'Imported from ' + AFileName;
    end;
  finally
    Stream.Free;
    FStream.Free;
  end;
end;

procedure TMainForm.StylesItemClick(Sender: TObject);
begin
  GraphPenDialog.Graph := LastGraph;
  GraphPenDialog.Execute;
end;

procedure TMainForm.EmptyMixtureItemClick(Sender: TObject);
begin
  if CloseQuery then begin
    SetMixture(TMixture.Create(nil));
    SetName('');
    ClearOverlays(True);
  end;
end;

procedure TMainForm.NewLinearMoleculeItemClick(Sender: TObject);
var
  Mix: TMixture;
begin
  if CloseQuery then begin
    Mix := TMixture.Create(nil);
    AddSampleLinearMolecule(Mix);
    SetMixture(Mix);
    SetName('');
    DoPlot;
    AllButtonClick(nil);
  end;
end;

procedure TMainForm.OldPgopherItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.dat';
  ImportDialog.FilterIndex := 1;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportPgopherFile);
end;

procedure TMainForm.OldD3hItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.dat';
  ImportDialog.FilterIndex := 1;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportD3hFile);
end;

procedure TMainForm.CALPGMfileItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.par';
  ImportDialog.FilterIndex := 2;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportCALPGMFile);
end;

procedure TMainForm.BinaryfileItemClick(Sender: TObject);
var
  ExperimenalPlot: TExperimentalPlot;
  Stream: TStream;
  BinaryFileLoader: TBinaryFileLoader;
begin
  if not MergeItem.Checked then begin
    if not ClearOverlays(False) then
      Exit;
    Overlays := TOverlays.Create(nil);
    PerformMNodeNotification(Overlays, mnNewBaseNode, 0);
  end;
  ImportDialog.DefaultExt := '.dat';
  ImportDialog.FilterIndex := 3;
  if ImportDialog.Execute then begin
    Stream := TFileStream.Create(ImportDialog.FileName, fmOpenRead);
    try
      ExperimenalPlot := TExperimentalPlot.Create(Overlays);
      try
        ExperimenalPlot.Name := ExtractFileName(ImportDialog.FileName);
        ExperimenalPlot.Comment := 'Imported from ' + ImportDialog.FileName;
        BeginUpdate;
        BinaryFileLoader := TBinaryFileLoader.Create(ExperimenalPlot,Stream,'');
        try
          if not BinaryFileLoader.Load(False) then
            FreeAndNil(ExperimenalPlot);
        finally
          BinaryFileLoader.Free;
          EndUpdate(True);
        end;
      except
        ExperimenalPlot.Free;
      end;
    finally
      Stream.Free;
    end;
  end;
end;

procedure TMainForm.MRUFileListGetDisplayName(Sender: TObject;
  AFilename: string; var ADisplayName: string);
begin
  ADisplayName := ExtractRelativepath(GetCurrentDir, AFilename);
end;

procedure TMainForm.ExportPlotItemClick(Sender: TObject);
var
  StreamClass: TStreamCanvasClass;
  Stream: TStream;
begin
  if ExportPlotDialog.Execute then begin
    StreamClass := GetStreamCanvasClassFromFilename(ExportPlotDialog.FileName);
    Stream := TFileStream.Create(ExportPlotDialog.FileName,fmCreate);
    try
      LastGraph.ExportTo(Stream, StreamClass);
    finally
      Stream.Free;
    end;
  end;
end;

procedure TMainForm.WMDropFiles(var Message: TMessage);
var
  Strings: TStringList;
  i: Integer;
begin
  Strings := DroppedFiles(Message);
  try
    for i := 0 to Strings.Count-1 do
      LoadGeneralFile(Strings[i],MergeItem.Checked);
  finally
    Strings.Free;
  end;
  ConsiderReplot(not Assigned(LineList));
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  if HandleAllocated then
    DragAcceptFiles(Handle, False);
end;

procedure TMainForm.TimingItemClick(Sender: TObject);
begin
  ReShow(ProfilerForm);
  {$IFDEF RecyclerMM}
  RMMSnapShot.Show;
  {$ENDIF}
end;

procedure TMainForm.InvertItemClick(Sender: TObject);
begin
  if Assigned(Overlays) then
    Overlays.Invert;
  ConsiderReplot(False,False);
end;

procedure TMainForm.DeleteLastPeakItemClick(Sender: TObject);
var
  Plot: TPlot;
begin
  Plot := LastPlot(True);
  if Assigned(Plot) then begin
    Plot.DeleteLastPeak;
    ConsiderReplot(False,False);
  end;
end;

function TMainForm.LastPlot(WithPeak: Boolean): TPlot;
var
  i: Integer;
begin
  Result := nil;
  if not Assigned(Overlays) then Exit;
  for i := 0 to Overlays.Count - 1 do
    if Overlays[i] = FLastPlot then
      Break
    else if i = Overlays.Count - 1 then
      FLastPlot := nil;
  if Assigned(FLastPlot) and (not WithPeak or ( Length(FLastPlot.Peaks) > 0 )) then begin
    Result := FLastPlot;
    Exit;
  end;
  for i := Overlays.Count - 1 downto 0 do
    if not WithPeak or ( Length(Overlays[i].Peaks) > 0 ) then begin
      Result := Overlays[i];
      Exit;
    end;
end;

procedure TMainForm.ConsiderReplot(All: Boolean; PurgeLineSelection: Boolean);
begin
  if not Assigned(Mixture) or Mixture.AutoReplot then
    DoPlot(All,PurgeLineSelection);
end;

procedure TMainForm.DoPlot(All: Boolean; PurgeLineSelection: Boolean);
begin
  if PurgeLineSelection and Assigned(LineList) and LineList.Selected then
    LineList := nil;
  if All then
    FindLimits;
  ShowSimulation;
end;

procedure TMainForm.ClearAllPeaksItemClick(Sender: TObject);
var
  i: Integer;
begin
  if not Assigned(Overlays) then Exit;
  for i := 0 to Overlays.Count-1 do
    Overlays[i].ClearPeaks;
  ConsiderReplot(False,False);
end;

procedure TMainForm.LoadAnyObservations;
begin
  if LastObsWasLineForm and LineForm.HasObservations then
    LineForm.LoadObservations
  else
    LogForm.LoadObservations;
end;

procedure TMainForm.DoLineFit;
var
  i: Integer;
begin
  if BeginCalculation then Exit;
  try
    for i := 1 to Mixture.FitCycles do begin
      Mixture.FillParameters;
      Mixture.PushParameters(ftLine);
      PushProgress('Fitting');
      try
        FitLines(Mixture.FittingParameters, Observations, Mixture.Precision);
        PopProgress;
      except
        Mixture.DiscardFit;
        ResetProgress;
        raise;
      end;
    end;
  finally
    EndCalculation;
  end;
  ConsiderReplot;
end;

function TMainForm.BeginCalculation: Boolean;
begin
  Result := CalculationInProgress;
  FCalculationInProgress := True;
end;

procedure TMainForm.EndCalculation;
begin
  FCalculationInProgress := False;
end;

procedure TMainForm.SelectRegionItemClick(Sender: TObject);
var
  i: Integer;
begin
  if not Assigned(Overlays) or not LastGraph.AreaSelected then Exit;
  for i := 0 to Overlays.Count-1 do
    Overlays[i].SelectFrequencyRange(LastGraph.SelectedMathArea.Left, LastGraph.SelectedMathArea.Right);
  ConsiderReplot(False,False);
end;

procedure TMainForm.ClearSelectionItemClick(Sender: TObject);
var
  i: Integer;
begin
  if not Assigned(Overlays) then Exit;
  for i := 0 to Overlays.Count-1 do
    Overlays[i].ClearSelection;
  ConsiderReplot(False,False);
end;

procedure TMainForm.WhatIsItemClick(Sender: TObject);
begin
  SetInHelp(True);
end;

procedure TMainForm.IndexItemClick(Sender: TObject);
begin
  {$IFDEF VCL_}
  Application.HelpSystem.ShowTableOfContents;
  {$ENDIF}
end;

procedure TMainForm.SearchItemClick(Sender: TObject);
begin
  {$IFDEF VCL_}
  PHtmlHelpViewer.ShowHtmlSearch;
  {$ENDIF}
end;

procedure TMainForm.ProceduresItemClick(Sender: TObject);
begin
  Application.HelpKeyword('Procedures');
end;

procedure TMainForm.UsingHelpItemClick(Sender: TObject);
begin
  Application.HelpKeyword('Using Help');
end;

procedure TMainForm.ShowPartsItemClick(Sender: TObject);
begin
  ShowPartsButton.Down := not ShowPartsButton.Down;
  ChangeMixtureSetting(ShowPartsIndex, ShowPartsButton);
end;

procedure TMainForm.ShowSumItemClick(Sender: TObject);
begin
  ShowSumButton.Down := not ShowSumButton.Down;
  ChangeMixtureSetting(ShowSumIndex, ShowSumButton);
end;

procedure TMainForm.ShowFortratItemClick(Sender: TObject);
begin
  ShowFortratButton.Down := not ShowFortratButton.Down;
  ChangeMixtureSetting(ShowFortratIndex, ShowFortratButton);
end;

procedure TMainForm.DumpItemClick(Sender: TObject);
begin
  {$IFDEF HELPHELP} MakeSkeletonHelp(Screen.ActiveForm); {$ENDIF}
end;

procedure TMainForm.GContextPopup(Sender: TObject; MousePos: TPoint;
  var Handled: Boolean);
begin
  Handled := True; { Don't want default context menu }
end;

procedure TMainForm.StatusBarHint(Sender: TObject);
begin
  if Application.Hint = '' then
    StatusBar.Panels[5].Text := FLastLineInfo
  else
    StatusBar.Panels[5].Text := Application.Hint;
end;

procedure TMainForm.ChangeMixtureSetting(AIndex: Integer; Sender: TObject);
begin
  if Assigned(Mixture) then
    Mixture.SetFlag(AIndex, (Sender as TSpeedButton).Down);
  ShowSimulation;
end;

procedure TMainForm.ExportLineListItemClick(Sender: TObject);
var
  F: TextFile;
  i: Integer;
  Sep: Char;
  Threshold: Double;
begin
  if not Assigned(Mixture) then
    Exit;
  if not Assigned(Mixture.LineList) then
    Mixture.MakeLineList;
  if not Assigned(Mixture.LineList) then
    Exit;
  if Mixture.Name <> '' then
    ExportLineListDialog.FileName := ChangeFileExt(Mixture.Name, '');
  if ExportLineListDialog.Execute and (LineExportOptions.ShowModal = mrOK) then begin
    AssignFile(F, ExportLineListDialog.FileName);
    if LowerCase(ExtractFileExt(ExportLineListDialog.FileName)) = '.csv' then
      Sep := ','
    else if LowerCase(ExtractFileExt(ExportLineListDialog.FileName)) = '.txt' then
      Sep := #9
    else
      Sep := ' ';
    try
      ReWrite(F);
      with Mixture.LineList do begin
        Threshold := LargestIntensity*LineExportOptions.ThresholdSpin.Value;
        HeadingPrinted := False;
        for i := 0 to nLines-1 do
          ShowLine(i,F,Sep,Threshold,LineExportOptions.FitFormatCheck.Checked);
        ShowFooter(F,Sep);
      end;
    finally
      CloseFile(F);
    end;
  end;
end;

procedure TMainForm.WavenumbersItemClick(Sender: TObject);
begin
  if not Assigned(Mixture) then
    Exit;
  Mixture.ChangePlotUnits(TResonanceUnits((Sender as TComponent).Tag));
  UpdateToolbar;
  ConsiderReplot;
end;

procedure TMainForm.FoffsetKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then begin
    Key := 0;
    ShowSimulation;
  end;
end;

procedure TMainForm.UpperStateItemClick(Sender: TObject);
begin
  Mixture.UseUpper := not Mixture.UseUpper;
  ConsiderReplot;
end;

initialization
{$IFDEF FPC}
{$IFDEF CPUI386}
  SetExceptionMask(GetExceptionMask+[exInvalidOp]);
{$ENDIF}
{$IFDEF DARWIN}
//  asm
//    mtfsfi 6,9
//  end;
{$ENDIF}
{$ENDIF}
{$IFDEF LCL_}
  {$I Main.lrs}
{$ENDIF}

end.
