unit OvrlyFrm;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms, Messages,
  Dialogs, ConstFrm, Menus, ComCtrls, Grids, NGrid, StdCtrls, Buttons,
  ExtCtrls, Split, Overlay, GridEx, MNode, ExtDlgs, ScrlCtrl, Graphs;

type
  TOverlayForm = class(TConstantsForm)
    OpenButton: TSpeedButton;
    SaveButton: TSpeedButton;
    SaveAsButton: TSpeedButton;
    SaveDialog: TSaveDialog;
    OpenDialog: TOpenDialog;
    InvertItem: TMenuItem;
    DetailsPanel: TPanel;
    SplitV: TSplit;
    BitmapPanel: TPanel;
    BitmapToolbar: TPanel;
    CompactButton: TButton;
    CancelButton: TButton;
    OneToOneButton: TSpeedButton;
    Label2: TLabel;
    Xpixel1Button: TButton;
    Xpixel2Button: TButton;
    BaselineButton: TButton;
    MaximumButton: TButton;
    ScrollBox: TScrollBox;
    PasteOverlayButton: TSpeedButton;
    CopyButton: TSpeedButton;
    ExportCurrentButton: TSpeedButton;
    AddOpenButton: TSpeedButton;
    OpenPictureDialog: TOpenPictureDialog;
    SavePictureDialog: TSavePictureDialog;
    PlotPanel: TPanel;
    PlotToolbar: TPanel;
    ExportCurrentButton2: TSpeedButton;
    AddOpenButton2: TSpeedButton;
    GraphD: TGraphD;
    ClearPlotPeaksButton: TButton;
    ClearBitmapPeaksButton: TButton;
    PaintBox: TPaintBox;
    CropButton: TButton;
    UndoButton: TButton;
    PlateButton: TButton;
    CopyPeaksButton: TButton;
    procedure CopyPeaksButtonClick(Sender: TObject);
    procedure PlateButtonClick(Sender: TObject);
    procedure CropButtonClick(Sender: TObject);
    procedure UndoButtonClick(Sender: TObject);
    procedure PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBoxPaint(Sender: TObject);
    procedure ClearPlotPeaksButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure GraphDMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ExportCurrentButtonClick(Sender: TObject);
    procedure AddOpenButtonClick(Sender: TObject);
    procedure PasteOverlayButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MaximumButtonClick(Sender: TObject);
    procedure OptGridDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
      State: TGridDrawState);
    procedure CompactButtonClick(Sender: TObject);
    procedure OneToOneButtonClick(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure SaveAsButtonClick(Sender: TObject);
    procedure OpenButtonClick(Sender: TObject);
    procedure InvertItemClick(Sender: TObject);
  private
    FMouseSetting: Integer;
    Selecting: Boolean; { Mouse is selecting an area }
    FAreaSelected: Boolean;
    FSelectedArea:TRect;
    function Overlays: TOverlays;
    procedure SetMouseSetting(const Value: Integer);
    procedure ScaleCoords(var X, Y: Integer);
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
    procedure CancelSelection;
    procedure DrawSelectedArea;
    procedure UpdateSelection(X, Y: Integer; NewPoint:Boolean);
    procedure ShowCurrentPlot;
  protected
    procedure MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer); override;
  public
    procedure AddOverlayFile(FileName: string);
    function BaseNodeClass: TClass; override;
    destructor Destroy; override;
    function DefaultCaption: string; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    property MouseSetting: Integer read FMouseSetting write SetMouseSetting;
    procedure UpdateAlwaysVisible; override;
  end;

var
  OverlayForm: TOverlayForm;

implementation

uses {$IFDEF LCL_} LResources, LCLType, {$ENDIF}
     {$IFNDEF CONSOLE} ClipBrd, {$IFDEF MSWINDOWS} ShellAPI, {$ENDIF} {$ENDIF}
     Huge, Mixture, MathSup, Main, ExpBitmap, ExpPlot, ChildForm, GraphRange, Plate, WCompat;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

function TOverlayForm.Overlays: TOverlays;
begin
  Result := BaseNode as TOverlays;
end;

destructor TOverlayForm.Destroy;
begin
  if Self = OverlayForm then
    OverlayForm := nil;
  inherited Destroy;
end;

procedure TOverlayForm.SaveButtonClick(Sender: TObject);
begin
  if Overlays <> nil then
    with Overlays do begin
      if (Name = '') or (Name = 'Overlays') then
        SaveAsButtonClick(Sender)
      else
        SaveToFile(Name);
    end;
end;

procedure TOverlayForm.SaveAsButtonClick(Sender: TObject);
begin
  if (Overlays <> nil) and SaveDialog.Execute then
    Overlays.SaveToFile(SaveDialog.FileName);
  ReStuff;
end;

procedure TOverlayForm.OpenButtonClick(Sender: TObject);
begin
  if OpenDialog.Execute then
    MainForm.LoadGeneralFile(OpenDialog.FileName,False);
end;

function TOverlayForm.DefaultCaption: string;
begin
  Result := 'Overlays';
end;

function TOverlayForm.BaseNodeClass: TClass;
begin
  Result := TOverlays;
end;

procedure TOverlayForm.InvertItemClick(Sender: TObject);
begin
  if Assigned(CurrentMNode) and (CurrentMNode is TPlot) then
    TPlot(CurrentMNode).Invert;
  ShowItem(CurrentMNode);
end;

procedure TOverlayForm.OneToOneButtonClick(Sender: TObject);
begin
  if OneToOneButton.Down then begin
    PaintBox.Align := alNone;
    PaintBox.Left := 0;
    PaintBox.Top := 0;
    if Assigned(CurrentMNode) and (CurrentMNode is TBitmapPlot) then
      with CurrentMNode as TBitmapPlot do
        PaintBox.SetBounds(0, 0, Picture.Width, Picture.Height);
  end else
    PaintBox.Align := alClient;
  PaintBox.Invalidate;
end;

procedure TOverlayForm.CompactButtonClick(Sender: TObject);
begin
  if Assigned(CurrentMNode) then
    with CurrentMNode as TBitmapPlot do
      Compact;
  ShowItem(CurrentMNode);
end;

procedure TOverlayForm.ShowCurrentPlot;
var
  PlotRanges: TPlotRanges;
  d: Double;
begin
  GraphD.Start;
  if CurrentMNode is TOverlays then begin
    Overlays.FixupProperties;
    GraphD.Map(Overlays.HMin, Overlays.HMax, 0, 1);
    GraphD.Scales(Overlays.UnitsName, '');
    Overlays.Show(GraphD, Overlays.Units, PlotRanges, True);
  end else
    with CurrentMNode as TExperimentalPlot do begin
      if IsLineList then begin
        d := (HMax - HMin)/50;
        GraphD.Map(HMin-d, HMax+d, 0, 1);
      end else
        GraphD.Map(HMin, HMax, 0, 1);
      Show(GraphD, 0, 1, Units, PlotRanges, False);
      if Length(PlotRanges) = 1 then begin
        GraphD.ScaledAxis(PlotRanges[0].yi, PlotRanges[0].yx, 0, 1, GraphD.Xmin, False, '');
        GraphD.Axis(GraphD.Xmin, GraphD.Xmax, GraphD.Ymin, true, UnitsName);
      end;
    end;
  GraphD.Finish;
end;

procedure TOverlayForm.ShowItem(MNode: TMNode);
var
  PlotRanges: TPlotRanges;
begin
  BeginUpdate;
  inherited;
  MouseSetting := -1;
  if Assigned(MNode) and (MNode is TBitmapPlot) then
    with MNode as TBitmapPlot do begin
      BitmapPanel.Visible := True;
      BitmapPanel.Align := alClient;
      PlotPanel.Visible := False;
      CancelSelection;
      if Assigned(Picture.Graphic) then begin
       { with Image.Picture.Bitmap do begin
          Height := Picture.Graphic.Height;
          Width := Picture.Graphic.Width;
          Canvas.Brush.Color := clWhite;
          Canvas.FillRect(Rect(0,0,Width, Height));
          Canvas.Draw(0,0,Picture.Graphic);
        end }
        CompactButton.Enabled := not ((Picture.Graphic is TBitmap)
                                        and (TBitmap(Picture.Graphic).PixelFormat = pf1bit))
      end else
        CompactButton.Enabled := False;
      OneToOneButtonClick(nil);
    end
  else begin
    PlotPanel.Visible := True;
    PlotPanel.Align := alClient;
    BitmapPanel.Visible := False;
    if Assigned(CurrentMNode) then
        ShowCurrentPlot
    else
      GraphD.PlotChanged;
  end;
  EndUpdate(False);
end;

procedure TOverlayForm.SetMouseSetting(const Value: Integer);
begin
  FMouseSetting := Value;
  OptGrid.Invalidate;
  CancelButton.Enabled := Value > 0;
end;

procedure TOverlayForm.OptGridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
begin
  if (MouseSetting > 1) and (MouseSetting = ARow+1) and (ACol = 1) then
    ColouredTextInGrid(OptGrid.Canvas, Rect, 'Click to set', clRed);
end;

procedure TOverlayForm.MaximumButtonClick(Sender: TObject);
begin
  MouseSetting := (Sender as TComponent).Tag;
end;

procedure TOverlayForm.PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if Selecting then
    UpdateSelection(X, Y, True);
  ScaleCoords(X,Y);
  StatusBar.Panels[0].Text := 'Pixels';
  StatusBar.Panels[2].Text := IntToStr(X); { TODO : Scale to true pixels }
  StatusBar.Panels[4].Text := IntToStr(Y);
  if CurrentMNode is TBitmapPlot then
    with TBitmapPlot(CurrentMNode) do
      if Picture.Width > 0 then begin
        StatusBar.Panels[6].Text := RoundedStr(HMin + X *(HMax-HMin)/nDF, (HMax-HMin)/nDF);
        StatusBar.Panels[8].Text := RoundedStr(IMin + Y *(IMax-IMin)/Picture.Height, (IMax-IMin)/Picture.Height);
      end
  else begin
    StatusBar.Panels[6].Text := '';
    StatusBar.Panels[8].Text := '';
  end;
end;

procedure TOverlayForm.ScaleCoords(var X, Y: Integer);
var
  Rect: TRect;
begin
  { Convert from client Co-ordinates to bitmap coordinates }
  if not (CurrentMNode is TBitmapPlot) then
    Exit;
  Rect := PaintBox.ClientRect;
  if Rect.Right <> Rect.Left then
    X := Round((X-Rect.Left)/(Rect.Right-Rect.Left)*TBitmapPlot(CurrentMNode).Picture.Width);
  if Rect.Bottom <> Rect.Top then
    Y := Round((Y-Rect.Bottom)/(Rect.Top-Rect.Bottom)*TBitmapPlot(CurrentMNode).Picture.Height);
end;

procedure TOverlayForm.PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  v: Integer;
begin
  if (MouseSetting > 1) and Assigned(CurrentMNode) and (CurrentMNode is TBitmapPlot) then
    with CurrentMNode as TBitmapPlot do begin
      ScaleCoords(X,Y);
      if MouseSetting >= 7 then
        v := Y
      else
        v := X;
      SetSetting(MouseSetting,v);
    end;
  MouseSetting := -1;
  if Button = mbLeft then begin
    CancelSelection;
    Selecting := True;
    FSelectedArea.Left := X;
    FSelectedArea.Top := Y;
  end;
end;

procedure TOverlayForm.PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) and Selecting then begin
    UpdateSelection(X, Y, True);
    Selecting := False;
  end;
  UpdateAlwaysVisible;
end;

procedure TOverlayForm.CancelButtonClick(Sender: TObject);
begin
  MouseSetting := -1;
end;

procedure TOverlayForm.CopyButtonClick(Sender: TObject);
var
  AFormat: Word;
  {$IFDEF MSWINDOWS}
  AData: Cardinal;
  APalette: HPALETTE;
  {$ENDIF}
begin
  if Assigned(CurrentMNode) then
    try
      Clipboard.Open;
      if CurrentMNode is TBitmapPlot then begin
        TBitmapPlot(CurrentMNode).Picture.SaveToClipboardFormat(AFormat{$IFDEF VCL_} , AData, APalette {$ENDIF});
        {$IFDEF VCL_}
        if APalette <> 0 then
          Clipboard.SetAsHandle(CF_PALETTE, APalette);
        Clipboard.SetAsHandle(AFormat, AData);
        {$ENDIF}
      end;
      CurrentMNode.CopyToClipboard;
    finally
      Clipboard.Close;
    end;
end;

procedure TOverlayForm.AddOverlayFile(FileName: string);
begin
  MainForm.AddOverlay(FileName,True);
  if Assigned(MainForm.Overlays) then begin
    BaseNode := MainForm.Overlays;
    if SelectorComboBox.Items.Count > 0 then
      SelectorComboBox.ItemIndex := SelectorComboBox.Items.Count-1;
  end;
end;

procedure TOverlayForm.PasteOverlayButtonClick(Sender: TObject);
begin
  if not PasteItemsIntoTree then
    AddOverlayFile('');
end;

procedure TOverlayForm.AddOpenButtonClick(Sender: TObject);
begin
  OpenPictureDialog.Filter :=
    'Data File (*.dat)|*.dat|Overlay File (*.ovr)|*.ovr|'
    +PlotLoaderFilter
    +GraphicFilter(TGraphic)
    +'|All Files (*.*)|*.*';
  { TODO: hook onselection change to preview other styles of file }
  if OpenPictureDialog.Execute then
    AddOverlayFile(OpenPictureDialog.FileName);
end;

procedure TOverlayForm.ExportCurrentButtonClick(Sender: TObject);
begin
  if Assigned(CurrentMNode) then
    if CurrentMNode is TBitmapPlot then
      with CurrentMNode as TBitmapPlot do begin
        SavePictureDialog.DefaultExt := GraphicExtension(TGraphicClass(Picture.Graphic.ClassType));
        SavePictureDialog.Filter := GraphicFilter(TGraphicClass(Picture.Graphic.ClassType));
        if SavePictureDialog.Execute then
          Picture.Graphic.SaveToFile(SavePictureDialog.FileName);
      end;
    { TODO: Handle Experimental Plot }
    { TODO: Allow encapsulation in .ovr }
end;

procedure TOverlayForm.GraphDMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  i: Integer;
  MNode: TMNode;
begin
  MNode := CurrentMNode;
  if (CurrentMNode is TOverlays) and (CurrentMNode.Count = 1) then
    MNode := CurrentMNode.Items[0];
  if MNode is TExperimentalPlot then
    with MNode as TExperimentalPlot do begin
      IntensityAt(GraphD.LastMouseX, i);
      StatusBar.Panels[0].Text := 'Point';
      StatusBar.Panels[2].Text := IntToStr(i);
      StatusBar.Panels[4].Text := RoundedStr(Intensity[i], (Imax-Imin)/Height);
      { TODO: Show weights also }
    end
  else begin
    StatusBar.Panels[0].Text := '';
    StatusBar.Panels[2].Text := '';
    StatusBar.Panels[4].Text := '';
  end;
  StatusBar.Panels[6].Text := RoundedStr(GraphD.LastMouseX, (GraphD.Xmax-GraphD.Xmin)/Width);
  StatusBar.Panels[8].Text := RoundedStr(GraphD.LastMouseY, (GraphD.Ymax-GraphD.Ymin)/Height);
end;

procedure TOverlayForm.WMDropFiles(var Message: TMessage);
var
  Strings: TStringList;
  i: Integer;
begin
  Strings := DroppedFiles(Message);
  try
    for i := 0 to Strings.Count-1 do
      AddOverlayFile(Strings[i]);
  finally
    Strings.Free;
  end;
end;

procedure TOverlayForm.FormDestroy(Sender: TObject);
begin
  if HandleAllocated then
    DragAcceptFiles(Handle, False);
  inherited;
end;

procedure TOverlayForm.FormCreate(Sender: TObject);
begin
  inherited;
  DragAcceptFiles(Handle, True);
end;

procedure TOverlayForm.UpdateAlwaysVisible;
var
  i: Integer;
  SomePeaks: Boolean;
begin
  inherited;
  if Visible and Assigned(BaseNode) then begin
    StatusBar.Panels[5].Text := (BaseNode as TStreamableMNode).UnitsTypeName;
    SomePeaks := False;
    if Assigned(CurrentMNode) then begin
      if CurrentMNode is TBitmapPlot then
        UndoButton.Enabled := Assigned(TBitmapPlot(CurrentMNode).LastGraphic);
      if CurrentMNode is TPlot then
        SomePeaks := High(TPlot(CurrentMNode).Peaks) >= 0
      else with CurrentMNode as TOverlays do
        for i := 0 to ComponentCount-1 do
          if High(Items[i].Peaks) >= 0 then
            SomePeaks := True;
    end;
    ClearPlotPeaksButton.Enabled := SomePeaks;
    ClearBitmapPeaksButton.Enabled := SomePeaks;
    CropButton.Enabled := FAreaSelected;
  end;
end;

procedure TOverlayForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  inherited;
  if Assigned(CurrentMNode) and not (CurrentMNode is TBitmapPlot) then
    ShowCurrentPlot;
end;

procedure TOverlayForm.ClearPlotPeaksButtonClick(Sender: TObject);
var
  i: Integer;
begin
  if Assigned(CurrentMNode) then
    if CurrentMNode is TPlot then
      TPlot(CurrentMNode).ClearPeaks
    else with CurrentMNode as TOverlays do
      for i := 0 to ComponentCount-1 do
        Items[i].ClearPeaks;
  ShowItem(CurrentMNode);
end;

procedure TOverlayForm.PaintBoxPaint(Sender: TObject);
begin
  PaintBox.Canvas.FillRect(ClientRect);
  if Assigned(CurrentMNode) and (CurrentMNode is TBitmapPlot) then
    with CurrentMNode as TBitmapPlot do
      PaintBox.Canvas.StretchDraw(PaintBox.ClientRect, Picture.Graphic);
  DrawSelectedArea;
end;

procedure TOverlayForm.CancelSelection;
begin
  DrawSelectedArea;
  Selecting := False;
  FAreaSelected := False;
end;

procedure TOverlayForm.DrawSelectedArea;
begin
  if FAreaSelected then
    DrawFocusRectHelper(PaintBox.Canvas, FSelectedArea);
end;

procedure TOverlayForm.UpdateSelection(X, Y: Integer; NewPoint: Boolean);
begin
  DrawSelectedArea;
  if NewPoint and Selecting then begin
    FSelectedArea.Right := X;
    FSelectedArea.Bottom := Y;
    { Make sure rectangle shown includes current cursor point }
    if FSelectedArea.Right > FSelectedArea.Left then
      Inc(FSelectedArea.Right);
    if FSelectedArea.Bottom > FSelectedArea.Top then
      Inc(FSelectedArea.Bottom);
    FAreaSelected := ( abs(FSelectedArea.Right - FSelectedArea.Left) > SmallX )
                    or (abs(FSelectedArea.Bottom - FSelectedArea.Top) > SmallY );
    DrawSelectedArea;
  end;
end;

procedure TOverlayForm.UndoButtonClick(Sender: TObject);
begin
  if not Assigned(CurrentMNode) or not (CurrentMNode is TBitmapPlot) then
    Exit;
  with TBitmapPlot(CurrentMNode) do
    Undo;
  ShowItem(CurrentMNode);
end;

procedure TOverlayForm.CropButtonClick(Sender: TObject);
begin
  if not Assigned(CurrentMNode) or not (CurrentMNode is TBitmapPlot) or not FAreaSelected then
    Exit;
  ScaleCoords(FSelectedArea.Left, FSelectedArea.Top);
  ScaleCoords(FSelectedArea.Right, FSelectedArea.Bottom);
  with TBitmapPlot(CurrentMNode) do
    Crop(FSelectedArea);
  ShowItem(CurrentMNode);
end;

procedure TOverlayForm.PlateButtonClick(Sender: TObject);
var
  DeltaX: Integer;
  s: string;
begin
  if not Assigned(CurrentMNode) or not (CurrentMNode is TBitmapPlot) then
    Exit;
  if FAreaSelected then begin
    ScaleCoords(FSelectedArea.Left, FSelectedArea.Top);
    ScaleCoords(FSelectedArea.Right, FSelectedArea.Bottom);
    with FSelectedArea do
      if Top > Bottom then
        DeltaX := Right - Left
      else
        DeltaX := Left - Right;
  end else begin
    s := '0';
    if not InputQuery('Y displacement for a line', 'pixels:', s) then
      Exit;
    DeltaX := StrToInt(s);
  end;
  TBitmapPlot(CurrentMNode).ScanPlateSpectrum(DeltaX);
  ShowItem(BaseNode);
end;

procedure TOverlayForm.CopyPeaksButtonClick(Sender: TObject);
var
  Stream: THugeMemoryStream;
  s: string;
  i: Integer;
  procedure Process(Plot: TPlot);
  var
    i: Integer;
  begin
    for i := 0 to High(Plot.Peaks) do begin
      s := FloatToStr(Plot.Peaks[i]) + sLineBreak;
      Stream.Write(s[1], Length(s));
    end;
  end;
begin
  if not Assigned(CurrentMNode) then
    Exit;
  Stream := THugeMemoryStream.Create;
  try
    if CurrentMNode is TOverlays then
      with TOverlays(CurrentMNode) do
        for i := 0 to Count-1 do
          Process(Items[i])
    else
      Process(CurrentMNode as TPlot);
    Stream.CopyToClipboard(CF_TEXT);
  finally
    Stream.Free;
  end;
end;

procedure TOverlayForm.MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
begin
  if Change = mnGraphicChanged then
    GraphD.PlotChanged;
  inherited;
end;

initialization
{$IFDEF LCL_}
  {$I ovrlyfrm.lrs}
{$ENDIF}
end.
