unit PickDest;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, Messages,  {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Buttons, WhatsThis, Dialogs, Grids, GridEx, StdCtrls, MNode;

type

  TPickDestForm = class(THelpForm)
    Grid: TStringGridEx;
    OKButton: TButton;
    CancelButton: TButton;
    Label1: TLabel;
    procedure GridGetEditStyle(Sender: TObject; ACol, ARow: Integer;
      var Style: TEditStyle);
    procedure GridGetPickListItems(ACol, ARow: Integer; Items: TStrings);
    procedure OKButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
  private
    nIntensity, nWeighted: Integer;
    ChannelDest: TIntegerArray;
  public
    function Execute(var AChannelDest: TIntegerArray; var AnWeights: Integer): Integer;
  end;

var
  PickDestForm: TPickDestForm;

implementation

{$IFDEF LCL_}
uses LResources;
{$ENDIF}

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

function TPickDestForm.Execute(var AChannelDest: TIntegerArray; var AnWeights: Integer): Integer;
var
  i: Integer;
begin
  ChannelDest := AChannelDest;
  Grid.RowCount := Length(ChannelDest)+1;
  Grid.Cells[1,1] := 'Frequency';
  Grid.Cells[1,2] := 'Intensity';
  for i := 3 to Length(ChannelDest) do
    Grid.Cells[1,i] := 'Discard';
  for i := 1 to Length(ChannelDest) do
    Grid.Cells[0,i] := IntToStr(i);
  Grid.Cells[0,0] := 'Channel';
  Grid.Cells[1,0] := 'Destination';
  AnWeights := 0;
  if ShowModal = mrOK then begin
    Result := nIntensity;
    AnWeights := nWeighted;
  end else
    Result := 0;
end;

procedure TPickDestForm.GridGetEditStyle(Sender: TObject; ACol, ARow: Integer; var Style: TEditStyle);
begin
  Style := esPickList;
end;

procedure TPickDestForm.GridGetPickListItems(ACol, ARow: Integer;
  Items: TStrings);
begin
  if Items.Count <> 0 then Exit;
  Items.Add('Frequency');
  Items.Add('Intensity');
  Items.Add('Weights');
  Items.Add('Discard');
end;

procedure TPickDestForm.OKButtonClick(Sender: TObject);
var
  HadFrequency: Boolean;
  i: Integer;
begin
  nIntensity := 0;
  HadFrequency := False;
  nWeighted := 1000;
  for i := 0 to High(ChannelDest) do
    if Grid.Cells[1,i+1] = 'Frequency' then begin
      if HadFrequency then
        raise Exception.Create('More than one frequency channel');
      ChannelDest[i] := 0;
      HadFrequency := True;
    end else if Grid.Cells[1,i+1] = 'Intensity' then begin
      Inc(nIntensity);
      ChannelDest[i] := nIntensity;
    end else if Grid.Cells[1,i+1] = 'Weights' then begin
      ChannelDest[i] := nWeighted;
      Inc(nWeighted);
    end else
      ChannelDest[i] := -1;
  if not HadFrequency then
    raise Exception.Create('No frequency channel specified');
  Dec(nWeighted,1000);
  if (nWeighted > 0) and (nWeighted <> nIntensity) then
    raise Exception.Create('Specify same number of weight and intensity channels');
  ModalResult := mrOk;
end;

procedure TPickDestForm.CancelButtonClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

initialization
{$IFDEF LCL_}
  {$I PickDest.lrs}
{$ENDIF}

end.
