unit LineFrm;

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF} { Messages, }
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, NGrid, ExtCtrls, Units,
  Mixture, StdCtrls, Buttons, Menus, QuantumNumber, WhatsThis;

type

  TStateDef = record
    Manifold: string;
    J:        TQuantumNumber;
    Sym:      Integer;
  end;

  TLineW = record
    Molecule:  string;
    Upper:     TStateDef;
    Lower:     TStateDef;
    Weight:    Single;
    Line:      TLine; { Note that eigenvalues not always set }
  end;

  TLineSortFunction = function (const A, B: TLineW): Integer;

  { TLineForm }

  TLineForm = class(THelpForm)
    Panel: TPanel;
    Grid: TNumberGrid;
    ClearButton: TButton;
    AssignButton: TButton;
    AllButton: TButton;
    NoneButton: TButton;
    CopyButton: TSpeedButton;
    CutButton: TSpeedButton;
    DeleteButton: TSpeedButton;
    CloseButton: TButton;
    DeleteUnassigned: TButton;
    MinusButton: TButton;
    Difference: TLabel;
    SortOnButton: TButton;
    SortPopupMenu: TPopupMenu;
    FrequencySortItem: TMenuItem;
    UpperStateSortItem: TMenuItem;
    LowerStateSortItem: TMenuItem;
    StrengthSortItem: TMenuItem;
    WidthSortItem: TMenuItem;
    WeightSortItem: TMenuItem;
    FitButton: TSpeedButton;
    AutoCheck: TCheckBox;
    BranchItem: TMenuItem;
    procedure BranchItemClick(Sender: TObject);
    procedure GridGetCellLabel(Sender: TObject; Col, Row: Integer;
      var s: string; var IsText: Boolean);
    procedure ClearButtonClick(Sender: TObject);
    procedure GridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure AllButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure NoneButtonClick(Sender: TObject);
    procedure CutButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure DeleteButtonClick(Sender: TObject);
    procedure CloseButtonClick(Sender: TObject);
    procedure AssignButtonClick(Sender: TObject);
    procedure GridGetEditText(Sender: TObject; ACol, ARow: Integer;
      var Value: String);
    procedure GridChangeCellQuery(Sender: TObject; Col, Row: Integer;
      var CanChange: Boolean);
    procedure GridSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: String);
    procedure DeleteUnassignedClick(Sender: TObject);
    procedure MinusButtonClick(Sender: TObject);
    procedure StrengthSortItemClick(Sender: TObject);
    procedure UpperStateSortItemClick(Sender: TObject);
    procedure FrequencySortItemClick(Sender: TObject);
    procedure LowerStateSortItemClick(Sender: TObject);
    procedure WidthSortItemClick(Sender: TObject);
    procedure WeightSortItemClick(Sender: TObject);
    procedure SortOnButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FitButtonClick(Sender: TObject);
  private
    NotEmpty: Boolean;
    Lines: array of TLineW;
    nLines: Integer;
    LastTemperature: Double;
    LastTvib: Double;
    nPendingLines: Integer;
    FLongestNameLength: Integer;
    procedure InternalAddLines(First, Last: Integer);
    function nVisibleLines: Integer;
    function StringRep(ARow, ACol: Integer; var buffer: string): Boolean;
  protected
    procedure UpdateActions; override;
  public
    procedure AddLine(ALine: TLine; AUnits: TResonanceUnits);
    procedure ShowLines(Temperature, TVib: Double);
    procedure BringUp;
    procedure ClearTLines;
    function HasObservations:Boolean;
    procedure LoadObservations;
    procedure SortLines(SortFunction: TLineSortFunction; First, Last: Integer);
    procedure CopyToStream(Stream: TStream; Row1, Row2: Integer; All: Boolean);
  end;

var
  LineForm: TLineForm;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF} Math, MathSup, TextAssign, Huge, ClipBrd, SlaveEdit,
  Main, LogUnit, Obs;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

var
  CF_Measurement, CF_Difference: DWORD;

{ TLineForm }

procedure TLineForm.AddLine(ALine: TLine; AUnits: TResonanceUnits);
  function StateDef(Eigenvalues: TEigenvalues; AIndex: Integer):TStateDef;
  begin
    with Result do begin
      Manifold := Eigenvalues.Manifold.Name;
      J := Eigenvalues.J;
      Sym := Eigenvalues.Sym;
    end;
  end;
begin
  if nLines >= Length(Lines) then
    SetLength(Lines, nLines+50);
  with Lines[nLines] do begin
    Line := ALine;
    ConvertLine(Line, AUnits, MainForm.Mixture.PlotUnits);
    Molecule := ALine.EigenvaluesL.Manifold.Molecule.Name;
    Upper := StateDef(ALine.EigenvaluesU, ALine.IndexU);
    Lower := StateDef(ALine.EigenvaluesL, ALine.IndexL);
    SetNan(Weight);
  end;
  Inc(nLines);
end;

procedure TLineForm.SortLines(SortFunction: TLineSortFunction; First, Last: Integer);
var
  N, I, J, Offset: Integer;
  IsInOrder: Boolean;
  T: TLineW;
begin
  Grid.EditorMode := False;
  { ShellSort }
  N := Last-First+1;
  Offset := N;
  while Offset > 1 do begin
    Offset := Offset shr 1;
    repeat
      IsInOrder := True;
      for J := First to First + N - Offset - 1 do begin
        I := J + Offset;
        if SortFunction(Lines[I], Lines[J]) > 0 { I > J? } then begin
          IsInOrder := False;
          T := Lines[I];
          Lines[I] := Lines[J];
          Lines[J] := T;
        end;
      end;
    until IsInOrder;
  end;
  InternalAddLines(First, Last);
end;

function ByIntensity(const A, B: TLineW): Integer;
begin
 Result := Sign(A.Line.Strength - B.Line.Strength);
end;

function ByFrequency(const A, B: TLineW): Integer;
begin
 Result := -Sign(A.Line.Frequency - B.Line.Frequency);
end;

function ByWidth(const A, B: TLineW): Integer;
begin
 Result := Sign(A.Line.Width - B.Line.Width);
end;

function ByWeight(const A, B: TLineW): Integer;
begin
 if IsNanOrInf(A.Weight) then
   if IsNanOrInf(B.Weight) then
     Result := 0
   else
     Result := -1
 else if IsNanOrInf(B.Weight) then
   Result := 1
 else
   Result := Sign(A.Weight - B.Weight);
end;

function ByEigenvalues(const A, B: TStateDef): Integer;
begin
  Result := CompareText(A.Manifold, B.Manifold);
  if Result = 0 then
    Result := A.J - B.J;
  if Result = 0 then
    Result := A.Sym - B.Sym;
end;

function ByUpperState(const A, B: TLineW): Integer;
begin
 Result := CompareText(A.Molecule, B.Molecule);
 if Result = 0 then
   Result := ByEigenvalues(A.Upper, B.Upper);
 if Result = 0 then
   Result := A.Line.IndexU - B.Line.IndexU;
 Result := -Result;
end;

function ByLowerState(const A, B: TLineW): Integer;
begin
 Result := CompareText(A.Molecule, B.Molecule);
 if Result = 0 then
   Result := ByEigenvalues(A.Lower, B.Lower);
 if Result = 0 then
   Result := A.Line.IndexL - B.Line.IndexL;
 Result := -Result;
end;

procedure SplitCachedName(var Line: TLine; var Branch, TxName: string);
var
  p: Integer;
begin
  TxName := MakeTxName(Line);
  p := Pos(':',TxName);
  if p > 0 then begin
    Branch := Copy(TxName,1,p-1);
    if (Length(Branch) > 0) and (Branch[Length(Branch)] = ' ') then
      Delete(Branch,Length(Branch),1);
    Delete(TxName,1,p);
    if (TxName <> ' ') and (TxName[1] = ' ') then
      Delete(TxName,1,1);
  end else
    Branch := '';
end;

function ByBranch(const A, B: TLineW): Integer;
var
  BranchA, BranchB: string;
  function RemoveBrackets(const L: TLineW):string;
  { Branch label without (J) }
  var
    Line: TLine;
    Junk: string;
    s, e: Integer;
  begin
    Line := L.Line;
    SplitCachedName(Line, Result, Junk);
    s := Pos('(',Result);
    e := Pos(')',Result);
    if (s > 0) and (e > s) then
      Delete(Result, s, e-s+2);
  end;
begin
  BranchA := RemoveBrackets(A);
  BranchB := RemoveBrackets(B);
  Result := CompareText(BranchB, BranchA);
  if Result = 0 then
    Result := B.Lower.J - A.Lower.J;
end;

procedure TLineForm.ShowLines(Temperature, Tvib: Double);
var
  i, f: Integer;
  Rect: TGridRect;
begin
  if nLines <= nVisibleLines then Exit;
  Grid.BeginUpdate;
  LastTemperature := Temperature;
  LastTvib := Tvib;
  f := nVisibleLines;
  for i := f to nLines - 1 do
    with Lines[i].Line do
      if Assigned(EigenvaluesU) and Assigned(EigenvaluesL) then
        Strength := MainForm.Mixture.LineIntensity(Lines[i].Line, Temperature, Tvib);
  SortLines({$IFDEF FPC} @ {$ENDIF}ByIntensity, nVisibleLines, nLines-1);
  InternalAddLines(f, nLines - 1);
  nPendingLines := 0;
  Rect.Top := f+1;
  Rect.Bottom := Grid.RowCount-1;
  Rect.Left := 1;
  Rect.Right := Grid.ColCount-1;
  Grid.Selection := Rect;
  BringUp;
  Grid.EndUpdate;
end;

procedure TLineForm.InternalAddLines(First, Last: Integer);
var
  i, dn: Integer;
  procedure MakeTitle(U, L: TBasisState);
  begin
    Grid.Title[14] := U.TxNameKey(U,L);
    Grid.Title[15] := U.StateLabelKey + ' - ' + L.StateLabelKey;
  end;
begin
  if First > Last then Exit;
  dn := Last - nVisibleLines + 1; { Number added }
  if dn > 0 then
    if NotEmpty then
      Grid.RowCount := Grid.RowCount + dn
    else
      Grid.RowCount := Grid.RowCount + dn - 1;
  NotEmpty := True;
  if First = 0 then begin
    if Assigned(Lines[0].Line.EigenvaluesU) then
      MakeTitle(Lines[0].Line.EigenvaluesU.Basis[1], Lines[0].Line.EigenvaluesL.Basis[1])
    else begin
      Grid.Title[14] := 'Branch';
      Grid.Title[15] := 'Name';
    end;
    Grid.AutoAdjustColumn(14);
  end;
  for i := First+1 to Last+1 do
    with Lines[i-1] do begin
      Grid[i,3] := Upper.J;
      Grid[i,4] := Upper.Sym;
      Grid[i,5] := Line.IndexU;
      Grid[i,7] := Lower.J;
      Grid[i,8] := Lower.Sym;
      Grid[i,9] := Line.IndexL;
      Grid[i,10] := Line.Frequency;
      Grid[i,11] := Weight;
      Grid[i,12] := Line.Strength;
      Grid[i,13] := Line.Width;
    end;
  Grid.Invalidate;
end;

procedure TLineForm.BringUp;
begin
  if not Visible then
    Show
  else
    BringToFront;
end;

procedure TLineForm.ClearButtonClick(Sender: TObject);
var
  i: Integer;
begin
  Grid.RowCount := 2;
  NotEmpty := False;
  nLines := 0;
  for i := 1 to Grid.ColCount-1 do
    Grid.Valid[1,i] := False;
  Grid.ColWidths[Grid.ColCount-1] := 150;
  Grid.Invalidate;
end;

procedure TLineForm.GridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  buffer: string;
  Pitch: TFontPitch;
  FontName: TFontName;
  L: Integer;
begin
  if StringRep(ARow, ACol, buffer) then
    if ACol = 15 then begin
      Pitch := Grid.Canvas.Font.Pitch;
      {$IFDEF MSWINDOWS}
      FontName := Grid.Canvas.Font.Name;
      Grid.Canvas.Font.Name := 'Courier New';
      Grid.Canvas.Font.Size := Grid.Canvas.Font.Size + 1;
      {$ENDIF}
      Grid.Canvas.Font.Pitch := fpFixed;
      L := Grid.Canvas.TextWidth(buffer) + 10;
      if FLongestNameLength < L then
        FLongestNameLength := L;
      TextInGrid(Grid.Canvas, Rect, buffer);
      {$IFDEF MSWINDOWS}
      Grid.Canvas.Font.Size := Grid.Canvas.Font.Size - 1;
      Grid.Canvas.Font.Name := FontName;
      {$ENDIF}
      Grid.Canvas.Font.Pitch := Pitch;
    end else
      TextInGrid(Grid.Canvas, Rect, buffer);
end;

procedure TLineForm.AllButtonClick(Sender: TObject);
var
  Rect: TGridRect;
begin
  Rect.Top := 1;
  Rect.Bottom := Grid.RowCount-1;
  Rect.Left := 1;
  Rect.Right := Grid.ColCount-1;
  Grid.Selection := Rect;
end;

procedure TLineForm.FormCreate(Sender: TObject);
begin
  {$IFDEF LCL_}
  DeleteUnassigned.Width := DeleteUnassigned.Width + 4;
  {$ENDIF}
end;

procedure TLineForm.NoneButtonClick(Sender: TObject);
begin
  Grid.CancelSelection;
end;

procedure TLineForm.CutButtonClick(Sender: TObject);
begin
  CopyButtonClick(Sender);
  DeleteButtonClick(Sender);
end;

procedure TLineForm.CopyToStream(Stream: TStream; Row1, Row2: Integer; All: Boolean);
var
  F: TextFile;
  i, j: Integer;
  buffer: string;
begin
  AssignStream(F, Stream);
  Rewrite(F);
  for i := Row1 to Row2 do begin
    if All or Grid.Valid[i,11] then
      for j := 1 to Grid.ColCount-1 do
        case j of
         10: Write(F, Grid[i,j]:12:4, ' ');
         11: if Grid.Valid[i,j] then
               Write(F, FloatToStr(Grid[i,j]), ' ')
             else
               Write(F, '- ');
         12, 13: Write(F, dtosm(Grid[i,j], 6, -1):6, ' '); { Limit Strength and width to 6 characters }
         14: begin StringRep(i,j,buffer); Write(F,': ', buffer); end;
         15: begin StringRep(i,j,buffer); WriteLn(F,' : ', buffer); end;
        else
          if StringRep(i, j, buffer) then
            Write(F, buffer:3, ' ')
          else
            Write(F, FloatToStr(Grid[i,j]):2, ' ');
        end;
  end;
  Write(F,#0);
  System.Close(F);
  Stream.Position := 0;
end;


procedure TLineForm.CopyButtonClick(Sender: TObject);
var
  s: THugeMemoryStream;
begin
  if not NotEmpty then Exit;
  s := THugeMemoryStream.Create;
  try
    CopyToStream(s, Grid.Selection.Top, Grid.Selection.Bottom, True);
    s.CopyToClipboard(CF_TEXT);
  finally
    s.Free;
  end;
end;

procedure TLineForm.DeleteButtonClick(Sender: TObject);
var
  i, d, n, e: Integer;
begin
  { Remember how many lines to recopy }
  d := Grid.Selection.Top;
  n := Grid.Selection.Bottom+1-d;
  e := Grid.RowCount - n - 2;
  for i := d-1 to e do
    Lines[i] := Lines[i+n];
  nLines := e+1;
  if d = 1 then
    ClearButtonClick(nil)
  else
    Grid.RowCount := d;
  { Copy remaining lines down }
  InternalAddLines(d-1, e);
end;

procedure TLineForm.CloseButtonClick(Sender: TObject);
begin
  Close;
end;

type
  ClipboardMeasurement = packed record
    Value: Double;
    Signature: array [0..3] of Char;
    Filename: Char;
  end;
  PClipboardMeasurement = ^ClipboardMeasurement;

procedure TLineForm.AssignButtonClick(Sender: TObject);
var
  Value, StrMax: Double;
  CHandle: TGHandle;
  Clip: PClipboardMeasurement;
  FileName, s, Branch: string;
  i, p: Integer;
begin
  if NotEmpty and Visible and ClipBoard.HasFormat(CF_TEXT) then begin
    if CF_Measurement = 0 then
      CF_Measurement := RegisterClipboardFormat('Measurement');
    Value := StrToFloat(ClipBoard.AsText);
    if ClipBoard.HasFormat(CF_Measurement) then begin
      CHandle := TGHandle.Create;
      try
        CHandle.SetFromClipboard(CF_Measurement);
        Clip := PClipboardMeasurement(CHandle.Base);
        if Clip^.Signature = 'MEAS' then begin
          Value := Clip^.Value;
          FileName := ExtractFileName(PChar(@Clip^.FileName));
        end;
      finally
        CHandle.Free;
      end;
    end else
      FileName := '';
    StrMax := Lines[Grid.Selection.Top-1].Line.Strength;
    for i := Grid.Selection.Top+1 to Grid.Selection.Bottom do
      if StrMax < Lines[i-1].Line.Strength then
        StrMax := Lines[i-1].Line.Strength;
    for i := Grid.Selection.Top to Grid.Selection.Bottom do
      if Lines[i-1].Line.Strength > StrMax/4 then begin
        Lines[i-1].Line.Frequency := Value;
        SplitCachedName(Lines[i-1].Line, Branch, s);
        p := Pos(':', s);
        if p > 0 then begin
          if (p > 1) and (s[p-1] = ' ') then
            Dec(p); { Trim trailing space }
          Delete(s,p,Length(s));
        end;
        if FileName <> '' then
          s := s + ' : ' + ExtractFileName(FileName);
        if Branch <> '' then
          s := Branch + ' : ' + s;
        Lines[i-1].Line.CachedName := s;
        Lines[i-1].Weight := 1;
      end;
    InternalAddLines(Grid.Selection.Top-1, Grid.Selection.Bottom-1);
  end;
end;

function TLineForm.StringRep(ARow, ACol: Integer; var buffer: string): Boolean;
var
  Junk: string;
begin
  Result := NotEmpty and (ARow > 0);
  if Result then
    case ACol of
      1: buffer := Lines[ARow-1].Molecule;
      2: buffer := Lines[ARow-1].Upper.Manifold;
      6: buffer := Lines[ARow-1].Lower.Manifold;
     14: SplitCachedName(Lines[ARow-1].Line, buffer, Junk);
     15: SplitCachedName(Lines[ARow-1].Line, Junk, buffer);
    else
      Result := False;
    end;
end;

procedure TLineForm.GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: String);
begin
  StringRep(ARow, ACol, Value);
end;

procedure TLineForm.GridChangeCellQuery(Sender: TObject; Col, Row: Integer; var CanChange: Boolean);
begin
  // Was : CanChange := CanChange and (Col in [10,11,14]);
end;

procedure TLineForm.GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
var
  v: Double;
  Branch, TxName: string;
begin
  if ACol in [10,11] then
    if not TryStrToFloat(Value, v) then
      SetNan(v);
  case ACol of { TODO: Is this copy back necessary for numerical values? }
     1: Lines[ARow-1].Molecule := Value;
     2: Lines[ARow-1].Upper.Manifold := Value;
     6: Lines[ARow-1].Upper.Manifold := Value;
    10: Lines[ARow-1].Line.Frequency := v;
    11: Lines[ARow-1].Weight := v;
    14,15:
      begin
        SplitCachedName(Lines[ARow-1].Line, Branch, TxName);
        if ACol = 14 then
          Branch := Value
        else
          TxName := Value;
        if Branch <> '' then
          TxName := Branch + ' : ' + Value;
        Lines[ARow-1].Line.CachedName := TxName;
      end;
  end;
end;

function TLineForm.nVisibleLines: Integer;
begin
  if NotEmpty then
    Result := Grid.RowCount-1
  else
    Result := 0
end;

procedure TLineForm.DeleteUnassignedClick(Sender: TObject);
var
  s, e: Integer;
  r: TGridRect;
begin
  Grid.BeginUpdate;
  repeat
    e := nVisibleLines;
    while (e > 0) and Grid.Valid[e,11] do
      Dec(e);
    s := e;
    while (s > 0) and not Grid.Valid[s,11] do
      Dec(s);
    Inc(s);
    if e >= s then begin
      r.Top := s;
      r.Bottom := e;
      r.Left := 1;
      r.Right := 2;
      Grid.Selection := r;
      DeleteButtonClick(nil);
    end else
      Break;
  until False;
  Grid.EndUpdate;
end;


procedure TLineForm.MinusButtonClick(Sender: TObject);
var
  d: Double;
  Stream: THugeMemoryStream;
begin
  if Grid.Selection.Top >= Grid.Selection.Bottom then begin
    MessageDlg('The button takes the difference between the first and last selected line', mtInformation, [mbOK], 0);
    Exit;
  end;
  if CF_Difference = 0 then
    CF_Difference := RegisterClipboardFormat('Difference');
  with Grid.Selection do
    d := Abs(Grid[Bottom,10] - Grid[Top,10]);
  Difference.Caption := dtosm(d, 12, 4);
  Stream := THugeMemoryStream.Create;
  Clipboard.Open;
  try
    Clipboard.AsText := FloatToStr(d);
    Stream.Write(d, Sizeof(Double));
    Stream.CopyToClipboard(CF_Difference);
  finally
    Clipboard.Close;
    Stream.Free;
  end;
  BroadcastName('WM_New_Difference');
end;

procedure TLineForm.StrengthSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByIntensity, 0, nLines-1);
end;

procedure TLineForm.UpperStateSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByUpperState, 0, nLines-1);
end;

procedure TLineForm.BranchItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByBranch, 0, nLines-1);
end;

procedure TLineForm.FrequencySortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByFrequency, 0, nLines-1);
end;

procedure TLineForm.LowerStateSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByLowerState, 0, nLines-1);
end;

procedure TLineForm.WidthSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByWidth, 0, nLines-1);
end;

procedure TLineForm.WeightSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByWeight, 0, nLines-1);
end;

procedure TLineForm.SortOnButtonClick(Sender: TObject);
begin
  SortPopupMenu.PopupComponent := SortOnButton;
  SortPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TLineForm.FormDestroy(Sender: TObject);
begin
  LineForm := nil;
end;

procedure TLineForm.FitButtonClick(Sender: TObject);
begin
  if not NotEmpty then Exit;
  LoadObservations;
  MainForm.LastObsWasLineForm := True;
  LogForm.Show;
  MainForm.DoLineFit;
end;

procedure TLineForm.ClearTLines;
var
  i: Integer;
begin
  for i := 0 to High(Lines) do
    with Lines[i] do begin
      Line.EigenvaluesU := nil;
      Line.EigenvaluesU := nil;
    end;
end;

function TLineForm.HasObservations: Boolean;
var
  i: Integer;
begin
  Result := False;
  if not NotEmpty then Exit;
  for i := 0 to Grid.RowCount-2 do
    if not (IsNan(Lines[i].Weight)) and (Lines[i].Weight > 0) then begin
      Result := True;
      Exit;
    end;
end;

procedure TLineForm.LoadObservations;
var
  s: THugeMemoryStream;
begin
  s := THugeMemoryStream.Create;
  try
    CopyToStream(s, 1, Grid.RowCount-1, False);
    Obs.LoadObservations(MainForm.Mixture, s, '<LineList>');
  finally
    s.Free;
  end;
end;

procedure TLineForm.UpdateActions;
begin
  inherited;
  if FLongestNameLength > Grid.ColWidths[Grid.ColCount-1] then
    Grid.ColWidths[Grid.ColCount-1] := FLongestNameLength;
  FLongestNameLength := 0;
end;

procedure TLineForm.GridGetCellLabel(Sender: TObject; Col, Row: Integer;
  var s: string; var IsText: Boolean);
begin
  if (Row > 0) and (Col > 0) then
    IsText := StringRep(Row, Col, s);
end;

initialization
{$IFDEF LCL_}
  {$I linefrm.lrs}
{$ENDIF}
end.
