unit LogUnit;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, ActnList,
  {$IFNDEF LCL_} FileNameEdit, {$ELSE} EditBtn, {$ENDIF}
  Dialogs, ExtCtrls, ScrlCtrl, Log, Buttons, StdCtrls, AlEdit, EditExtr,
  NewSpin, MNode, WhatsThis;

type

  { TLogForm }

  TLogForm = class(THelpForm)
    Log: TLog;
    Panel1: TPanel;
    ObsFileName: TFileNameEdit;
    FitSpeed: TSpeedButton;
    EditButton: TSpeedButton;
    ContourFitButton: TSpeedButton;
    ClearButton: TSpeedButton;
    CopyButton: TSpeedButton;
    SaveDialog: TSaveDialog;
    SaveLogButton: TSpeedButton;
    UndoFitButton: TSpeedButton;
    RedoFitButton: TSpeedButton;
    ActionList: TActionList;
    UndoFitAction: TAction;
    RedoFitAction: TAction;
    SpinCycles: TSpinBox;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FitSpeedClick(Sender: TObject);
    procedure EditButtonClick(Sender: TObject);
    procedure ContourFitButtonClick(Sender: TObject);
    procedure ClearButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure LogFormDestroy(Sender: TObject);
    procedure SaveLogButtonClick(Sender: TObject);
    procedure UndoFitActionExecute(Sender: TObject);
    procedure UndoFitActionUpdate(Sender: TObject);
    procedure RedoFitActionExecute(Sender: TObject);
    procedure RedoFitActionUpdate(Sender: TObject);
    procedure FitSpeedMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SpinCyclesChange(Sender: TObject);
  private
    procedure MNodeNotification(Sender: TMNode; Change: TMNodeChange; Index: Integer);
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
  public
    procedure LoadObservations(AName: string = '');
  end;

var
  LogForm: TLogForm;

implementation

uses {$IFDEF MSWINDOWS} Windows, ShellAPI, {$ENDIF} {$IFDEF LCL_} LResources, {$ENDIF}
   TextAssign, Obs, Detach, Fitter, Main, Mixture, SlaveEdit, Huge, ClipBrd, WCompat;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

procedure TLogForm.FormCreate(Sender: TObject);
begin
  ClearButtonClick(nil);
  AddMNodeNotification({$IFDEF FPC} @ {$ENDIF}MNodeNotification);
  DragAcceptFiles(Handle, True);
end;

procedure TLogForm.EditButtonClick(Sender: TObject);
begin
  LaunchEditor(ObsFileName.FileName);
  {$IFDEF OLD}
  if not FileExists(ObsFileName.FileName) then
    if MessageDlg('File ' + ObsFileName.FileName + ' does not exist. Create?', mtConfirmation, [mbYes,mbNo], 0) = mrNo then
      Exit
    else
      FileClose(FileCreate(ObsFileName.FileName));
  {$IFDEF MSWINDOWS}
  if ShellExecute(0, nil, PChar(ObsFileName.FileName), nil, PChar(ExtractFileDir(ObsFileName.FileName)), SW_SHOW) <= 32 then
    Detach1('notepad', ObsFileName.FileName);
  {$ELSE}
  Detach1(ObsFileName.FileName, '');
  {$ENDIF}
  {$ENDIF}
end;

procedure TLogForm.LoadObservations(AName: string);
var
  f: TStream;
begin
  if AName = '' then
    AName := ObsFileName.FileName;
  if AName = '' then
    AName := '<clipboard>';
  if AnsiCompareText(AName, '<clipboard>') = 0 then begin
    f := THugeMemoryStream.Create;
    THugeMemoryStream(f).SetFromClipboard(CF_TEXT);
  end else begin
    MainForm.MRUFileList.AddItem(ExpandFileName(AName));
    f := TFileStream.Create(AName, fmOpenRead);
  end;
  try
    Obs.LoadObservations(MainForm.Mixture, f, AName);
  finally
    f.Free;
  end;
  MainForm.LastObsWasLineForm := False;
end;

procedure TLogForm.FitSpeedClick(Sender: TObject);
begin
  if not SaveAllWindows then Exit;
  LoadObservations(ObsFileName.FileName);
  MainForm.DoLineFit;
end;

procedure TLogForm.FitSpeedMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if mbRight = Button then begin
    LoadObservations('<clipboard>');
    MainForm.DoLineFit;
  end;
end;

procedure TLogForm.ContourFitButtonClick(Sender: TObject);
begin
  MainForm.ContourFit;
end;

procedure TLogForm.ClearButtonClick(Sender: TObject);
begin
  AssignStrings(Output, Log.Lines);
  Rewrite(Output);
end;

procedure TLogForm.CopyButtonClick(Sender: TObject);
begin
  if not Log.AreaSelected then
    Log.SelectAll;
  Log.CopyToClipboard
end;

procedure TLogForm.LogFormDestroy(Sender: TObject);
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeNotification);
  {$IFDEF LINUX}
  AssignFile(Output,'/proc/self/fd/1');
  Rewrite(Output);
  {$ELSE}
  CloseFile(Output);
  {$ENDIF}
end;

procedure TLogForm.SaveLogButtonClick(Sender: TObject);
begin
  if SaveDialog.Execute then
    Log.Lines.SaveToFile(SaveDialog.FileName);
end;

procedure TLogForm.UndoFitActionExecute(Sender: TObject);
begin
  MainForm.Mixture.UndoFit;
end;

procedure TLogForm.UndoFitActionUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled := Assigned(MainForm.Mixture) and MainForm.Mixture.CanUndoFit;
end;

procedure TLogForm.RedoFitActionExecute(Sender: TObject);
begin
  MainForm.Mixture.RedoFit;
end;

procedure TLogForm.RedoFitActionUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled := Assigned(MainForm.Mixture) and MainForm.Mixture.CanRedoFit;
end;

procedure TLogForm.MNodeNotification(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  if Assigned(MainForm.Mixture) and (Sender = MainForm.Mixture) then
    SpinCycles.Value := MainForm.Mixture.FitCycles;
end;

procedure TLogForm.SpinCyclesChange(Sender: TObject);
begin
  if Assigned(MainForm.Mixture) then
    MainForm.Mixture.FitCycles := SpinCycles.Value;
end;

procedure TLogForm.WMDropFiles(var Message: TMessage);
begin
  MainForm.WMDropFiles(Message);
end;

initialization
{$IFDEF LCL_}
  {$I LogUnit.lrs}
{$ENDIF}
end.
