unit Nlfit;

interface

uses Vector, MatrixVec;

procedure LeastSQ(
                  a: TMatrix;   { d(obs)/d(par) }
                  obs: TColumnVector; { normally obs - calc }
                  par: TColumnVector; { Parameters - changed by this routine }
                  weights: TColumnVector; { nil or The relative standard deviation of each value in obs }
                  { Output: }
                  var OldSD, NewSD: Double; { Previous and estimated next standard deviations }
                  var UnWtOldSD, UnWtNewSD: Double; { Previous and estimated next standard deviations ignoring weights }
                  b: TColumnVector;  { Standard Deviations of the parameters}
                  ada: TMatrix; { Correlation Matrix }
                  SVD: TMatrix; SVDw: TColumnVector; { Details of Singular Value Decomposition }
                  SVDThresh: Double; { Values smaller than this are discarded }
                  {$IFDEF NOSVD}
                    damp: Double {Damping Factor - normally 0}
                  {$ELSE}
                    ScaleChanges: Double { Scale changes by this - normally 1 }
                  {$ENDIF}
                 );

implementation

uses {$IFDEF NOSVD} GaussJ; {$ELSE} SVD; {$ENDIF}

{ Makes least squares adjustment of parameters in par to minimise
  sum (obs(i)/weights(i))**2 given d(obs)/d(par) in a }
procedure LeastSQ(
                  a: TMatrix;   { d(obs)/d(par) }
                  obs: TColumnVector; { normally obs - calc }
                  par: TColumnVector; { Parameters - changed by this routine }
                  weights: TColumnVector; { nil or The relative standard deviation of each value in obs }
                  var OldSD, NewSD: Double; { Previous and estimated next standard deviations }
                  var UnWtOldSD, UnWtNewSD: Double; { Previous and estimated next standard deviations ignoring weights }
                  b: TColumnVector;  { Standard Deviations of the parameters }
                  ada: TMatrix; { Correlation Matrix }
                  SVD: TMatrix; SVDw: TColumnVector; { Details of Singular Value Decomposition }
                  SVDThresh: Double; { Values smaller than this are discarded }
                  {$IFDEF NOSVD}
                    damp: Double {Damping Factor - normally 0}
                  {$ELSE}
                    ScaleChanges: Double { Scale changes by this - normally 1 }
                  {$ENDIF}
                 );
{$IFNDEF NOSVD}
var
  U: TMatrix;
  obsb: TColumnVector;
  w: Double;
  acum, acum1, s: Extended;
  i, j, k: Integer;
begin
  U := nil;
  obsb := nil;
  try
    U := TMatrix.Create(obs.Size, par.Size);
    obsb := TColumnVector.Create(obs.Size);
    if weights <> nil then begin
      for i := 1 to obs.Size do begin
        w := weights[i];
        obsb[i] := obs[i] / w;
        for j := 1 to par.Size do
          U[i,j] := a[i,j] / w;
      end;
    end else begin
      U.Assign(a);
      obsb.Assign(obs);
    end;
    { Intial Error }
    acum := 0;
    acum1 := 0;
    for i := 1 to obs.Size do begin
     acum := acum + Sqr(obsb[i]);
     acum1 := acum1 + Sqr(obs[i]);
    end;
    if obs.Size > par.Size then begin
      OldSD := Sqrt(acum/(obs.Size - par.Size));
      UnWtOldSD := Sqrt(acum1/(obs.Size - par.Size));
    end else begin
      OldSD := Sqrt(acum);
      UnWtOldSD := Sqrt(acum1);
    end;
{   ShowMatrix(U, 'A', nil); }
    SV_Decompose(U, SVDw, U, SVD);
{   ShowMatrix(U, 'U', nil);
    ShowMatrix(SVD, 'V', SVDw); }
    if SVDThresh > 0 then begin
      s := 0;
      for i := 1 to par.Size do
        if s < SVDw[i] then s := SVDw[i];
      s := s * SVDThresh;
      for i := 1 to par.Size do
        if SVDw[i] < s then SVDw[i] := 0;
    end;
    SV_Solve(U, SVDw, SVD, obsb, b);
    { Calculate estimated residuals
      calc =  a * b
      and standard deviation of fit }
    acum := 0;
    acum1 := 0;
    for i := 1 to obs.Size do begin
     s := 0;
     for j := 1 to par.Size do
       s := s + a[i,j]*b[j];
     if weights <> nil then
      acum := acum + Sqr(obs[i]/weights[i]-s);
     acum1 := acum1 + Sqr(obs[i]-s);
    end;
    if weights = nil then
      acum := acum1;
    if obs.Size > par.Size then begin
      NewSD := Sqrt(acum/(obs.Size - par.Size));
      UnWtNewSD := Sqrt(acum1/(obs.Size - par.Size));
    end;
    for i := 1 to par.Size do
      par[i] := par[i] + ScaleChanges*b[i];
    for i := 1 to par.Size do
      if SVDw[i] > 0 then
        obsb[i] := 1 / Sqr(SVDw[i])
      else obsb.Valid[i] := False;
    { Variance - covariance matrix }
    for i := 1 to par.Size do
      for j := 1 to i do
        if obsb.Valid[i] and obsb.Valid[j] then begin
          s := 0;
          for k := 1 to par.Size do
            if obsb.Valid[k] then
              s := s + SVD[i,k] * SVD[j,k] * obsb[k];
          ada[i,j] := s;
          ada[j,i] := s;
        end else begin
          ada.Valid[i,j] := False;
          ada.Valid[i,j] := False;
        end;
    for i := 1 to par.Size do
      if obsb.Valid[i] and (ada[i,i] > 0) then
        b[i] := Sqrt(ada[i,i])
      else
        b.Valid[i] := False;
    for i := 1 to par.Size do
      for j := 1 to i do begin
        if b.Valid[i] and b.Valid[j] then
          ada[i,j] := ada[i,j] / (b[i] * b[j])
        else
          ada.Valid[i,j] := False;
      end;
    for i := 1 to par.Size do
     if b.Valid[i] then
       b[i] := NewSD * b[i];
  finally
    U.Free;
    obsb.Free;
  end;
end;
{$ELSE}
var
  acum, acum1, s: Extended;
  i, j: Integer;
begin
  acum1 := obs.DotProduct(obs);
  if weights <> nil then begin
    { b := transpose(a)*obs }
    for i := 1 to par.Size do begin
      acum := 0;
      for j := 1 to obs.Size do begin
        a[j, i] := a[j, i] / weights[j];
        acum := acum + a[j,i] * obs[j] / weights[j];
      end;
      b[i] := acum;
    end;
    acum := 0;
    for i := 1 to obs.Size do
      acum := acum + Sqr(obs[i])/Sqr(weights[i]);
  end else begin
{   for i := 1 to par.Size do begin
     acum := 0;
      for j := 1 to obs.Size do
        acum := acum + a[j,i] * obs[j];
      b[i] := acum;
    end;
    acum := 0;
    for i := 1 to obs.Size do
      acum := acum + Sqr(obs[i]); }
    for i := 1 to par.Size do
      b[i] := a.Column[i].DotProduct(obs);
    acum := acum1;
  end;
  if obs.Size > par.Size then begin
    OldSd := Sqrt(acum/(obs.Size-par.Size));
    UnWtOldSd := Sqrt(acum1/(obs.Size-par.Size));
  end;
  for i := 1 to par.Size do
    for j := 1 to par.Size do
      ada[i, j] := a.Column[i].Dotproduct(a.Column[j]);
  if damp <> 0 then
    for i := 1 to par.Size do
      ada[i,i] := ada[i,i] * (1+damp);
  { invert ada and calculate change in parameters,
    b := ada**-1 * b }
  GaussJordan(ada, b);
  { Calculate estimated residuals
    calc =  a * b
    and standard deviation of fit }
  acum := 0;
  acum1 := 0;
  for i := 1 to obs.Size do begin
   s := 0;
   for j := 1 to par.Size do
     s := s + a[i,j]*b[j];
   if weights <> nil then
    acum := acum + Sqr(obs[i]/weights[i]-s);
   acum1:= acum1+ Sqr(obs[i]-s);
  end;
  if weights = nil then
    acum := acum1;
  if obs.Size > par.Size then begin
    NewSD := Sqrt(acum/(obs.Size - par.Size));
    UnWtNewSD := Sqrt(acum1/(obs.Size - par.Size));
  end;
  for i := 1 to par.Size do
   par[i] := par[i] + b[i];
  { correlation  matrix }
  for i := 1 to par.Size do
    if ada[i,i] > 0 then
      b[i] := Sqrt(ada[i,i])
    else
      b.Valid[i] := False;
  for i := 1 to par.Size do
    for j := 1 to i do begin
      if b.Valid[i] and b.Valid[j] then
        ada[i,j] := ada[i,j] / (b[i] * b[j])
      else
        ada.Valid[i,j] := False;
    end;
  for i := 1 to par.Size do
   if b.Valid[i] then
     b[i] := NewSD * b[i];
end;
{$ENDIF}

end.

