unit RenForm;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  WhatsThis, Dialogs, StdCtrls, Buttons, MNode;

type
  TRenameForm = class(THelpForm)
    OKButton: TButton;
    CancelButton: TButton;
    BraComboBox: TComboBox;
    Label1: TLabel;
    Label2: TLabel;
    KetComboBox: TComboBox;
  private
    MNodePair: TMNodePair;
  public
    function StuffComboBox(MNode: TMNode; ComboBox: TComboBox; Current: TMNode): string;
    procedure Rename(MNode: TMNodePair; ChangingMNode: TMNode = nil);
    procedure RenameMNodePairsAsRequired(MNode: TMNode);
    function UnStuffComboBox(MNode: TMNode; ComboBox: TComboBox): string;
  end;

var
  RenameForm: TRenameForm;

implementation

{$IFDEF LCL_}
uses LResources;
{$ENDIF}

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

{ TRenameForm }

procedure TRenameForm.Rename(MNode: TMNodePair; ChangingMNode: TMNode);
var
  E: Exception;
  BraDef, KetDef: string;
begin
  MNodePair := MNode;
  if (MNode.BraOwner = nil) or (MNode.KetOwner = nil) then
    raise Exception.Create('Fix parent node first');
  BraDef := StuffComboBox(MNode.BraOwner, BraComboBox, MNode.Bra);
  KetDef := StuffComboBox(MNode.KetOwner, KetComboBox, MNode.Ket);
  if not(Assigned(ChangingMNode) and (BraDef <> '') and (KetDef <> '')) then begin
    Caption := 'Rename ' + MNode.UniqueName;
    if (ShowModal <> mrOK) and not Assigned(ChangingMNode) then
      Exit; { No action taken }
  end;
  E := MNode.SetBraKetNameE(UnStuffComboBox(MNode.BraOwner, BraComboBox),
                            UnStuffComboBox(MNode.KetOwner, KetComboBox), ChangingMNode);
  { If child nodes need renaming as a result of this name change,
    prompt for this }
  if Assigned(E) then begin
    if Assigned(ChangingMNode) then
      RenameMNodePairsAsRequired(ChangingMNode)
    else
      RenameMNodePairsAsRequired(MNode);
    { Re-issue notifications, as these may have been skipped }
    MNode.Modified := True;
    PerformMNodeNotification(MNode, mnNameChange, -1);
  end;
end;

procedure TRenameForm.RenameMNodePairsAsRequired(MNode: TMNode);
var
  ProblemMNode: TMNodePair;
  E: Exception;
begin
  { If a paste MNode or other change means that the Bra or Ket of mnode pairs need
    changing, this will prompt as required }
  E := MNode.FixupProperties;
  while Assigned(E) do
    if E is ECantFindState then begin
      ProblemMNode := ECantFindState(E).MNodePair;
      E.Free;
      Rename(ProblemMNode, MNode);
      E := MNode.FixupProperties;
    end else if Assigned(E) then
      raise E;
end;

function TRenameForm.StuffComboBox(MNode: TMNode; ComboBox: TComboBox; Current: TMNode): string;
var
  I: Integer;
begin
  ComboBox.Clear;
  for I := 0 to MNode.ComponentCount-1 do
    if MNode.Components[I] is MNodePair.BraKetClass then begin
      ComboBox.Items.Add(TMNode(MNode.Components[i]).Name);
      if MNode.Components[I] = Current then
        ComboBox.ItemIndex := ComboBox.Items.Count-1;
    end;
  if ComboBox.Items.Count = 1 then begin
    Result := ComboBox.Items[0];
    ComboBox.ItemIndex := 0;
  end else
   Result := '';
end;

function TRenameForm.UnStuffComboBox(MNode: TMNode; ComboBox: TComboBox): string;
var
  I, J: Integer;
begin
  Result := '';
  J := ComboBox.ItemIndex;
  for I := 0 to MNode.ComponentCount-1 do
    if MNode.Components[I] is MNodePair.BraKetClass then
      if J = 0 then begin
        Result := TMNode(MNode.Components[I]).Name;
        Exit;
      end else
        Dec(J);
end;

initialization
{$IFDEF LCL_}
  {$I RenForm.lrs}
{$ENDIF}
end.
