{ Loosely based on Z_Prof tool by A. Baars (antonie.baars@wxs.nl) }

unit Profile;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

procedure TStart(var s:string);
procedure TEnd(var s:string);

procedure CalculateOverhead;

type
  TMark = record
    Passes: Longint;
    Start, PrevTime, TotalTime, Fastest: int64;
    Active: boolean;
    Name: string;
  end;

var
  Marks: array of TMark;
  Overhead: Integer;         { Time taken between null TStart() .. TEnd() }
  ExternalOverhead: Integer; { Time to perform null TStart() .. TEnd() }
  MarkLevel: Integer;
  ShowTimingAtEnd: Boolean = True;
  TimeChanged: Boolean = False;

implementation

uses Clock, Math, MathSup, SysUtils;

function TSection(var s:string):Integer;
begin
  if (Length(s) <> 4) or (s[4] <> #0) then begin
    Result := 0;
    { Search through current entries }
    while Result <= High(Marks) do begin
      if (s = Marks[Result].Name) or (Marks[Result].Name = '') then
        Break;
      Inc(Result);
    end;
    if Result > High(Marks) then
      SetLength(Marks, Length(Marks)+20);
    if Marks[Result].Name = '' then begin
      { New entry }
      Marks[Result].Name := s;
    end;
    SetLength(s,4);
    PInteger(@s[1])^ := Result;
  end else
    Result := PInteger(@s[1])^;
end;

procedure TStart(var s:string);
var
  m: Integer;
begin
  if Length(s) <> 4 then
    m := TSection(s)
  else
    m := PInteger(@s[1])^;
  if m > High(Marks) then
    m := TSection(s);
  with Marks[m] do
    {$IFDEF DEBUG_TIME}
    if Active then begin
      Active := False;
      raise Exception.CreateFmt('TStart(%s)', [Name])
    end else
    {$ENDIF}
    begin
      Active := True;
      Inc(MarkLevel);
      Start := Clocks;
    end;
end;

procedure TEnd(var s: string);
var
  t: int64;
  m: Integer;
begin
  t := Clocks;
  if Length(s) <> 4 then
    m := TSection(s)
  else
    m := PInteger(@s[1])^;
  if m > High(Marks) then
    m := TSection(s);
  with Marks[m] do
    {$IFDEF DEBUG_TIME}
    if not Active then
      raise Exception.CreateFmt('TEnd(%d)', [m])
    else
    {$ENDIF}
    begin
      PrevTime := t - Start - Overhead;
      Active := False;
      TotalTime := TotalTime + PrevTime;
      if (PrevTime < Fastest) or (Passes = 0) then
        Fastest := PrevTime;
      Inc(Passes);
      Dec(MarkLevel);
      if MarkLevel <= 0 then begin
        TimeChanged := True;
        MarkLevel := 0;
      end;
    end;
end;

procedure CalculateOverhead;
var
  s, s1: string;
  i, n: Integer;
begin
  s1 := '__Overhead';
  n := TSection(s1);
  s := '';
  TSection(s); { Use a blank entry }
  Overhead := 0;
  TStart(s1);
  for i := 1 to 1000 do begin
    TStart(s); TEnd(s);
  end;
  TEnd(s1);
  with Marks[TSection(s)] do begin
    Overhead := Round(TotalTime/1000);
    Passes := 0;
    TotalTime := 0;
  end;
  ExternalOverhead := Round(Marks[n].PrevTime / 1000);
  Marks[n].Name := '';
end;

var
  i: Integer;
  Musec: Double;

initialization
finalization
  if ShowTimingAtEnd and (Length(Marks) > 0) then begin
    Musec := ClockFrequency/1e6;
    CalculateOverhead;
    WriteLn('Timing (microseconds) on a ', Round(Musec), ' MHz Processor');
    WriteLn('  Count  Fastest       Last    Average      Total');
    for i := 0 to High(Marks) do
      with Marks[i] do
        if Passes > 0 then begin
          Write(Passes:7, ' ', Fastest/Musec:10:1, ' ', PrevTime/Musec:10:1,  ' ', TotalTime/Passes/Musec:10:1, ' ', Round(TotalTime/Musec):10, ' ', Name);
          if Active then
            WriteLn(' (unclosed)')
          else
            WriteLn;
         end;
  end;
end.
