unit Sysinfo;

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}

interface

function is_buggy_pentium:Boolean;
function DetermineCpu:string;

implementation

uses {$IFDEF MSWINDOWS} Windows, Huge, {$ENDIF}
     {$IFDEF LINUX} Libc, {$ENDIF}
     {$IFDEF CLX_} Qt, {$ENDIF}
     SysUtils;

{$IFDEF CPU86}
{$IFDEF FPC}
{$ASMMODE intel}
{$ENDIF}
const
  numerator:Longint = 4195835;
  denominator:Longint = 3145727;

function is_buggy_pentium:Boolean;
var
  t:Smallint;
begin
  asm
    fild dword ptr numerator
    fidiv dword ptr denominator
    fimul dword ptr denominator
    fisub dword ptr numerator
    fistp word ptr t
  end;
  Result := t <> 0;
end;
{$ELSE}
function is_buggy_pentium:Boolean;
begin
  Result := False;
end;
{$ENDIF}

{$IFDEF LINUX}
function GetCpuName: string;
var
  buffer: array [0..600] of Char;
  f, l: Integer;
  function FindField(const Name: string): string;
  var
    s, e: PChar;
  begin
    s := strstr(buffer, PChar(Name));
    if s <> nil then begin
      s := strscan(s, ':');
      if s <> nil then begin
        Inc(s);
        e := strscan(s, #10);
        SetString(Result, s, e-s);
      end;
    end;
  end;
begin
  Result := '';
  f := open('/proc/cpuinfo', O_RDONLY);
  if f <> -1 then begin
    l := __read(f,buffer,High(buffer));
    if l > 0 then begin
      buffer[l] := #0;
    end;
    __close(f);
    Result := FindField('MHz');
    if Result <> '' then
      Result := Concat(Result, ' MHz');
    Result := Concat(FindField('model name'), Result);
  end;
end;

function DetermineCpu:string;
var
 buf:TUTSName;
 infobuf: string;
 {$IFDEF CLX_}
 function ModuleFileName(const Name:string):string;
 var
   module: HMODULE;
 begin
   Result := '';
   module := GetModuleHandle(PChar(Name));
   if module <> 0 then begin
     Result := GetModuleName(module);
   end;
   if Result = '' then Result := Name;
 end;
 {$ENDIF}
begin
  // TODO: Add cpu info
  infobuf := GetCpuName;
  uname(buf);
  Result := Format('%s %s %s %s %s%s glibc-%s-%s',
      [buf.sysname, buf.release, buf.version, buf.machine, buf.nodename, infobuf, gnu_get_libc_version, gnu_get_libc_release]);
 {$IFDEF CLX_}
 Result := Concat(Result, ' ', ModuleFileName(QtIntf));
  if Qt.QtIntf <> Qt.QtLib then
    Result := Concat(Result, ' ', ModuleFileName(QtLib));
 {$ENDIF}
end;
{$ENDIF}

{$IFDEF BSD}
function DetermineCpu:string;
begin
  Result := 'BSD'; { TODO: More info }
end;
{$ENDIF}

{$IFDEF WIN32}
function DetermineCpu:string;
var
  o:TOSVersionInfo;
  s:TSystemInfo;
  CPUVendorString: string[13];
  CPUid: Longint;
  Processor: Integer;
  HeapStatus: THeapStatus;
begin
  o.dwOSVersionInfoSize := SizeOf(o);
  GetVersionEx(o);
  case o.dwPlatformId of
     VER_PLATFORM_WIN32s:        Result := 'Win32s on Windows 3.1';
     VER_PLATFORM_WIN32_WINDOWS: Result := 'Win32c on Windows 95';
     VER_PLATFORM_WIN32_NT:      Result := 'Windows NT';
  else
    Result :='Windows ?';
  end;
  Result := Result + Format(' %d.%d.%d %s'#10#13, [o.dwMajorVersion, o.dwMinorVersion,
                            LOWORD(o.dwBuildNumber), o.szCSDVersion]);
  GetSystemInfo(s);
  processor := s.dwProcessorType;
  if is_buggy_pentium then begin
    Result := Result + 'Broken ';
    processor := 586;
  end;
  if processor >= 586 then begin
    CPUVendorString[0] := Chr(12);
    try
      asm
        push ebx
        xor ebx,ebx
        xor eax,eax
        db $0f, $a2  // CPUID
        or ebx,ebx
        je @@done
                  // After B0 stepping
        mov dword ptr CPUVendorString[1], ebx
        mov dword ptr CPUVendorString[5], edx
        mov dword ptr CPUVendorString[9], ecx
        mov eax,1
        db $0f, $a2  // CPUID
      @@done:
        mov CPUid, eax
      end;
    except
      CPUid := 0;
    end;
  end;
  if CPUid <> 0 then begin
    processor := ((CPUid shr 8) and $f)*100 + 86;
    Result := Result + Format('Processor: %d, Model: %d, Step: %d, %s',
                    [processor, (cpuid shr 4) and $f, cpuid and $f, CPUVendorString]);
  end else
    Result := Result + Format('Processor: %d', [processor]);
  HeapStatus := GetHeapStatus;
  Result := Result + Format(#10#13'Memory Used: %d', [HeapStatus.TotalAllocated + GMemoryCount]);
end;
{$ENDIF}

{$IFDEF DFS_WIN16}
function GetCPUID(var s: string): Longint;
var
  ID: string[13];
begin
  ID[0] := Chr(0);
  asm
    db $66  { 32 bit prefix }
    xor bx, bx
    db $66  { 32 bit prefix }
    xor ax, ax
    db $0f, $a2  { CPUID }
    db $66  { 32 bit prefix }
    or bx, bx
    je @done		{ After B0 stepping }
    db $66  { 32 bit prefix }
    mov word ptr ID[1],bx
    db $66  { 32 bit prefix }
    mov word ptr ID[5],dx
    db $66  { 32 bit prefix }
    mov word ptr ID[9],cx
    mov byte ptr ID[0], 12
    db $66  { 32 bit prefix }
    xor ax, ax
    db $66  { 32 bit prefix }
    xor ax, ax
    inc ax
    db $0f, $a2  { CPUID }
@done:
    db $66  { 32 bit prefix }
    mov word ptr Result, ax
  end;
  s := ID;
end;

function DetermineCpu:string;
var
  dwVersion, dwFlags, CPUid: Longint;
  Processor: Integer;
  OSType: string;
  CPUVendorString: string[13];
begin
  dwVersion := GetVersion;
  dwFlags := GetWinFlags;
  if (dwFlags and $4000) <> 0 then
    OSType := 'NT'
  else if (LOBYTE(LOWORD(dwVersion)) > 3) or (HIBYTE(LOWORD(dwVersion)) > 50) then
    OSType := '95'
  else
    OSType := '';
  Result := Format('Windows %s %d.%d ', [OSType, LOBYTE(LOWORD(dwVersion)), HIBYTE(LOWORD(dwVersion))]);
  if (dwFlags and WF_ENHANCED) <> 0 then
    Result := Result + 'enhanced mode '
  else
    Result := Result + 'standard mode ';
  if (dwFlags and WF_CPU286) <> 0 then
    Processor := 286
  else if (dwFlags and WF_CPU386) <> 0 then
    Processor := 386
  else if (dwFlags and WF_CPU486) <> 0 then
    Processor := 486
  else
    Processor := 0;
  if is_buggy_pentium then begin
    Result := Result + 'Broken ';
    processor := 586;
  end;
  Result := Result + #13#10;
  try
    CPUVendorString := '';
    CPUid := GetCPUID(CPUVendorString);
  except
    CPUid := 0;
  end;
  if CPUid <> 0 then begin
    processor := ((CPUid shr 8) and $f)*100 + 86;
    Result := Result + Format('Processor: %d, Model: %d, Step: %d, %s',
                    [processor, (cpuid shr 4) and $f, cpuid and $f, CPUVendorString]);
  end else
    Result := Result + Format('Processor: %d', [processor]);
  if (dwFlags and WF_80x87) <> 0 then
    Result := Result + ' + coprocessor'
end;
{$ENDIF}
end.
