unit TextAssign;
{ Connect TextFile output to a Stream or TStrings object }
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes, SysUtils;

procedure AssignStrings(var F: TextFile; Strings: TStrings);
procedure AssignStream(var F: TextFile; Stream: TStream);

implementation

uses UCompat;

{ Note that no error checking is done - all the actions are assumed to succeed }

function NoAction(Var t:TTextRec):Integer;
begin
  Result := 0;
end;

type
  TStringsData = record
    Strings: TStrings;
    Line: Integer;
  end;
  PStringsData = ^TStringsData;

function StringsReadFunc(var T:TTextRec):Integer;
begin
  with PStringsData(@T.UserData)^ do
    if Line >= Strings.Count then
      T.BufEnd := 0
    else begin
      { Note that this will trunacate the string if it is too long }
      T.BufEnd := StrLen(StrPLCopy(PChar(T.BufPtr), Strings[Line]+lineend, T.BufSize));
      Inc(Line);
    end;
  T.BufPos := 0;
  Result := 0;
end;

{ Write a string to the line number in the string list.
  If the string ends with a line ending, then bump the
  line number afterwards so the next output starts the next line.
  A line ending is forced to be LF only in open to avoid problems
  with a line ending spaning two calls to OutputStrings }
function StringsWriteFunc(var T:TTextRec):Integer;
var
  Str: string;
  NewLine: Boolean;
  s, p, e: PChar;
begin
  with T do
    { Is there anything to do? }
    if BufPos > 0 then
      with PStringsData(@UserData)^ do begin
        s := PChar(BufPtr);
        p := s;
        e := s + BufPos;
        repeat
          if p = e then
            NewLine := False
          else
            NewLine := p^ in [#10,#13];
          if NewLine or (p=e) then begin
            SetString(Str, s, p-s);
            { Appending to an existing string? }
            if Line < Strings.Count then
              Strings[Strings.Count-1] := Strings[Strings.Count-1] + Str
            else begin
              Strings.Add(Str);
              Line := Strings.Count-1; { Add does not have to cause Count to change by +1 }
            end;
            if NewLine then
              Line := Line + 1;
            s := p + 1;
            if (s < e) and (s^ in [#10,#13]) and (s^ <> p^) then
              Inc(s); { Discard other newline character }
            p := s;
          end else
            Inc(p);
        until p > e;
        BufPos := 0;
      end;
  Result := 0;
end;

function StringsOpenFunc(var T:TTextRec):Integer;
begin
  with T do begin
    if Mode = fmInput then begin
      InOutFunc := @StringsReadFunc;
      PStringsData(@UserData)^.Line := 0;
    end else begin
      InOutFunc := @StringsWriteFunc;
      FlushFunc := InOutFunc;
      if Mode = fmOutput then
        PStringsData(@UserData)^.Strings.Clear
      else { Mode = fmInOut }
        Mode := fmOutput;
      PStringsData(@UserData)^.Line := PStringsData(@UserData)^.Strings.Count;
    end;
  end;
  Result := 0;
end;

procedure ResetTextFile(var F: TextFile);
begin
  FillChar(F,SizeOf(TTextRec),0);
  with TTextRec(F) do begin
    Handle := {$IFDEF FPC} UnusedHandle {$ELSE} -1 {$ENDIF};
    Mode := fmClosed;
    OpenFunc := @NoAction;
    CloseFunc := @NoAction;
    InOutFunc := @NoAction;
    FlushFunc := @NoAction;
    BufSize := SizeOf(Buffer);
    BufPtr := @Buffer;
    {$IFDEF FPC}
    LineEnd := LineEnding;
    {$ENDIF}
  end;
end;

procedure AssignStrings(var F: TextFile; Strings: TStrings);
begin
  ResetTextFile(F);
  with TTextRec(F) do begin
    PStringsData(@UserData)^.Strings := Strings;
    Name := 'TStrings';
    OpenFunc := @StringsOpenFunc;
  end;
end;

type
  PStream = ^TStream;

function StreamReadFunc(var T:TTextRec):Integer;
begin
  T.BufEnd := PStream(@T.UserData)^.Write(T.BufPtr^, T.BufSize);
  T.BufPos := 0;
  Result := 0;
end;

function StreamWriteFunc(var T:TTextRec):Integer;
begin
  T.BufPos := T.BufPos - PStream(@T.UserData)^.Write(T.BufPtr^, T.BufPos);
  Result := 0;
end;

function StreamOpenFunc(var T:TTextRec):Integer;
begin
  with T do
    if Mode = fmInput then
      InOutFunc := @StreamReadFunc
    else begin
      InOutFunc := @StreamWriteFunc;
      FlushFunc := InOutFunc;
      if Mode = fmOutput then
        PStream(@T.UserData)^.Position := 0
      else { Mode = fmInOut }
        Mode := fmOutput;
    end;
  Result := 0;
end;

procedure AssignStream(var F: TextFile; Stream: TStream);
begin
  ResetTextFile(F);
  with TTextRec(F) do begin
    PStream(@UserData)^ := Stream;
    {$IFNDEF FPC}
    Flags := tfCRLF * Byte(DefaultTextLineBreakStyle);
    {$ENDIF}
    Name := 'Stream';
    OpenFunc := @StreamOpenFunc;
  end;
end;

end.
