unit Vector;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Huge;

type

  TVector = class(TGVectorN)
  private
    function GetValid(i:Longint):Boolean;
    procedure SetValid(i:Longint; IsValid:Boolean);
    function GetValue(i:Longint):Double;
    procedure SetValue(i:Longint; v:Double);
    procedure SetSize(n:Longint);
  public
    constructor Create(ASize: Longint);
    procedure Clear; { to NAN }
    function GetCell(i:Longint; var v: Double): Boolean;
    property Size read GetSize write SetSize;
    property Valid[i:Longint]:Boolean read GetValid write SetValid;
    property Value[i:Longint]:Double read GetValue write SetValue; default;
    procedure ClearRange(First, Count:Longint);
    function DotProduct(v:TGVectorN):Double;
    procedure Sqr;
    procedure Add(v: TGVectorN);
    procedure Zero;
    function FirstItem: PDouble;
  end;

  TColumnVector = class(TVector) { As TVector, but first index is 1 }
  public
    constructor Create(ASize: Longint);
  end;

implementation

uses MathSup, SysUtils;

constructor TVector.Create(ASize: Longint);
begin
  inherited Create(SizeOf(Double));
  SetSize(ASize);
end;

constructor TColumnVector.Create(ASize: Longint);
begin
  inherited Create(ASize);
  Offset := 1;
end;

procedure TVector.SetSize(n:Longint);
var
  Old: Longint;
begin
  Old := Size;
  inherited Size := n;
  if Size > Old then
    ClearRange(Old+Offset, Size-Old);
end;

procedure TVector.Clear;
begin
  ClearRange(Offset, Size);
end;

procedure TVector.ClearRange(First, Count:Longint);
var
  p:^Double;
{$IFDEF VER80}
  n:Word;
{$ENDIF}
begin
  p := Item[First];
  while Count > 0 do begin
{$IFNDEF VER80}
    SetNan(p^);
    Inc(p,1);
    Dec(Count);
{$ELSE}
    n := (65536-Ofs(p^))div 8;
    if n > Count then n := Count;
    asm
      les di,p
      add di,6
      mov ax,$7ff0
      mov cx,n
    @l:
      mov es:[di],ax
      add di,8
      dec cx
      jne @l
    end;
    HugeInc(Pointer(p), Longint(n)*8);
    Count := Count - n;
{$ENDIF}
  end;
end;

function TVector.GetValid(i:Longint):Boolean;
var
  p:^Double;
begin
  p := Item[i];
  Result := not IsNanOrInf(p^);
end;

procedure TVector.SetValid(i:Longint; IsValid:Boolean);
var
  p:^Double;
begin
  p := Item[i];
  if IsValid then
    p^ := 0
  else
    SetNan(p^);
end;

function TVector.GetValue(i:Longint):Double;
{$IFNDEF FPC}
type
  PDouble = ^Double;
{$ENDIF}
begin
  Result := PDouble(Item[i])^;
end;

procedure TVector.SetValue(i:Longint; v:Double);
var
  p:^Double;
begin
  p := Item[i];
  p^ := v;
end;

function TVector.DotProduct(v:TGVectorN):Double;
var
  p, p1: ^Double;
  i: Longint;
  Sum: Extended;
begin
  if v.Size <> Size then
    raise Exception.CreateFmt('TVector.DotProduct(%d,%d)', [Size, v.Size]);
  p := Base;
  p1 := v.Base;
  Sum := 0;
  for i := 1 to Size do begin
    Sum := Sum + p^*p1^;
    {$IFNDEF VER80}
      Inc(p);
      Inc(p1);
    {$ELSE}
      HugeInc(Pointer(p), Sizeof(Double));
      HugeInc(Pointer(p1), Sizeof(Double));
    {$ENDIF}
  end;
  Result := Sum;
end;

function TVector.FirstItem: PDouble;
begin
  Result := Base;
end;

function TVector.GetCell(i: Integer; var v: Double): Boolean;
var
  p: ^Double;
begin
  p := Item[i];
  Result := not IsNanOrInf(p^);
  System.Move(p^, v, SizeOf(Double));  { So that specific Invalid values can be seen }
end;

procedure TVector.Sqr;
var
  p: ^Double;
  i: Longint;
begin
  p := Base;
  for i := 1 to Size do begin
    p^ := System.Sqr(p^);
    {$IFNDEF VER80}
      Inc(p);
    {$ELSE}
      HugeInc(Pointer(p), Sizeof(Double));
    {$ENDIF}
  end;
end;

procedure TVector.Add(v: TGVectorN);
var
  p, p1: ^Double;
  i: Longint;
begin
  if v.Size <> Size then
    raise Exception.CreateFmt('TVector.Add(%d,%d)', [Size, v.Size]);
  p := Base;
  p1 := v.Base;
  for i := 1 to Size do begin
    p^ := p^ + p1^;
    {$IFNDEF VER80}
      Inc(p);
      Inc(p1);
    {$ELSE}
      HugeInc(Pointer(p), Sizeof(Double));
      HugeInc(Pointer(p1), Sizeof(Double));
    {$ENDIF}
  end;
end;

procedure TVector.Zero;
begin
  FillChar(FirstItem^, Size*SizeOf(Double), #0);
end;

end.
