unit f2c;
{$I DFS.inc}
interface

type
  DoubleComplex = record
    r: Double;
    i: Double;
  end;
  PDoubleComplex = ^DoubleComplex;
  FtnLen = Integer;
  LogBool = LongBool;

{$IFDEF VCL_}
{$DEFINE PROVIDE_G77}
function _sqrt(a: Double): Double; cdecl;
function _log(a: Double): Double; cdecl;
function _exp(a: Double): Double; cdecl;
function _sin(a: Double): Double; cdecl;
function _cos(a: Double): Double; cdecl;
function _tan(a: Double): Double; cdecl;
function _atan(a: Double): Double; cdecl;
function _atan2(a, b: Double): Double; cdecl;
{$ENDIF}
{$IFDEF FPC}
{$IFDEF LINUX}
{$DEFINE PROVIDE_GFORTRAN}
{$LINKLIB m}
{$ELSE}
{$DEFINE PROVIDE_G77}
{$ENDIF}
{$ENDIF}

{$IFDEF PROVIDE_GFORTRAN}
function gfortran_pow_r8_i4(a:Double; p:Integer): Double; cdecl;
procedure gfortran_concat_string(destlen: FtnLen; dest: PChar; len1: FtnLen; s1: PChar; len2: FtnLen; s2: PChar); cdecl;
function gfortran_compare_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar): Integer; cdecl;
procedure gfortran_copy_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar); cdecl;
{$DEFINE PROVIDE_G77}
{$ENDIF}

{$IFDEF PROVIDE_G77}
function _pow_di(var a:Double; var p:Integer): Double; cdecl;

procedure _z_div(var c, a, b: DoubleComplex); cdecl;
function _d_sign(var a, b: Double): Double; cdecl;
function _r_sign(var a, b: Single): Single; cdecl;

function _i_dnnt(var a: Double): Integer; cdecl;

function s_rnge(varn: PChar; offset: Integer; procn: PChar; line: Integer):Integer; cdecl;
function s_cat(lp: PChar; rpp: PPChar; rnp: PInteger; var np: Integer; ll: FtnLen):Integer; cdecl;
function s_cmp(a, b: PChar; la, lb: FtnLen): Integer; cdecl;
procedure s_copy(a, b: PChar; la, lb: FtnLen); cdecl;
{$ENDIF}

function errstop_(s: PChar; l: Integer):Integer; cdecl;
function errstopf_(fmt: PChar; args: Pointer { ... } ):Integer; cdecl;

function xerror_(MESSG: PChar; NMESSG, NERR, LEVEL: PInteger; len: Integer): Integer; cdecl;
function xerbla_(SRNAME: PChar; var Info: Integer; len: FtnLen): Integer; cdecl;
function dlamc3_(var a: Double; var b: Double): Double; cdecl;
function overflowtrapped_:Boolean; cdecl;
function ieeeck_(var ispec: Integer; var zero, one: double): integer;

function _tend_(var Id:Integer):Integer; cdecl;
function _tstart_(var Id:Integer):Integer; cdecl;

{$IFDEF VCL_}
function __ftol(x:Double):Integer; cdecl;

var
  __turboFloat: Integer;
{$ENDIF}

implementation

uses { Compat, } Math, Sysutils, Profile;

{$IFDEF FPC}
{$MACRO ON}
{$IFDEF DARWIN}
{$DEFINE alias_:=alias: #95}
{$ELSE}
{$DEFINE alias_:=alias:}
{$ENDIF}
{$ENDIF}

{$IFDEF VCL_}
function _sqrt(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'sqrt'; {$ENDIF}
begin
  Result := Sqrt(a);
end;

function _log(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'log'; {$ENDIF}
begin
  Result := Ln(a);
end;

function _exp(a: Double): Double; cdecl;
begin
  Result := Exp(a);
end;

function _sin(a: Double): Double; cdecl;
begin
  Result := Sin(a);
end;

function _cos(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'cos'; {$ENDIF} 
begin
  Result := Cos(a);
end;

function _tan(a: Double): Double; cdecl;
begin
  Result := Tan(a);
end;

function _atan(a: Double): Double; cdecl;
begin
  Result := ArcTan(a);
end;

function _atan2(a, b: Double): Double; cdecl;
begin
  Result := ArcTan2(a, b);
end;
{$ENDIF}

{$IFDEF PROVIDE_G77}
procedure _z_div(var c, a, b: DoubleComplex); cdecl;
var
  ratio, den : Double;
  abr, abi, cr: Double;
begin
  abr := Abs(b.r);
  abi := Abs(b.i);
  if abr <= abi then begin
    if abi = 0 then
      raise Exception.Create('complex division by zero');
    ratio := b.r / b.i ;
    den := b.i * (1 + ratio*ratio);
    cr := (a.r*ratio + a.i) / den;
    c.i := (a.i*ratio - a.r) / den;
  end else begin
    ratio := b.i / b.r ;
    den := b.r * (1 + ratio*ratio);
    cr := (a.r + a.i*ratio) / den;
    c.i := (a.i - a.r*ratio) / den;
  end;
  c.r := cr;
end;

function _d_sign(var a, b: Double): Double; cdecl;
begin
  if a >= 0 then
    Result := a
  else
    Result := -a;
  if b < 0 then
    Result := -Result;
end;

function _r_sign(var a, b: Single): Single; cdecl;
begin
  if a >= 0 then
    Result := a
  else
    Result := -a;
  if b < 0 then
    Result := -Result;
end;

function _i_dnnt(var a: Double): Integer; cdecl;
begin
  Result := Round(a);
end;

function _pow_di(var a:Double; var p:Integer): Double; cdecl;  {$IFDEF FPC} alias: '_pow_di'; {$ENDIF}
begin
  Result := IntPower(a,p);
end;

function s_rnge(varn: PChar; offset: Integer; procn: PChar; line: Integer):Integer; cdecl;
var
  msg: string;
begin
  msg := Format('Subscript out of range on file line %d, procedure %s', [line, procn]);
  msg := msg + {$IFDEF FPC}LineEnding{$ELSE}sLineBreak{$ENDIF} + Format('Attempt to access the %d-th element of variable %s', [offset+1, varn]);
  raise Exception.Create(msg);
end;

function s_cat(lp: PChar; rpp: PPChar; rnp: PInteger; var np: Integer; ll: FtnLen):Integer; cdecl; {$IFDEF FPC} alias: '_s_cat'; {$ENDIF}
var
  n, i, nc: Integer;
  rp: PChar;
  { Copy of f2c routine assuming NO_OVERWRITE defined, so lhs must not contain rhs }
begin
  n := np;
  for i := 0 to n-1 do begin
    nc := ll;
    if rnp^ < nc then
      nc := rnp^;
    Inc(rnp);
    Dec(ll, nc);
    rp := rpp^;
    Inc(rpp);
    Move(rp^, lp^, nc);
  end;
  FillChar(lp^, ll, ' ');
  Result := 0;
end;

function s_cmp(a, b: PChar; la, lb: FtnLen): Integer; cdecl; {$IFDEF FPC} alias: '_s_cmp'; {$ENDIF}
var
  aend, bend: PChar;
begin
  aend := a + la;
  bend := b + lb;
  if la <= lb then begin
    while a < aend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while b < bend do
      if b^ <> ' ' then begin
        Result := Ord(' ') - Ord(b^);
        Exit;
      end else
        Inc(b);
  end else begin
    while b < bend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while a < aend do
      if a^ <> ' ' then begin
        Result := Ord(a^) - Ord(' ');
        Exit;
      end else
        Inc(a);
  end;
  Result := 0;
end;

procedure s_copy(a, b: PChar; la, lb: FtnLen); cdecl; {$IFDEF FPC} alias: '_s_copy'; {$ENDIF}
{ Copy of f2c routine assuming NO_OVERWRITE defined, so lhs must not contain rhs }
begin
  if la <= lb then
    Move(b^, a^, la)
  else begin
    Move(b^, a^, lb);
    FillChar((a+lb)^, la-lb, ' ');
  end;
end;

{$ENDIF}

{$IFDEF PROVIDE_GFORTRAN}
function gfortran_pow_r8_i4(a:Double; p:Integer): Double; cdecl;  {$IFDEF FPC} alias: '_gfortran_pow_r8_i4'; {$ENDIF}
begin
  Result := IntPower(a,p);
end;

procedure gfortran_concat_string(destlen: FtnLen; dest: PChar; len1: FtnLen; s1: PChar; len2: FtnLen; s2: PChar); cdecl; {$IFDEF FPC} alias: '_gfortran_concat_string'; {$ENDIF}
begin
  if len1 <= destlen then
    Move(s1^, dest^, destlen)
  else begin
    Move(s1^, dest^, len1);
    Inc(dest,len1);
    Dec(destlen,len1);
    if len2 <= destlen then
      Move(s2, dest, destlen)
    else begin
      Move(s2^, dest^, len2);
      FillChar(dest^, destlen-len2, ' ');
    end;
  end;
end;

function gfortran_compare_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar): Integer; cdecl; {$IFDEF FPC} alias: '_gfortran_compare_string'; {$ENDIF}
begin
  Result := s_cmp(a, b, la, lb);
end;

procedure gfortran_copy_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar); cdecl;  {$IFDEF FPC} alias: '_gfortran_copy_string'; {$ENDIF}
begin
  s_copy(a, b, la, lb);
end;
{$ENDIF}

{$IFDEF VCL_}
function __ftol(x:Double):Integer; cdecl;
var
  cw: Word;
  r: Int64;
asm
  { Based on Borland C run time library }
        fstcw   cw                // save the control word
        fwait
        mov     ax, cw

        // the last bit in 0c01h disables the invalid fpu op exception
        // if the result overflows 64bits. this is done to remain
        // consistent with casts to 32bit ints which can overflow
        // w/o causing invalid fpu op.
        or      cw, $0C01               // set rounding control to chop
        fldcw   cw
        fistp   r                       // convert to 64-bit integer
        mov     cw, ax
        fldcw   cw                      // restore the control word
        mov     eax, dword ptr r        // return LS 32 bits
        mov     edx, dword ptr [r+4]    //        MS 32 bits
end;
{$ENDIF}

function errstop_(s: PChar; l: Integer):Integer; cdecl;
var
  str: string;
begin
  SetString(str, s, l);
  raise Exception.Create(str);
end;

function errstopf_(fmt: PChar; args: Pointer { ... } ):Integer; cdecl; {$IFDEF FPC} alias_'errstopf_'; {$ENDIF}
var
  w: string;
  p: PChar;
begin
  w := '';
  p := fmt;
  while p^ <> #0 do begin
    case p^ of
     '\': begin Inc(p); w := w + p^; end;
     '#', #0: Break;
     '%': begin
            Inc(p);
            case p^ of
              'd': w := w + IntToStr(PInteger(args)^);
              'g': w := w + FloatToStr(PSingle(args)^);
              'l': begin Inc(p); w := w + FloatToStr(PDouble(args)^); end;
            end;
            Inc(PChar(args),SizeOf(PChar));
          end;
    else
      w := w + p^;
    end;
    Inc(p);
  end;
  raise Exception.Create(w);
  Floor(0.1);
end;

function xerror_(MESSG: PChar; NMESSG, NERR, LEVEL: PInteger; len: FtnLen): Integer; cdecl;
{ SLATEC error routine }
begin
  raise Exception.CreateFmt('%.*s %d %d', [NMESSG^, MESSG, NERR^, LEVEL^]);
end;

function xerbla_(SRNAME: PChar; var Info: Integer; len: FtnLen): Integer; cdecl; {$IFDEF FPC} alias_'xerbla_'; {$ENDIF}
{ BLAS error routine }
var
  s: string;
begin
  SetString(s, SRNAME, len);
  raise Exception.CreateFmt('Error in %s: Parameter %d had an illegal value', [s, Info]);
end;

function overflowtrapped_:Boolean; cdecl; {$IFDEF FPC} alias_'overflowtrapped_'; {$ENDIF}
begin
  Result :=
  {$IFDEF CPUPOWERPC}
    True;
  {$ENDIF}
  {$IFDEF CPU86}
     exZeroDivide in GetExceptionMask;
  {$ENDIF}
end;

function dlamc3_(var a: Double; var b: Double): Double; cdecl; {$IFDEF FPC} alias_'dlamc3_'; {$ENDIF}
begin
  Result := a + b; { This routine must not keep any values in the floating point registers }
end;

function ieeeck_(var ispec: Integer; var zero, one: double): integer; {$IFDEF FPC} alias_'ieeeck_'; {$ENDIF}
begin
  { Copy of ieeeck routine from lapack }
  if overflowtrapped_ then
    Result := 0
  else
    Result := 1; { Assume IEEE behaviour }
end;

var
  TimingStrings: array of AnsiString;

function _tend_(var Id:Integer):Integer; cdecl;
begin
  TEnd(TimingStrings[Id]);
  Result := 0; { Not used }
end;

function _tstart_(var Id:Integer):Integer; cdecl;
begin
  if Id > High(TimingStrings) then
    SetLength(TimingStrings, Id+10);
  if Length(TimingStrings[Id]) = 0 then
    TimingStrings[Id] := Format('Fortran %d', [Id]);
  TStart(TimingStrings[Id]);
  Result := 0; { Not used }
end;

end.
