unit Detach;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

function ApplicationExecutable: string;
function Detach1(AppName: string; Argument: string):Boolean;

implementation

uses {$IFDEF MSWINDOWS} Windows, UCompat, {$ENDIF}
     {$IFNDEF CLX_} Forms, {$ELSE CLX_} QForms, {$ENDIF}
     {$IFDEF UNIX} {$IFNDEF DARWIN} Libc, {$ENDIF} CompatMsg, BaseUnix, Unix, {$ENDIF}
     SysUtils;

{$IFDEF UNIX}

var
  LastStatus: Integer; { TODO: These should be arrays }
  LastPid: Integer;

procedure HandleChildSignal(SigNum: Integer); cdecl;
begin
  LastPid := {$IFDEF KYLIX} Libc.waitpid {$ELSE} FpWaitpid {$ENDIF} (-1, @LastStatus, WNOHANG);
end;
{$ENDIF}

function ApplicationExecutable: string;
{ Replacement for Application.Exename, which does not return the full
  path of the application under KYLIX }
{$IFDEF LINUX}
var
  l: Integer;
  Buffer: array[0..FILENAME_MAX] of Char;
{$ENDIF}
begin
  Result := Application.ExeName;
  {$IFDEF LINUX}
  if (Length(Result) <= 1) or (Result[1] <> '/') then begin
    l := readlink('/proc/self/exe', Buffer, FILENAME_MAX);
    if l <> -1 then begin
      if l >= FILENAME_MAX then l := FILENAME_MAX-1;
      Buffer[l] := #0;
      SetString(Result, Buffer, l);
    end;
  end;
  {$ENDIF}
end;

function Detach1(AppName: string; Argument: string):Boolean;
var
  {$IFDEF MSWINDOWS}
  w: Word;
  cline: string;
  {$ELSE}
  {$IFDEF KYLIX}
  pid, wpid: __pid_t;
  PAppName, PArgument: PChar;
  status: Integer;
  {$ELSE}
  Wpid, Pid: TPid;
  Status: cInt;
  {$ENDIF}
  {$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  if Length(Argument) > 0 then
    cline := Concat(AppName, ' ', Argument)
  else
    cline := AppName;
  w := WinExec(AddressOfFirstChar(cline+#0), SW_SHOWNORMAL);
  Result := w >= 32;
  {$ELSE}
  {$IFDEF KYLIX}
    PAppName := PChar(AppName);
    if Length(Argument) > 0 then
      PArgument := PChar(Argument)
    else
      PArgument := nil;
    Pid := Libc.fork;
  {$ELSE}
    Pid := FpFork;
  {$ENDIF}
  if Pid < 0 then
    Result := False { Fork failed }
  else if pid = 0 then begin
    { In Child }
    {$IFDEF FPC}
      if Length(Argument) > 0 then
        FpExecLP(AppName, [Argument])
      else
        FpExecLP(AppName, []);
      Halt(127); { Somthing nasty happened }
    {$ELSE}
      Libc.execlp(PAppName, PAppName, PArgument, nil);
      Libc._Exit(127);
    {$ENDIF}
  end else begin
    Wpid := {$IFDEF KYLIX} Libc.waitpid {$ELSE} FpWaitpid {$ENDIF} (pid, @Status, WNOHANG);
    if (Wpid > 0) and not WIFEXITED(Status) then begin
      {$IFDEF KYLIX} Libc.__sleep(1); {$ELSE} Sleep(1000); {$ENDIF}
      Wpid := {$IFDEF KYLIX} Libc.waitpid {$ELSE} FpWaitpid {$ENDIF}(pid, @Status, WNOHANG);
    end;
    if (Wpid < 0) and (Pid = LastPid) then begin
      Wpid := LastPid;
      Status := LastStatus;
    end;
    if Wpid = 0 then begin
      // WriteLn(wpid, ' ', GetLastOSError, ' ', SysErrorMessage(GetLastOSError), ' ', AppName);
      {$IFDEF KYLIX}
      Libc.signal(SIGCHLD, HandleChildSignal);
      {$ELSE}
      FpSignal(SIGCHLD, @HandleChildSignal);
      {$ENDIF}
      Result := True; { Assume running ok }
    end else begin
      { Finished }
      if (Wpid > 0) and WIFEXITED(Status) then
        Result := WEXITSTATUS(Status) = 0
      else
        Result := False;
    end;
  end;
  {$ENDIF}
end;

end.
