unit ExpCan;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} Graphics, {$ELSE} QGraphics, {$ENDIF}
     {$IFDEF LCL_} PostScriptPrinter, {$ELSE} Types, {$ENDIF} Classes;

type

  TExportableCanvas = class
  protected
    procedure SetFont(const AValue: TFont); virtual;
    function GetFont: TFont; virtual; abstract;
    procedure SetPixels(X, Y: Integer; const Value: TColor); virtual; abstract;
  public
    function RealCanvas: TCanvas; virtual;
    function Pen: TPen; virtual; abstract;
    function Brush: TBrush; virtual; abstract;
    procedure Ellipse(X1, Y1, X2, Y2: Integer); virtual; abstract;
    procedure FillRect(const Rect: TRect); virtual; abstract;
    procedure Polyline(Points: array of TPoint; const nPoints: Integer); virtual; abstract;
    procedure Polygon(Points: array of TPoint; const nPoints: Integer); virtual; abstract;
    procedure RotatedTextOut(x, y: Integer; const str: string; RealTextAngle: Integer; Justify: TAlignment); virtual; abstract;
    procedure StretchDraw(var Rect: TRect; Graphic: TGraphic); virtual; abstract;
    procedure TextOut(X,Y: Integer; const Text: String); virtual;
    procedure TextRect(ARect: TRect; X, Y: integer; const Text : string); virtual;
    property Pixels[X, Y: Integer]: TColor write SetPixels;
    property Font: TFont read GetFont write SetFont;
  end;

  TCanvasCanvas = class(TExportableCanvas) { A real canvas }
  private
    FCanvas: TCanvas;
  protected
    function GetFont: TFont; override;
  public
    constructor Create(ACanvas: TCanvas);
    function RealCanvas: TCanvas; override;
    function Pen: TPen; override;
    function Brush: TBrush; override;
    procedure SetPixels(X, Y: Integer; const Value: TColor); override;
    procedure Ellipse(X1, Y1, X2, Y2: Integer); override;
    procedure FillRect(const Rect: TRect); override;
    procedure Polyline(Points: array of TPoint; const nPoints: Integer); override;
    procedure Polygon(Points: array of TPoint; const nPoints: Integer); override;
    procedure RotatedTextOut(x, y: Integer; const str: string; RealTextAngle: Integer; Justify: TAlignment); override;
    procedure StretchDraw(var Rect: TRect; Graphic: TGraphic); override;
    procedure TextOut(X,Y: Integer; const Text: String); override;
    procedure TextRect(ARect: TRect; X, Y: integer; const Text : string); override;
  end;

  TCustomStreamCanvas = class(TExportableCanvas)
  public
    constructor Create(AStream: TStream; Xmm, Ymm: Double); virtual;
    function DPI: Integer; { Assumed isotropic } virtual; abstract;
  end;

  TStreamCanvasClass = class of TCustomStreamCanvas;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF} SysUtils;

{ TCanvasCanvas }

function TCanvasCanvas.Brush: TBrush;
begin
  Result := FCanvas.Brush;
end;

constructor TCanvasCanvas.Create(ACanvas: TCanvas);
begin
  FCanvas := ACanvas;
end;

procedure TCanvasCanvas.Ellipse(X1, Y1, X2, Y2: Integer);
begin
  FCanvas.Ellipse(X1, Y1, X2, Y2);
end;

procedure TCanvasCanvas.FillRect(const Rect: TRect);
begin
  FCanvas.FillRect(Rect);
end;

function TCanvasCanvas.GetFont: TFont;
begin
  Result := FCanvas.Font;
end;

function TCanvasCanvas.Pen: TPen;
begin
  Result := FCanvas.Pen;
end;

procedure TCanvasCanvas.Polygon(Points: array of TPoint; const nPoints: Integer);
begin
  FCanvas.Polygon({$IFDEF FPC}@Points[0], nPoints, False{$ELSE}Slice(Points,nPoints){$ENDIF});
end;

procedure TCanvasCanvas.Polyline(Points: array of TPoint; const nPoints: Integer);
begin
  FCanvas.Polyline({$IFDEF FPC}Points, 0, nPoints{$ELSE}Slice(Points,nPoints){$ENDIF});
end;

function TCanvasCanvas.RealCanvas: TCanvas;
begin
  Result := FCanvas;
end;

procedure TCanvasCanvas.RotatedTextOut(x, y: Integer; const str: string;
  RealTextAngle: Integer; Justify: TAlignment);
begin
  raise Exception.Create('TCanvasCanvas.RotatedTextOut'); { Not called }
end;

procedure TCanvasCanvas.TextOut(X, Y: Integer; const Text: String);
begin
  FCanvas.TextOut(X, Y, Text);
end;

procedure TCanvasCanvas.TextRect(ARect: TRect; X, Y: integer; const Text: string);
begin
  FCanvas.TextRect(ARect, X, Y, Text);
end;

procedure TCanvasCanvas.SetPixels(X, Y: Integer; const Value: TColor);
begin
  FCanvas.Pixels[X,Y] := Value;
end;

procedure TCanvasCanvas.StretchDraw(var Rect: TRect; Graphic: TGraphic);
begin
  FCanvas.StretchDraw(Rect, Graphic);
end;

{ TExportableCanvas }

procedure TExportableCanvas.SetFont(const AValue: TFont);
begin
  GetFont.Assign(AValue);
end;

function TExportableCanvas.RealCanvas: TCanvas;
begin
  Result := nil;
end;

procedure TExportableCanvas.TextOut(X, Y: Integer; const Text: String);
begin
  RotatedTextOut(X, Y, Text, 0, taLeftJustify);
end;

procedure TExportableCanvas.TextRect(ARect: TRect; X, Y: integer; const Text: string);
begin
  TextOut(X, Y, Text);
end;

{ TCustomStreamCanvas }

constructor TCustomStreamCanvas.Create(AStream: TStream; Xmm, Ymm: Double);
begin
  inherited Create;
end;

end.
