unit MarkChse;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} Controls, Graphics, Messages,
                    Forms, Dialogs, StdCtrls, Menus, ExtCtrls, Buttons,
     {$ELSE} QControls, QGraphics, QForms, QDialogs, QStdCtrls, QMenus,
                    QExtCtrls, QButtons, Types, {$ENDIF}
     {$IFDEF LCL_} LCLType, LMessages, {$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     FrameGrp, SysUtils, NewSpin, GStyle, Classes, AlEdit, EditExtr;

type
  TMarkChooseItem = (pcColour, pcSize, pcType);
  TMarkChooseItems = set of TMarkChooseItem;

  TMarkChoose = class(TFrameGroupBox)
    Colour: TPanel;
    SizeLbl: TLabel;
    TypeLbl: TLabel;
    MarkType: TComboBox;
    PenColorDialog: TColorDialog;
    Size: TFixedSpin;
    Dot: TCheckBox;
    SetColour: TBitBtn;
    procedure SetColourClick(Sender: TObject);
    procedure SetMarkStyle(AStyle:TMarkStyle);
    procedure SizeChange(Sender: TObject);
    procedure MarkTypeChange(Sender: TObject);
    procedure DotClick(Sender: TObject);
    procedure DotKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DotMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  protected
    procedure Loaded; override;
    {$IFNDEF CLX_}
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CreateHandle; override;
    {$ELSE}
    procedure FontChanged; override;
    procedure EnabledChanged; override;
    {$ENDIF}
    procedure MarkTypeDrawItem({$IFDEF CLX_}Sender: TObject;{$ELSE} Sender: TWinControl;{$ENDIF} Index: Integer;
      Rect: TRect; State: TOwnerDrawState {$IFDEF CLX_}; var Handled:Boolean{$ENDIF});
  private
    FMarkStyle: TMarkStyle;
    FVisibleItems: TMarkChooseItems;
    FEnabledItems: TMarkChooseItems;
    FVertSpacing: Integer;
    procedure MarkStyleChanged(Sender: TObject);
    procedure SetEnabledItems(const Value: TMarkChooseItems);
    procedure SetVisibleItems(const Value: TMarkChooseItems);
    procedure UpdateEnabled;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    property MarkStyle:TMarkStyle read FMarkStyle write SetMarkStyle;
    property VisibleItems: TMarkChooseItems read FVisibleItems write SetVisibleItems;
    property EnabledItems: TMarkChooseItems read FEnabledItems write SetEnabledItems;

    property Caption;
    property Color;
    {$IFDEF MSWINDOWS}
    property Ctl3D;
    property ParentCtl3D;
    property DragCursor;
    {$ENDIF}
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;

  end;

var
  MarkChoose: TMarkChoose;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF}
     ExpCan, Gmarks;



{$IFDEF MSWINDOWS}
{$IFNDEF LCL_}
{$R *.dfm}
{$ENDIF}
{$ENDIF}
{$IFDEF CLX_}
{$R *.xfm}
{$ENDIF}

procedure TMarkChoose.SetColourClick(Sender: TObject);
begin
  PenColorDialog.Color := MarkStyle.Colour;
  if PenColorDialog.Execute then
    MarkStyle.Colour := PenColorDialog.Color;
end;

procedure TMarkChoose.MarkStyleChanged(Sender: TObject);
var
  i: Integer;
begin
  if csLoading in ComponentState then Exit;
  if MarkType.Items.Count < 2 then begin
    for i := 1 to nMarkTypes do
      MarkType.Items.Add('');
    MarkType.ItemIndex := Ord(MarkStyle.MarkType);
  end;
  Colour.Font := Font;
  if SetColour.Enabled then begin
    Colour.Color := MarkStyle.Colour;
    Colour.Font.Color := MarkStyle.Colour xor clWhite;
  end else begin
    Colour.Color := clBtnFace;
    Colour.Font.Color := MarkStyle.Colour;
  end;
  Size.Value := MarkStyle.Size;
  Dot.Checked := MarkStyle.DotAtCentre;
  MarkType.ItemIndex := Ord(MarkStyle.MarkType);
  Invalidate;
end;

procedure TMarkChoose.SetMarkStyle(AStyle:TMarkStyle);
begin
  FMarkStyle.Assign(AStyle);
end;

procedure TMarkChoose.Loaded;
begin
  inherited Loaded;
  MarkType.OnDrawItem := {$IFDEF FPC}@{$ENDIF}MarkTypeDrawItem;
  {$IFNDEF MSWINDOWS}
  MarkStyleChanged(Self);
  UpdateEnabled;
  {$ENDIF}
end;

procedure TMarkChoose.MarkTypeDrawItem({$IFDEF CLX_}Sender: TObject;{$ELSE} Sender: TWinControl;{$ENDIF} Index: Integer;
  Rect: TRect; State: TOwnerDrawState {$IFDEF CLX_}; var Handled:Boolean{$ENDIF});
var
  mp: TMarkParameters;
  x, y: Integer;
  c: TColor;
  ACanvas: TExportableCanvas;
begin
  {$IFDEF CLX_}
  Handled := Sender is TComboBox;
  if not Handled then Exit;
  Dec(Rect.Top);
  Rect.Right := Rect.Left + MarkType.ItemHeight;
  Rect.Bottom := Rect.Top + MarkType.ItemHeight;
  {$ENDIF}
  x := (Rect.Left + Rect.Right) div 2;
  y := (Rect.Top + Rect.Bottom) div 2;
  mp.v := (Rect.Bottom - Rect.Top) div 6;
  if (odSelected in State) and (odFocused in State) then
    c := clHighlightText
  else
    c := clWindowText;
  mp.h := mp.v;
  mp.BlobV := 0;
  mp.BlobH := 0;
  mp.IsRealDevice := True;
  mp.DotAtCentre := Dot.Checked;
  mp.rgbColour := ColorToRGB(c);
  mp.PreparePoints := nil;
  mp.ForceSymmetricalMarks := True;
  with (Sender as TComboBox).Canvas do begin { draw on the control canvas, not on the form }
    FillRect(rect);
    Pen.Color := c;
    Brush.Color := c;
    ACanvas := TCanvasCanvas.Create((Sender as TComboBox).Canvas);
    try
      Mark(ACanvas, x, y, TMarkType(Index), mp);
    finally
      ACanvas.Free;
    end;
  end;
end;

{$IFDEF FPC}
var
  ResourcesLoaded: Boolean;
{$ENDIF}
constructor TMarkChoose.Create(AOwner: TComponent);
begin
  {$IFDEF FPC}
  if not ResourcesLoaded then
    {$I MarkChse.lrs}
  ResourcesLoaded := True; { Work around the fact that the name and the icon have the same name }
  {$ENDIF}
  FVisibleItems := [pcColour, pcSize, pcType];
  FEnabledItems := [pcColour, pcSize, pcType];
  FMarkStyle := TMarkStyle.Create;
  FMarkStyle.OnChange := {$IFDEF FPC}@{$ENDIF}MarkStyleChanged;
  inherited Create(AOwner);
  ParentFont := True;
  FVertSpacing := Size.Top - (Colour.Top + Colour.Height);
  SetBounds(Left, Top, 0, 1);
end;

{$IFNDEF CLX_}
procedure TMarkChoose.CMFontChanged(var Message: TMessage);
begin
  inherited;
  MarkStyleChanged(Self);
end;

procedure TMarkChoose.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  UpdateEnabled;
end;

procedure TMarkChoose.CreateHandle;
begin
  inherited CreateHandle;
  MarkStyleChanged(Self);
  UpdateEnabled;
end;

{$ELSE}

procedure TMarkChoose.FontChanged;
begin
  inherited;
  MarkStyleChanged(Self);
end;

procedure TMarkChoose.EnabledChanged;
begin
  inherited;
  UpdateEnabled;
end;
{$ENDIF}

destructor TMarkChoose.Destroy;
begin
  FMarkStyle.Free;
  inherited Destroy;
end;

procedure TMarkChoose.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  cp, mp: Integer;
  procedure Ck(var A: Integer; l: Integer);
  begin
    if A < l then A := l;
  end;
begin
  if Colour <> nil then
    if AHeight > AWidth div 2  then begin
      { Portrait }
      Ck(AWidth, 2*Colour.Left + Colour.Width {$IFDEF LCL_} + Width - ClientWidth {$ENDIF});
      cp := Colour.Top;
      if Colour.Visible then
        Inc(cp, Colour.Height + FVertSpacing);
      if Size.Visible then begin
        Size.Left := Colour.Left + Colour.Width - Size.Width - Size.ExtraWidth;
        Size.Top := cp;
        SizeLbl.Left := Colour.Left;
        SizeLbl.Top := Size.Top + Size.Height div 2 - SizeLbl.Height div 2;
        cp := Size.Top + Size.Height + FVertSpacing;
      end;
      if MarkType.Visible then begin
        MarkType.Top := cp;
        TypeLbl.Left := Colour.Left;
        TypeLbl.Top := MarkType.Top + MarkType.Height div 2 - TypeLbl.Height div 2;
        MarkType.Left := TypeLbl.Left + TypeLbl.Width + 3;
        Dot.Left := Colour.Left + Colour.Width - Dot.Width;
        Dot.Top := MarkType.Top + MarkType.Height div 2 - Dot.Height div 2;
        cp := MarkType.Top + MarkType.Height + FVertSpacing;
      end;
      if not (csLoading in ComponentState) then
        Ck(AHeight, cp + FVertSpacing div 2 {$IFDEF LCL_} + Height - ClientHeight {$ENDIF});
    end else begin
      if Colour.Visible then begin
        Ck(AHeight, Colour.Top + Colour.Height + FVertSpacing {$IFDEF LCL_} + Height - ClientHeight {$ENDIF});
        mp := Colour.Top + Colour.Height div 2;
        cp := Colour.Left + Colour.Width + Colour.Left;
      end else begin
        Ck(AHeight, Colour.Top + MarkType.Height + FVertSpacing {$IFDEF LCL_} + Height - ClientHeight {$ENDIF});
        mp := Colour.Top + MarkType.Height div 2;
        cp := Colour.Left;
      end;
      if Size.Visible then begin
        SizeLbl.Left := cp;
        SizeLbl.Top := mp - SizeLbl.Height div 2;
        Size.Left := SizeLbl.Left + SizeLbl.Width + 3;
        Size.Top := mp - Size.Height div 2;
        cp := Size.Left + Size.Width + Size.ExtraWidth + Colour.Left;
      end;
      if MarkType.Visible then begin
        TypeLbl.Left := cp;
        TypeLbl.Top := mp - TypeLbl.Height div 2;
        MarkType.Left := TypeLbl.Left + TypeLbl.Width + 3;
        MarkType.Top := mp - MarkType.Height div 2;
        Dot.Left := MarkType.Left + MarkType.Width + 5;
        Dot.Top := mp - Dot.Height div 2;
        cp := Dot.Left + Dot.Width + Colour.Left;
      end;
      if not (csLoading in ComponentState) then Ck(AWidth, cp {$IFDEF LCL_} + Width - ClientWidth {$ENDIF});
    end;
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  Invalidate;
end;

procedure TMarkChoose.SizeChange(Sender: TObject);
begin
  MarkStyle.Size := Size.Value;
end;

procedure TMarkChoose.MarkTypeChange(Sender: TObject);
begin
  MarkStyle.MarkType := TMarkType(MarkType.ItemIndex);
end;

procedure TMarkChoose.DotClick(Sender: TObject);
begin
  MarkStyle.DotAtCentre := Dot.Checked;
  MarkType.Invalidate;
end;

procedure TMarkChoose.DotKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  MarkStyle.DotAtCentre := Dot.Checked;
  MarkType.Invalidate;
end;

procedure TMarkChoose.DotMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MarkStyle.DotAtCentre := Dot.Checked;
  MarkType.Invalidate;
end;

procedure TMarkChoose.SetVisibleItems(const Value: TMarkChooseItems);
begin
  if Value <> FVisibleItems then begin
    FVisibleItems := Value;
    Colour.Visible := pcColour in VisibleItems;
    SizeLbl.Visible := pcSize in VisibleItems;
    Size.Visible := pcSize in VisibleItems;
    TypeLbl.Visible := pcType in VisibleItems;
    MarkType.Visible := pcType in VisibleItems;
    Dot.Visible := pcType in VisibleItems;
    SetBounds(Left, Top, Width, Height);
  end;
end;

procedure TMarkChoose.SetEnabledItems(const Value: TMarkChooseItems);
begin
  if Value <> FEnabledItems then begin
    FEnabledItems := Value;
    SetColour.Enabled := pcColour in FEnabledItems;
    SizeLbl.Enabled := pcSize in EnabledItems;
    Size.Enabled := pcSize in EnabledItems;
    TypeLbl.Enabled := pcType in EnabledItems;
    MarkType.Enabled := pcType in EnabledItems;
    Dot.Enabled := pcType in EnabledItems;
    MarkStyleChanged(Self);
  end;
end;

procedure TMarkChoose.UpdateEnabled;
begin
  if csLoading in ComponentState then Exit;
  SetColour.Enabled := Enabled;
  if Enabled then
    ParentFont := True
  else
    Font.Color := clGrayText;
end;

{$IFDEF LCL_}
initialization
  {$I MarkChse.lrs}
{$ENDIF}
end.
