unit WCompat;

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
{ Contains various compatibility routines to cope with differences in operating systems, compilers
  and windowing routines }
interface

uses
  {$IFDEF VER80} Forms, {$ENDIF}
  {$IFDEF CLX_}
    Qt, QControls, QGraphics,
  {$ELSE}
    Messages, Graphics, Types,
  {$ENDIF}
  {$IFDEF FPC} DynLibs, LMessages, {$ENDIF}
  Classes, IniFiles, SysUtils;


{$IFDEF UNIX}
{$IFDEF CLX_}
const
  VK_UP = Key_Up;
  VK_DOWN = Key_Down;
  VK_LEFT = Key_Left;
  VK_RIGHT = Key_Right;
  VK_SPACE = Key_Space;
  VK_PRIOR = Key_Prior;
  VK_NEXT = Key_Next;
  VK_RETURN = Key_Return;
  VK_BACK = Key_Backspace;
  VK_DELETE = Key_Delete;
  VK_HOME = Key_Home;
  VK_END = Key_End;
  VK_INSERT = Key_Insert;
  VK_ESCAPE = Key_Escape;
  VK_SHIFT = Key_Shift;
  VK_CONTROL = Key_Control;
  VK_MENU = Key_Alt;
  VK_PAUSE = Key_Pause;
  VK_TAB = Key_Tab;
  VK_F1 = Key_F1;
  VK_F2 = Key_F2;
  VK_F3 = Key_F3;
  VK_F4 = Key_F4;
  VK_F5 = Key_F5;
  VK_F6 = Key_F6;
  VK_F7 = Key_F7;
  VK_F8 = Key_F8;
  VK_F9 = Key_F9;
  VK_F10 = Key_F10;
  VK_F11 = Key_F11;
  VK_F12 = Key_F12;
  CF_TEXT = 1;
  IDCANCEL = mrCancel;
  IDNO = mrNo;
  IDYES = mrYes;

  function MinimizeName(const Filename: TFileName; Canvas: TCanvas; MaxLen: Integer): TFileName;

  function DroppedFiles(Event: QDropEventH):TStringList;
{$ENDIF}
  function RGB(Red:Byte; Green:Byte; Blue:Byte): TColor;
  function GetRValue(Colour: TColor):Byte;
  function GetGValue(Colour: TColor):Byte;
  function GetBValue(Colour: TColor):Byte;
{$ENDIF}

{$IFDEF FPC}
{$IFNDEF MSWINDOWS}
 procedure DragAcceptFiles(Handle: THandle; Enabled: Boolean);
{$ENDIF}
{$ENDIF}

function ClipboardHasText: Boolean;

function WheelScrollLines: Integer;
{$ifdef VER90} {$define ADDWHeel} {$endif}
{$IFDEF VER80} {$define ADDWHeel} {$endif}
{$ifdef ADDWHeel}
const
  WHEEL_DELTA = 120;      { Default value for rolling one notch }
  WM_MOUSEWHEEL = $20A;  { message that will be supported by the OS }
{$endif}

{$IFDEF LCL_}
type
  TWMMouseWheel = TLMMouseEvent;
const
  WHEEL_DELTA = 1;      { Default value for rolling one notch }
  WM_MOUSEWHEEL = $20A;  { message that will be supported by the OS }
{$ENDIF}
{$IFNDEF DFS_DELPHI_4_UP} {$DEFINE DECL} {$ENDIF}
{$IFDEF DECL}
type
  TWMMouseWheel = record
{$IFNDEF VER80}
    Msg: Cardinal;
    Keys:  Word;
    WheelDelta: Smallint;
{$else}
    Msg: Cardinal;
    WheelDelta: Integer;	{ this is the wParam }
{$endif}
    case Integer of	        { two ways to look at the lParam }
      0: (
        XPos: Smallint;	        { either as x- and y-coordinates... }
        YPos: Smallint);
      1: (
        Pos: TSmallPoint;	{ ...or as a single point }
        Result: Longint);	{ and finally, the result field }
  end;
{$ENDIF}

{$IFDEF MSWINDOWS}
var
  uMSH_MOUSEWHEEL:Word;
{$ENDIF}

function DroppedFiles(var Message: TMessage):TStringList;

implementation

uses
  {$IFDEF CLX_}
    QTypes, QForms, QClipBrd, DosCommand, Xlib,
  {$ELSE}
    Controls, ClipBrd,
    {$IFDEF MSWINDOWS}
      ShellAPI, Windows, UCompat,
    {$ELSE}
      {$IFNDEF GTK2}
        gtk, gdk, glib,
      {$ELSE}
        gtk2, gdk2, glib2,
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
  Math;

{$IFDEF UNIX}

{$IFDEF CLX}
procedure CutFirstDirectory(var S: TFileName); { From Windows FileCtrl }
var
  Root: Boolean;
  P: Integer;
begin
  if S = '\' then
    S := '';
  begin
    if S[1] = '\' then begin
      Root := True;
      Delete(S, 1, 1);
    end else
      Root := False;
    if S[1] = '.' then
      Delete(S, 1, 4);
    P := AnsiPos('\',S);
    if P <> 0 then begin
      Delete(S, 1, P);
      S := '...\' + S;
    end else
      S := '';
    if Root then
      S := '\' + S;
  end;
end;

{ Linux equivalent to FileCtrl.MinimizeName
  - Trim directory names until the resulting string fits on the given Canvas }
function MinimizeName(const Filename: TFileName; Canvas: TCanvas; MaxLen: Integer): TFileName;
var
  Dir: TFileName;
  Name: TFileName;
begin
  Result := FileName;
  Dir := ExtractFilePath(Result);
  Name := ExtractFileName(Result);

  while (Dir <> '') and (Canvas.TextWidth(Result) > MaxLen) do begin
    if Dir = '\...\' then
      Dir := '...\'
    else
      CutFirstDirectory(Dir);
    Result := Dir + Name;
  end;
end;

function ClipboardHasText: Boolean;
begin
  Result := Clipboard.Provides('text/plain');
end;
{$ENDIF}

{$IFDEF FPC}
var
  Targets: array [0..2] of TGtkTargetEntry = ((target:'text/plain'; flags:0; info:0), (target:'text/uri-list'; flags:0; info:1), (target:'STRING'; flags:0; info:1));

function DropCallBack(widget: PGtkWidget; dc: PGdkDragContext; x, y: gint; selection: PGtkSelectionData; info, t: guint; data: gPointer): GBoolean; cdecl;
var
  list: PGList;
  Message: TMessage;
begin
  if (data <> nil) and (selection <> nil) and (selection^.length >= 0) then begin
    list := dc^.targets;
    while list <> nil do begin
      // WriteLn(gdk_atom_name(TGDKAtom(list^.data)),':',selection^.data);
      if (selection^.data <> nil) and (StrPos(PChar(selection^.data), 'file://') = PChar(selection^.data)) then begin
        Message.Msg := WM_DROPFILES;
        Message.lParam := LongInt(selection^.data);
        TWinControl(data).Dispatch(Message);
        Exit;
      end;
      list := list^.next;
    end;
  end;
end;

function DroppedFiles(var Message: TMessage):TStringList;
var
  i: Integer;
begin
  Result := TStringList.Create;
  try
    Result.Text := PChar(Message.lParam);
    for i := 0 to Result.Count-1 do
      if Pos('file://', Result[i]) = 1 then
        Result[i] := Copy(Result[i], 8, Length(Result[i]));
  except
    Result.Free;
    Raise;
  end;
end;

procedure DragAcceptFiles(Handle: THandle; Enabled: Boolean);
begin
  if Enabled then begin
    gtk_drag_dest_set(PGtkWidget(Handle), GTK_DEST_DEFAULT_ALL, @Targets[0], Length(Targets), GDK_ACTION_COPY);
    gtk_signal_connect(PGtkObject(Handle), 'drag_data_received', TGTKSignalFunc(@DropCallBack), gPointer(FindOwnerControl(Handle)));
  end else
    gtk_drag_dest_unset(PGtkWidget(Handle));
end;

{$ENDIF}
{$ENDIF}

{$IFDEF testselection}
function GetPrimarySelection: string;
begin
  XConvertSelection(QtDisplay, XA_PRIMARY, XA_STRING, {None} XA_STRING,
                    QWidget_winId(Application.MainForm.Handle), CurrentTime);
  Result := '';
end;
{$ENDIF}

{$IFNDEF CLX_}
function ClipboardHasText: Boolean;
begin
  Result := Clipboard.HasFormat(CF_TEXT);
end;
{$ENDIF}

{$IFNDEF VCL_}
function RGB(Red:Byte; Green:Byte; Blue:Byte): TColor;
begin
  Result := Red or (Green shl 8) or (Blue shl 16);
end;

function GetRValue(Colour: TColor):Byte;
begin
  Result := Colour and 255;
end;

function GetGValue(Colour: TColor):Byte;
begin
  Result := (Colour shr 8) and 255;
end;

function GetBValue(Colour: TColor):Byte;
begin
  Result := (Colour shr 16) and 255;
end;
{$ENDIF}

function WheelScrollLines: Integer;
begin
  Result :=
  {$IFDEF VCL_}
    Mouse.WheelScrollLines;
  if Result = 0 then
    Result := 3;
  {$ELSE}
    3;
  {$ENDIF}
end;

{$IFDEF MSWINDOWS}
function DroppedFiles(var Message: TMessage):TStringList;
var
  n, nFiles: Integer;
  Buffer: array [0..600] of Char;
begin
  Result := TStringList.Create;
  try
    try
      nFiles := DragQueryFile(Message.wParam, UINT(MinusOne), nil, 0);
      for n := 0 to nFiles-1 do begin
        DragQueryFile(Message.wParam, n, @Buffer[0], 270);
        Result.Add(StrPas(Buffer));
      end;
    finally
      DragFinish(Message.wParam);
      Message.Result := 0;
    end;
  except
    Result.Free;
    Raise;
  end;
end;
{$ENDIF}

{$IFDEF CLX_}
function DroppedFiles(Event: QDropEventH):TStringList;
var
  StringList: QStringListH;
begin
  Result := nil;
  StringList := QStringList_create;
  try
    QUriDrag_decodeLocalFiles(QDropEvent_to_QMimeSource(Event), StringList);
    Result := QStringListToTStringList(StringList);
  finally
    QStringList_destroy(StringList);
  end;
end;
{$ENDIF}

{$IFDEF MSWINDOWS}
const
  MSH_MOUSEWHEEL = 'MSWHEEL_ROLLMSG';
  { wParam = wheel rotation expressed in multiples of WHEEL_DELTA
    lParam is the mouse coordinates }

initialization
  uMSH_MOUSEWHEEL := RegisterWindowMessage(MSH_MOUSEWHEEL);
{$ENDIF}
end.


