unit MuScroll;

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses
  {$IFDEF VCL_} Windows, Messages, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Menus, Buttons, ExtCtrls, {$ENDIF}
  {$IFDEF CLX_} Types, QGraphics, QControls, QForms, QDialogs, QStdCtrls, QMenus, QButtons, QExtCtrls, {$ENDIF}
  SysUtils, Classes;

const
  InitRepeatPause = 400;  { pause before repeat timer (ms) }
  RepeatPause     = 100;  { pause before hint window displays (ms)}

type

  TTimerSpeedButton = class;

{ TMicroScroll }

  TMicroScroll = class (TWinControl)
  private
    FUpButton: TTimerSpeedButton;
    FDownButton: TTimerSpeedButton;
    FFocusedButton: TTimerSpeedButton;
    FFocusControl: TWinControl;
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
    function CreateButton: TTimerSpeedButton;
    function GetUpGlyph: TBitmap;
    function GetDownGlyph: TBitmap;
    procedure SetUpGlyph(Value: TBitmap);
    procedure SetDownGlyph(Value: TBitmap);
    procedure BtnClick(Sender: TObject);
    procedure BtnMouseDown (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SetFocusBtn (Btn: TTimerSpeedButton);
    procedure AdjustButtonSizes (var W: Integer; var H: Integer);
    {$IFDEF VCL_}
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    {$ENDIF}
  protected
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    {$IFDEF VCL_}
    procedure CMEnabledchanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    {$ENDIF}
    {$IFDEF CLX_}
    procedure EnabledChanged; override;
    {$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    property Align;
    property DownGlyph: TBitmap read GetDownGlyph write SetDownGlyph;
    {$IFDEF VCL_}
    property DragCursor;
    {$ENDIF}
    property DragMode;
    property Enabled;
    property FocusControl: TWinControl read FFocusControl write FFocusControl;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
{    property TabOrder;
    property TabStop; }
    property OnEnter;
    property OnExit;
    property UpGlyph: TBitmap read GetUpGlyph write SetUpGlyph;
    property Visible;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
  end;

{$IFDEF OLDSPINBOX}
  TCustomSpinBox = class(TCustomEdit)
  private
    FAlignment: TAlignment;    { CMW }
    FButton: TMicroScroll;
    FEditorEnabled: Boolean;
    LastErr: string;
    {$IFDEF VCL_}
    function GetMinHeight: Integer;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMExit(var Message: TCMExit);   message CM_EXIT;
    procedure WMPaste(var Message: TWMPaste);   message WM_PASTE;
    procedure WMCut(var Message: TWMCut);   message WM_CUT;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure WMGetDlgCode(Var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure SetAlignment(value: TAlignment);
    {$ENDIF}
  protected
    nVisibleDigits: Integer;
{$ifdef DFS_COMPILER_2_UP}
{$ifdef DFS_COMPILER_3_UP}
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
{$else}
    procedure GetChildren(Proc: TGetChildProc); override;
{$endif}
{$endif}
    function IsValidChar(Key: Char): Boolean; virtual;
    procedure UpClick (Sender: TObject);
    procedure DownClick (Sender: TObject);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure ChangeClick(n:LongInt);
    function SetAsText(s:string):Boolean;
    {$IFDEF VCL_}
    procedure SetEditRect; virtual;
    {$ENDIF}
    {$IFDEF CLX_}
    procedure BoundsChanged; override;
    {$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {$IFDEF VCL_}
    procedure CreateParams(var Params: TCreateParams); override;
    {$ENDIF}
    property Button: TMicroScroll read FButton;
    function IncrementBy(n: LongInt):Boolean; virtual; abstract;
    function CheckText(s:string):Boolean; virtual; abstract;
  published
    property AutoSelect;
    property AutoSize;
    property Color;
    {$IFDEF VCL_}
    property Ctl3D;
    property ParentCtl3D;
    property DragCursor;
    property Alignment: TAlignment read FAlignment write SetAlignment default taRightJustify;
    {$ENDIF}
    {$IFDEF CLX_}
    property Alignment;
    {$ENDIF}
    property DragMode;
    property EditorEnabled: Boolean read FEditorEnabled write FEditorEnabled default True;
    property Enabled;
    property Font;
    property MaxLength;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

{ TSpinBox }

  TOnIncrementInteger = procedure (Sender: TObject; var NewValue:Longint; LastValue:Longint; n:Integer) of object;

  TSpinBox = class(TCustomSpinBox)
  private
    FMinValue: LongInt;
    FMaxValue: LongInt;
    FIncrement: LongInt;
    FOnIncrement: TOnIncrementInteger;
    LastValue: LongInt;
    function GetValue: LongInt;
    procedure SetValue (NewValue: LongInt);
  public
    constructor Create(AOwner: TComponent); override;
    function IncrementBy(n: LongInt):Boolean; override;
    function CheckText(s:string):Boolean; override;
  published
    property Increment: LongInt read FIncrement write FIncrement default 1;
    property MaxValue: LongInt read FMaxValue write FMaxValue;
    property MinValue: LongInt read FMinValue write FMinValue;
    property Value: LongInt read GetValue write SetValue default 0;
    property OnIncrement: TOnIncrementInteger read FOnIncrement write FOnIncrement;
  end;

{ TFixedSpin }

  TOnIncrementFixed = procedure (Sender: TObject; var NewValue: Double; LastValue:Double; n:Integer) of object;

  TFixedSpin = class(TCustomSpinBox)
  private
    FMinValue: Double;
    FMaxValue: Double;
    FIncrement: Double;
    LastValue: Double;
    FPlaces: Integer;
    FOnIncrement: TOnIncrementFixed;
    function GetValue: Double;
    procedure SetValue(NewValue: Double);
    procedure SetPlaces(NewPlaces: Integer);
  protected
{$IFDEF VCL_}
    procedure SetEditRect; override;
{$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;
    function IncrementBy(n: LongInt):Boolean; override;
    function CheckText(s:string):Boolean; override;
  published
    property Increment: Double read FIncrement write FIncrement;
    property MaxValue: Double read FMaxValue write FMaxValue;
    property MinValue: Double read FMinValue write FMinValue;
    property Places: Integer read FPlaces write SetPlaces default 2;
    property Value: Double read GetValue write SetValue;
    property OnIncrement: TOnIncrementFixed read FOnIncrement write FOnIncrement;
  end;
{$ENDIF}

{ TTimerSpeedButton }

  TTimeBtnState = set of (tbFocusRect, tbAllowTimer);

  TTimerSpeedButton = class(TSpeedButton)
  private
    FRepeatTimer: TTimer;
    FTimeBtnState: TTimeBtnState;
    FOnPressed: TNotifyEvent;
    procedure TimerExpired(Sender: TObject);
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;  X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    destructor Destroy; override;
    procedure Press;
    property TimeBtnState: TTimeBtnState read FTimeBtnState write FTimeBtnState;
    property OnPressed: TNotifyEvent read FOnPressed write FOnPressed;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

{$IFDEF SINGLE}
{$IFNDEF VER80}
{$R MuScroll.d32}
{$ELSE}
{$R Muscroll.d16}
{$ENDIF}
{$ENDIF}

uses MathSup;

{ TMicroScroll }

constructor TMicroScroll.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  TabStop := False;
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] +
    [csFramed, csOpaque];
  { ControlStyle := ControlStyle + [csNoFocus]; }
  {$IFDEF VCL_}
  Ctl3D := False;
  ParentCtl3D := False;
  {$ENDIF}

  FUpButton := CreateButton;
  FDownButton := CreateButton;
  UpGlyph := nil;
  DownGlyph := nil;

  Width := 20;
  Height := 25;
  FFocusedButton := FUpButton;
end;

function TMicroScroll.CreateButton: TTimerSpeedButton;
begin
  Result := TTimerSpeedButton.Create (Self);
  { Result.ControlStyle := Result.ControlStyle + [csNoFocus]; }
  Result.OnMouseDown := BtnMouseDown;
  Result.OnPressed := BtnClick;
  {$IFDEF VCL_}
  Result.OnClick := BtnClick;
  {$ENDIF}
  Result.Visible := True;
  Result.Enabled := True;
  Result.TimeBtnState := [tbAllowTimer];
  Result.NumGlyphs := 1;
  Result.Parent := Self;
  {$IFNDEF VER80}
  Result.Transparent := False;
  {$ENDIF}
end;

procedure TMicroScroll.AdjustButtonSizes (var W: Integer; var H: Integer);
begin
  if (FUpButton = nil) or (csLoading in ComponentState) then Exit;
  if W < 13 then W := 13;
  FUpButton.SetBounds (0, 0, W, (H+1) div 2);   { CMW: even up heights }
  FDownButton.SetBounds (0, FUpButton.Height - (H and 1), W, (H+1) div 2);
end;

procedure TMicroScroll.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  AdjustButtonSizes (W, H);
  inherited SetBounds (ALeft, ATop, W, H);
end;

{$IFDEF VCL_}
procedure TMicroScroll.WMSize(var Message: TWMSize);
var
  W, H: Integer;
begin
  inherited;

  { check for minimum size }
  W := Width;
  H := Height;
  AdjustButtonSizes (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds(Left, Top, W, H);
  Message.Result := 0;
end;

procedure TMicroScroll.WMSetFocus(var Message: TWMSetFocus);
begin
  FFocusedButton.TimeBtnState := FFocusedButton.TimeBtnState + [tbFocusRect];
  FFocusedButton.Invalidate;
end;

procedure TMicroScroll.WMKillFocus(var Message: TWMKillFocus);
begin
  FFocusedButton.TimeBtnState := FFocusedButton.TimeBtnState - [tbFocusRect];
  FFocusedButton.Invalidate;
end;

procedure TMicroScroll.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;
{$ENDIF}

procedure TMicroScroll.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_UP:
      begin
        SetFocusBtn (FUpButton);
        FUpButton.Press;
      end;
    VK_DOWN:
      begin
        SetFocusBtn (FDownButton);
        FDownButton.Press;
      end;
    VK_SPACE:
      FFocusedButton.Press;
  end;
end;

procedure TMicroScroll.BtnMouseDown (Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then begin
    SetFocusBtn (TTimerSpeedButton (Sender));
    if (FFocusControl <> nil) and FFocusControl.TabStop and
        FFocusControl.CanFocus and
{$IFDEF VCL_}
        (GetFocus <> FFocusControl.Handle)
{$ENDIF}
{$IFDEF CLX_}
      (Screen.ActiveControl <> FFocusControl) // TODO - MSWIN ok also?
{$ENDIF}
      then
      FFocusControl.SetFocus
    else if TabStop and
{$IFDEF VCL_}
     (GetFocus <> Handle)
{$ENDIF}
{$IFDEF CLX_}
      (Screen.ActiveControl <> self) // TODO - MSWIN ok also?
{$ENDIF}
     and CanFocus then
      SetFocus;
  end;
end;

procedure TMicroScroll.BtnClick(Sender: TObject);
begin
  if Sender = FUpButton then
  begin
    if Assigned(FOnUpClick) then FOnUpClick(Self);
  end
  else
    if Assigned(FOnDownClick) then FOnDownClick(Self);
end;

procedure TMicroScroll.SetFocusBtn (Btn: TTimerSpeedButton);
begin
  if TabStop and CanFocus and  (Btn <> FFocusedButton) then
  begin
    FFocusedButton.TimeBtnState := FFocusedButton.TimeBtnState - [tbFocusRect];
    FFocusedButton := Btn;
    if
{$IFDEF VCL_}
     (GetFocus = Handle)
{$ENDIF}
{$IFDEF CLX_}
     (Screen.ActiveControl = self) // TODO - MSWIN ok also?
{$ENDIF}
    then begin
       FFocusedButton.TimeBtnState := FFocusedButton.TimeBtnState + [tbFocusRect];
       Invalidate;
    end;
  end;
end;

procedure TMicroScroll.Loaded;
var
  W, H: Integer;
begin
  inherited Loaded;
  W := Width;
  H := Height;
  AdjustButtonSizes (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
end;

function TMicroScroll.GetUpGlyph: TBitmap;
begin
  Result := FUpButton.Glyph;
end;

procedure SetDefaultGlyph(Button:TTimerSpeedButton; Up:Boolean);
var
  p:array [0..2] of TPoint;
  base: Integer;
begin
  with Button.Glyph do begin
    if Up then
      Base := 3
    else
      Base := 0;
    p[0].x := 0; { Was 1 }
    p[0].y := Base;
    p[1].x := 3; { Was 4 }
    p[1].y := 3 - Base;
    p[2].y := Base;
    p[2].x := 6; { Was 7 }
    {$IFDEF VCL_}
    Canvas.Brush.Color := clBlack;
    Width := 7; { Was 8 }
    {$ENDIF}
    {$IFDEF CLX_}
    Canvas.Brush.Color := clWhite;
    Canvas.Pen.Color := clWhite;
    Transparent := False;
    Width := 8; { Was 9 }
    {$ENDIF}
    Height := 5;
    Monochrome := True;
    Canvas.Polygon(p);
  end;
  Button.NumGlyphs := 1;
  Button.Invalidate;
end;

procedure TMicroScroll.SetUpGlyph(Value: TBitmap);
begin
  if Value <> nil then
    FUpButton.Glyph := Value
  else
    SetDefaultGlyph(FUpButton, True); { CMW }
end;

function TMicroScroll.GetDownGlyph: TBitmap;
begin
  Result := FDownButton.Glyph;
end;

procedure TMicroScroll.SetDownGlyph(Value: TBitmap);
begin
  if Value <> nil then
    FDownButton.Glyph := Value
  else
    SetDefaultGlyph(FDownButton, False); { CMW }
end;

{$IFDEF VCL_}
procedure TMicroScroll.CMEnabledchanged(var Message: TMessage);
{$ENDIF}
{$IFDEF CLX_}
procedure TMicroScroll.EnabledChanged;
{$ENDIF}
begin
  inherited;
  FUpButton.Enabled := Enabled;
  FDownButton.Enabled := Enabled;
end;

{ TCustomSpinBox }
{$IFDEF OLDSPINBOX}

constructor TCustomSpinBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csFixedHeight, csSetCaption];
  FButton := TMicroScroll.Create (Self);
  FButton.Width := 15;
  FButton.Visible := True;
  FButton.Parent := Self;
  FButton.FocusControl := Self;
  FButton.OnUpClick := UpClick;
  FButton.OnDownClick := DownClick;
  FEditorEnabled := True;
  FAlignment := taRightJustify;
  Height := 19;
end;

destructor TCustomSpinBox.Destroy;
begin
  FButton := nil;
  inherited Destroy;
end;

{$ifdef DFS_COMPILER_2_UP}
{$ifdef DFS_COMPILER_3_UP}
procedure TCustomSpinBox.GetChildren(Proc: TGetChildProc; Root: TComponent);
{$else}
procedure TCustomSpinBox.GetChildren(Proc: TGetChildProc);
{$endif}
begin
{ Proc(FButton); }
end;
{$endif}

procedure TCustomSpinBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_UP:    UpClick (Self);
    VK_DOWN:  DownClick (Self);
    VK_PRIOR: ChangeClick(10);
    VK_NEXT:  ChangeClick(-10); 
  else
    inherited KeyDown(Key, Shift);
  end;
end;

procedure TCustomSpinBox.KeyPress(var Key: Char);
begin
  if not IsValidChar(Key) then
  begin
    Key := #0;
{$IFDEF VCL_}
    MessageBeep(0);
{$ENDIF}
{$IFDEF CLX_}
    Beep;
{$ENDIF}
  end;
  if Key <> #0 then inherited KeyPress(Key);
end;

function TCustomSpinBox.IsValidChar(Key: Char): Boolean;
begin
  Result := (Key in [DecimalSeparator, '+', '-', '0'..'9', 'e', 'E']) or
    ((Key < #32) and (Key <> Chr(VK_RETURN)));
  if not FEditorEnabled and Result and ((Key >= #32) or
      (Key = Char(VK_BACK)) or (Key = Char(VK_DELETE))) then
    Result := False;
end;

{$IFDEF CLX_}
procedure TCustomSpinBox.BoundsChanged;
var
  R: TRect;
begin
  inherited;
  if FButton <> nil then begin
    R := Rect(Width - FButton.Width, 0, FButton.Width, Height);
    if BorderStyle <> bsNone then begin
      InflateRect(R, 0, -2);
      OffsetRect(R, -2, 0);
    end;
    FButton.BoundsRect := R;
  end;
end;
{$ENDIF}

{$IFDEF VCL_}
procedure TCustomSpinBox.SetAlignment(value: TAlignment);
begin
  if value <> FAlignment then
  begin
    FAlignment := value;
    RecreateWnd;
  end;
end;

procedure TCustomSpinBox.CreateParams(var Params: TCreateParams);
const
   Alignments:array[TAlignment] of {$IFDEF DFS_COMPILER_3_UP} DWORD
                                   {$ELSE} Longint {$ENDIF} = (ES_LEFT,ES_RIGHT,ES_CENTER); { CMW }
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style and not (ES_AUTOVSCROLL {or ES_AUTOHSCROLL});
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN or ES_AUTOHSCROLL or Alignments[FAlignment];
end;

procedure TCustomSpinBox.CMCtl3DChanged(var Message: TMessage);
begin
  inherited;
  SetEditRect;
end;

procedure TCustomSpinBox.CMFontChanged(var Message: TMessage);
begin
  inherited;
  SetEditRect;
end;

procedure TCustomSpinBox.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  { do not call inherited message handler, instead specify wanted keys here }
  Msg.Result := DLGC_HASSETSEL or DLGC_WANTARROWS or DLGC_WANTCHARS;
end;

procedure TCustomSpinBox.SetEditRect;
var
  Loc: TRect;
  DC: HDC;
  SaveFont: HFont;
  Metrics: TTextMetric;
  HBorder, VBorder: Integer;
  ZeroSize: TSize;
begin
  if not HandleAllocated then Exit;
  { Determine Border Widths }
  {  First force full height }
  Loc.Left := 0;
  Loc.Right := ClientWidth;
  Loc.Top := 0;
  Loc.Bottom := 200;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));
  HBorder := Loc.Left;
  VBorder := Loc.Top;
  { Get Text Metrics of current font }
  DC := GetDC(0);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  GetTextExtentPoint(DC, AddressOfFirstChar('0123456789'), 10, ZeroSize);
  nVisibleDigits := MulDiv(ClientWidth - FButton.Width - 2, 10, ZeroSize.cx);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  Loc.Top := (ClientHeight - Metrics.tmAscent) div 2 - 1 - VBorder;
  Loc.Bottom := Loc.Top + Metrics.tmHeight + VBorder*2;
  Loc.Right := ClientWidth - FButton.Width - 2 + HBorder;
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
  { SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));  { debug }
end;

procedure TCustomSpinBox.CMShowingChanged(var Message: TMessage);
begin
  inherited;
  SetEditRect;
end;

procedure TCustomSpinBox.WMSize(var Message: TWMSize);
var
  MinHeight: Integer;
begin
  inherited;
  MinHeight := GetMinHeight;
    { text edit bug: if size set to less than minheight, then edit ctrl does
      not display the text }
  if Height < MinHeight then
    Height := MinHeight
  else begin
    if FButton <> nil then begin
{$ifdef WIN32}
      if NewStyleControls then
        if Ctl3D then
          FButton.SetBounds(Width - FButton.Width - 4, 0, FButton.Width, Height - 4)
        else
          FButton.SetBounds(Width - FButton.Width, 1, FButton.Width, Height-1)
      else
{$endif}
        FButton.SetBounds(Width - FButton.Width, 0, FButton.Width, Height);
    end;
    SetEditRect;
  end;
end;

function TCustomSpinBox.GetMinHeight: Integer;
{$ifdef ORIGINAL}
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  Result := Metrics.tmHeight + I div 4 + GetSystemMetrics(SM_CYBORDER) * 4 + 2;
{$else}
begin
  Result := 0;
{$endif}
end;

procedure TCustomSpinBox.WMPaste(var Message: TWMPaste);
begin
  if not FEditorEnabled or ReadOnly then Exit;
  inherited;
end;

procedure TCustomSpinBox.WMCut(var Message: TWMPaste);
begin
  if not FEditorEnabled or ReadOnly then Exit;
  inherited;
end;

procedure TCustomSpinBox.CMExit(var Message: TCMExit);
begin
  inherited;
  if not CheckText(Text) then begin
    ShowMessage(LastErr);
    SetFocus;
  end;
end;

procedure TCustomSpinBox.CMEnter(var Message: TCMGotFocus);
begin
  if AutoSelect and not (csLButtonDown in ControlState) then
    SelectAll;
  inherited;
end;

{$ENDIF}

procedure TCustomSpinBox.ChangeClick (n:LongInt);
begin
  if ReadOnly then
{$IFDEF VCL_}
    MessageBeep(0)
{$ENDIF}
{$IFDEF CLX_}
    Beep
{$ENDIF}
  else begin
    if not CheckText(Text) then
      n := 0;
    if IncrementBy(n) then
      Modified := True;
  end;
end;

procedure TCustomSpinBox.UpClick (Sender: TObject);
begin
  ChangeClick(1);
end;

procedure TCustomSpinBox.DownClick (Sender: TObject);
begin
  ChangeClick(-1);
end;

function TCustomSpinBox.SetAsText(s:string):Boolean;
var
  SelEnd: Integer;
begin
  Result := CheckText(s);
  if Result then
    if HandleAllocated then begin { CMW }
      SelEnd := SelStart + SelLength;
      Text := s;
      SelStart := SelEnd;
    end else
      Text := s;
end;

{ TSpinBox }

constructor TSpinBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Text := '0';
  FIncrement := 1;
end;

function TSpinBox.GetValue: LongInt;
begin
  if CheckText(Text) then
    Result := LastValue
  else if csDesigning in ComponentState then begin
    MessageDlg(Lasterr, mtWarning, [mbOk], 0);
    Value := LastValue;
    Result := LastValue;
  end else
    raise Exception.Create(LastErr);
end;

function TSpinBox.IncrementBy(n: LongInt):Boolean;
var
  NewValue, PrevValue: Longint;
begin
  PrevValue := LastValue;
  NewValue := LastValue+n*FIncrement;
  if Assigned(FOnIncrement) then FOnIncrement(Self, NewValue, LastValue, n);
  LastErr := '';
  SetValue(NewValue);
  if LastErr <> '' then
    SetValue(LastValue);
  Result := PrevValue <> LastValue;
end;

procedure TSpinBox.SetValue (NewValue: LongInt);
begin
  SetAsText(IntToStr(NewValue));
end;

function TSpinBox.CheckText(s: string):Boolean;
begin
  Result := True;
  try
    s := Trim(s);
    if (s = '') or (s = '-') or (s = '.') or (s = '-.') then
      LastValue := 0
    else
      LastValue := StrToInt(s);
  except
    Result := False;
    LastValue := FMinValue;
    LastErr := Format('Bad Number: %s', [Text]);
  end;
  if Result and (FMaxValue > FMinValue) then
    if (LastValue < FMinValue) or (LastValue > FMaxValue) then begin
      Result := False;
      if LastValue < FMinValue then
        LastValue := FMinValue
      else
        LastValue := FMaxValue;
      LastErr := Format('%s out of range (%d to %d)', [Text, FMinValue, FMaxValue]);
    end;
end;

{ TFixedSpin }

constructor TFixedSpin.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Text := '0.00';
  FIncrement := 0.01;
  FPlaces := 2;
end;

function TFixedSpin.GetValue: Double;
begin
  if CheckText(Text) then
    Result := LastValue
  else if csDesigning in ComponentState then begin
    MessageDlg(Lasterr, mtWarning, [mbOk], 0);
    Value := LastValue;
    Result := LastValue;
  end else
    raise Exception.Create(LastErr);
end;

function TFixedSpin.IncrementBy(n: LongInt):Boolean;
var
  NewValue, PrevValue: Double;
begin
  PrevValue := LastValue;
  NewValue := LastValue+n*FIncrement;
  if Assigned(FOnIncrement) then FOnIncrement(Self, NewValue, LastValue, n);
  LastErr := '';
  SetValue(NewValue);
  if LastErr <> '' then
    SetValue(LastValue);
  Result := PrevValue <> LastValue;
end;

procedure TFixedSpin.SetValue (NewValue: Double);
begin
  if Places >= 0 then
    SetAsText(FloatToStrF(NewValue, ffFixed, 15, Places))
  else
    SetAsText(DtosM(NewValue, nVisibleDigits-1, -1));
end;

function TFixedSpin.CheckText(s: string):Boolean;
begin
  Result := True;
  try
    s := Trim(s);
    if (s = '') or (s = '-') or (s = '.') or (s = '-.') then
      LastValue := 0
    else
      LastValue := StrToFloat(s);
  except
    Result := False;
    LastValue := FMinValue;
    LastErr := Format('Bad Number: %s', [Text]);
  end;
  if Result and (FMaxValue > FMinValue) then
    if (LastValue < FMinValue) or (LastValue > FMaxValue) then begin
      Result := False;
      if LastValue < FMinValue then
        LastValue := FMinValue
      else
        LastValue := FMaxValue;
      LastErr := Format('%s out of range (%g to %g)', [Text, FMinValue, FMaxValue]);
    end;
end;

procedure TFixedSpin.SetPlaces(NewPlaces: Integer);
begin
  if NewPlaces <> Places then begin
    FPlaces := NewPlaces;
    CheckText(Text); { Notionally SetValue(Value), but without exceptions on loading }
    SetValue(LastValue);
    if not (csLoading in ComponentState) then begin
      FIncrement := 1;
      if NewPlaces > 18 then NewPlaces := 18;
      while NewPlaces > 0 do begin
        FIncrement := FIncrement / 10;
        Dec(NewPlaces);
      end;
    end;
  end;
end;

{$IFDEF VCL_} { // TODO: Why? }
procedure TFixedSpin.SetEditRect;
begin
  inherited SetEditRect;
  if Places < 0 then begin
    CheckText(Text); { Notionally SetValue(Value), but without exceptions on loading }
    SetValue(LastValue);
  end;
end;
{$ENDIF}

{$ENDIF}

{TTimerSpeedButton}

destructor TTimerSpeedButton.Destroy;
begin
  if FRepeatTimer <> nil then
    FRepeatTimer.Free;
  inherited Destroy;
end;

procedure TTimerSpeedButton.Press;
begin
  if Assigned(FOnPressed) then
    FOnPressed(Self);
end;

procedure TTimerSpeedButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  {$IFDEF CLX_}
  { Click is called at a bad time, so do it this way under linux }
  if Button = mbLeft then
    Press;
  {$ENDIF}
  if tbAllowTimer in FTimeBtnState then begin
    if FRepeatTimer = nil then
      FRepeatTimer := TTimer.Create(Self);

    FRepeatTimer.OnTimer := TimerExpired;
    FRepeatTimer.Interval := InitRepeatPause;
    FRepeatTimer.Enabled  := True;
  end;
end;

procedure TTimerSpeedButton.MouseUp(Button: TMouseButton; Shift: TShiftState;
                                  X, Y: Integer);
begin
  inherited MouseUp (Button, Shift, X, Y);
  if FRepeatTimer <> nil then
    FRepeatTimer.Enabled  := False;
end;

procedure TTimerSpeedButton.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := RepeatPause;
  if (FState = bsDown) and MouseCapture then
  begin
    try
      Press;
    except
      FRepeatTimer.Enabled := False;
      raise;
    end;
  end;
end;

procedure TTimerSpeedButton.Paint;
var
  R: TRect;
begin
  inherited Paint;
  if tbFocusRect in FTimeBtnState then
  begin
    R := Bounds(0, 0, Width, Height);
    InflateRect(R, -3, -3);
    if FState = bsDown then
      OffsetRect(R, 1, 1);
{$IFDEF VCL_}
    { TODO: Linux? }
    DrawFocusRect(Canvas.Handle, R);
{$ENDIF}
  end;
end;

{$IFDEF SINGLE}
procedure Register;
begin
  RegisterComponents('CMW', [TMicroScroll, TSpinBox, TFixedSpin]);
end;
{$ENDIF}

end.










